//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================

/**
*******************************************************************************
* @file  sdi.c
* @brief SDcard low level driver
*
* Copyright (C) 2007 Parrot S.A.
*
* @author     florent.bayendrian@parrot.fr
* @date       2007-01-08
*******************************************************************************
*/

#include <cyg/hal/hal_arch.h> /* CYGNUM_HAL_STACK_SIZE_TYPICAL */
#include <cyg/hal/drv_api.h>
#include <cyg/hal/hal_io.h> // HAL_WRITE_UINT32
#include <cyg/hal/s3c2412x.h> // interrupt controller reg
#include <cyg/kernel/kapi.h> // cyg_semaphore_post();
#include <cyg/utils/mmu.h>
#include <assert.h>

#include "sdi.h"
#include "sdi_dma.h"
#include "sdcard.h"

#define SDI_DMA_INT_PRIORITY 4
#define SDI_BASE_ADDRESS ((unsigned int *) 0x5a000000)
#define SDI_INTERRUPT 21
#define SDI_SUBSRCPND_BIT 13

// interrupt handlers
static cyg_handle_t sdcard_int_handle;
static cyg_interrupt intr;

static cyg_uint32 sdi_isr(cyg_vector_t vector, cyg_addrword_t data);
static void sdi_dsr(cyg_vector_t vector, cyg_ucount32 count, cyg_addrword_t data);

// synchronization variables
static cyg_sem_t sdcard_sem_cmd_done;
static cyg_sem_t sdcard_sem_data_done;
static cyg_mutex_t sdcard_mutex;
static cyg_mutex_t sdcard_data_mutex;

cyg_uint8 sdcard_dma_buffer[SDI_MAX_TRANSFER_SIZE] __attribute__ ((aligned(32)));

// Global SDI State var
SDI_sStateStruct SDI_State;


#define SDI_SET_DATA_CMD_INT_MASK() \
    SDI_REGISTER(SDI_REG_SDIINTMSK) = SDI_DATA_CMD_INTMASK;

#define SDI_DELAY 20

/**
 * Reset whole sdmmc block
 */
void SDI_Reset(void)
{
    cyg_uint32 reg = 0;

    BIT_SET(reg, SDI_SDICON_SDRESET, 1);
    SDI_REGISTER(SDI_REG_SDICON) = reg;
}

/**
 * Enable or Disable the SDI clock
 *
 * @param b enable or disable the clock
 * @return none
 */
void SDI_ClockEnable(cyg_bool b)
{
    cyg_uint32 reg;

    reg = SDI_REGISTER(SDI_REG_SDICON);
    if (b) {
        BIT_SET(reg, SDI_SDICON_ENCLK, 1);
        BIT_SET(reg, SDI_SDICON_CTYP, 1);
    } else {
        BIT_SET(reg, SDI_SDICON_ENCLK, 0);
        BIT_SET(reg, SDI_SDICON_CTYP, 0);
    }
    SDI_REGISTER(SDI_REG_SDICON) = reg;
}

/**
 * Initialize the SD interface and set up the interrupt callback.
 *
 */
void SDI_Init(void)
{
    SDI_Reset();
    cyg_mutex_init(&sdcard_mutex);
    cyg_mutex_init(&sdcard_data_mutex);
    cyg_semaphore_init(&sdcard_sem_cmd_done, 0);
    cyg_semaphore_init(&sdcard_sem_data_done, 0);

    SDI_State.Base = SDI_BASE_ADDRESS;
    SDI_State.eCmdState = SDI_STATE_CMD_NOINIT;
    SDI_State.eDataState = SDI_STATE_DATA_NOINIT;
    SDI_State.BlockLength = SDCARD_BLOCK_SIZE;
    SDI_State.it_count = 0;

    // interrupt init
    cyg_drv_interrupt_create(
            SDI_INTERRUPT,
            SDI_DMA_INT_PRIORITY,
            (cyg_addrword_t) 0,
            (cyg_ISR_t*) &sdi_isr,
            (cyg_DSR_t*) &sdi_dsr,
            &sdcard_int_handle,
            &intr);
    cyg_drv_interrupt_attach(sdcard_int_handle);
    cyg_drv_interrupt_unmask(SDI_INTERRUPT);

    SDI_ClockEnable(false);
    SDI_REGISTER(SDI_REG_SDIINTMSK) = SDI_CMD_INTMASK;
}

/**
 * Interrupt service routine used by eCos
 *
 * @param vector not used
 * @param data not used
 *
 * @return CYG_ISR_CALL_DSR
 */
static cyg_uint32 sdi_isr(cyg_vector_t vector, cyg_addrword_t data)
{
    cyg_drv_interrupt_mask(SDI_INTERRUPT);
    cyg_drv_interrupt_acknowledge(SDI_INTERRUPT);

    return CYG_ISR_CALL_DSR;
}

/**
 * Delayed Service Routine used by eCos
 *
 * @param vector not used
 * @param count not used
 * @param data not used
 */
static void sdi_dsr(cyg_vector_t vector, cyg_ucount32 count, cyg_addrword_t data)
{
    cyg_uint32 c_status = SDI_REGISTER(SDI_REG_SDICMDSTA);
    cyg_uint32 d_status = SDI_REGISTER(SDI_REG_SDIDATSTA);

    SDI_State.it_count++;

    if (d_status & (SDI_SDIDATSTA_DATFIN | SDI_SDIDATSTA_DATCRC | SDI_SDIDATSTA_CRCSTA | SDI_SDIDATSTA_DATTOUT)) {
        if (d_status & (SDI_SDIDATSTA_DATCRC | SDI_SDIDATSTA_CRCSTA)) {
            SDI_State.eLastDataError = SDI_ERROR_CRCFAIL;
        } else if (d_status & SDI_SDIDATSTA_DATTOUT) {
            SDI_State.eLastDataError = SDI_ERROR_TIMEOUT;
        } else {
            SDI_State.eLastDataError = SDI_ERROR_NONE;
        }
        SDI_State.eDataState = SDI_STATE_DATA_NOINIT;
        SDI_REGISTER(SDI_REG_SDIDATCON) = 0;
        SDI_REGISTER(SDI_REG_SDIFSTA) = 0xFFFFFFFF;
        SDI_REGISTER(SDI_REG_SDIINTMSK) = SDI_CMD_INTMASK;
        cyg_semaphore_post(&sdcard_sem_data_done);
    }

    // clear flags
    SDI_REGISTER(SDI_REG_SDIDATSTA) = d_status;

    // command interrupt
    if (c_status & SDI_SDICMDSTA_EVENT) {
        cyg_bool need_wakeup = false;
        switch (SDI_State.eCmdState) {
            case SDI_STATE_CMD_NOINIT:
                CYG_ASSERT(0, "Spurious command event");
            break;
            case SDI_STATE_CMD:
                // check errors
                if (c_status & SDI_SDICMDSTA_CMDTOUT) {
                    SDI_State.eLastCmdError = SDI_ERROR_TIMEOUT;
                } else if (c_status & SDI_SDICMDSTA_RESPCRC) {
                    SDI_State.eLastCmdError = SDI_ERROR_CRCFAIL;
                } else {
                    SDI_State.eLastCmdError = SDI_ERROR_NONE;
                }

                if ((c_status & SDI_SDICMDSTA_RSPFIN) && (SDI_State.cmd_response_type != SDI_RESPONSE_NONE)) {
                    // get the response from the SDI
                    if (SDI_State.cmd_response_index)
                        *SDI_State.cmd_response_index = BIT_GET(c_status, SDI_SDICMDSTA_RSPINDEX);

                    switch (SDI_State.cmd_response_type) {
                        case SDI_RESPONSE_SHORT:
                            if (SDI_State.cmd_response)
                                *SDI_State.cmd_response = SDI_REGISTER(SDI_REG_SDIRSP0);
                        break;
                        case SDI_RESPONSE_LONG:
                            if (SDI_State.cmd_response) {
                                SDI_State.cmd_response[0] = SDI_REGISTER(SDI_REG_SDIRSP0);
                                SDI_State.cmd_response[1] = SDI_REGISTER(SDI_REG_SDIRSP1);
                                SDI_State.cmd_response[2] = SDI_REGISTER(SDI_REG_SDIRSP2);
                                SDI_State.cmd_response[3] = SDI_REGISTER(SDI_REG_SDIRSP3);
                            }
                        break;
                        default:
                            CYG_ASSERT(0, "Unknown reponse type !!");
                        break;
                    }
                    need_wakeup = true;
                } else if ((c_status & SDI_SDICMDSTA_CMDSENT) && (SDI_State.cmd_response_type == SDI_RESPONSE_NONE)) {
                    need_wakeup = true;
                }
                if (SDI_State.eLastCmdError != SDI_ERROR_NONE) {
                    need_wakeup = true;
                }
                // clear flags
                SDI_REGISTER(SDI_REG_SDICMDSTA) = c_status & SDI_SDICMDSTA_EVENT;

                if (need_wakeup) {
                    SDI_State.eCmdState = SDI_STATE_CMD_NOINIT;
                    SDI_REGISTER(SDI_REG_SDICMDCON) = 0;
                    cyg_semaphore_post(&sdcard_sem_cmd_done);
                }
            break;
            default:
                CYG_ASSERT(0, "Invalid CMD state");
            break;
        }
    }

    // clear the corresponding flags in the interrupt controller
    HAL_WRITE_UINT32(SUBSRCPND, 1<<SDI_SUBSRCPND_BIT);
    HAL_WRITE_UINT32(SRCPND, 1<<SDI_INTERRUPT);
    HAL_WRITE_UINT32(INTPND, 1<<SDI_INTERRUPT);

    cyg_drv_interrupt_unmask(SDI_INTERRUPT);
}


/**
 * Send a SDCard command and wait for all done
 *
 * @param command_index : command
 * @param argument
 * @param response_type :    SDI_RESPONSE_NONE
 *                        or SDI_RESPONSE_SHORT (response_index + response(32bits))
 *                        or SDI_RESPONSE_LONG  (response_index + response(4*32bits))
 * @param response_index : used if SDI_RESPONSE_SHORT or SDI_RESPONSE_LONG was specified
 * @param response : used if SDI_RESPONSE_LONG was specified
 */
SDI_eError SDI_SendCommand(cyg_uint32 command_index, cyg_uint32 argument, SDI_eResponseType response_type,
                cyg_uint32 * response_index, cyg_uint32 * response)
{
    cyg_uint32 command_reg = 0;

    cyg_mutex_lock(&sdcard_mutex);

    SDI_State.eCmdState = SDI_STATE_CMD;
    SDI_State.cmd_response_type = response_type;
    SDI_State.cmd_response_index = response_index;
    SDI_State.cmd_response = response;

    SDI_REGISTER(SDI_REG_SDICMDARG) = argument;

    BIT_SET(command_reg, SDI_SDICMDCON_CMDINDEX, command_index | SDI_SDICMDCON_SENDERHOST);
    BIT_SET(command_reg, SDI_SDICMDCON_CMST, 1);
    switch (response_type) {
        case SDI_RESPONSE_NONE:
        break;
        case SDI_RESPONSE_SHORT:
            BIT_SET(command_reg, SDI_SDICMDCON_WAITRSP, 1);
        break;
        case SDI_RESPONSE_LONG:
            BIT_SET(command_reg, SDI_SDICMDCON_WAITRSP, 1);
            BIT_SET(command_reg, SDI_SDICMDCON_LONGRSP, 1);
        break;
        default:
            assert(0);
        break;
    }
    SDI_REGISTER(SDI_REG_SDICMDCON) = command_reg;

    // wait for transfer and reponse if needed
    cyg_semaphore_wait(&sdcard_sem_cmd_done);

    cyg_mutex_unlock(&sdcard_mutex);

    return SDI_State.eLastCmdError;
}

/**
 * Set the current frequency of the interface
 *
 * @param freq frequency in Hz
 */
void SDI_SetClockFreq(cyg_uint32 freq)
{
    if (freq != 0) {
        int clkdiv = SDI_MCLK_SPEED/freq - 1;

        if ((clkdiv > 0) && (clkdiv < (1<<bwSDI_SDIPRE_VALUE))) {
            SDI_REGISTER(SDI_REG_SDIPRE) = clkdiv;
        } else {
            SDI_REGISTER(SDI_REG_SDIPRE) = 1;
        }
    } else {
        SDI_ClockEnable(false);
    }
}

/*
 * Set data bus width to 4 or 1
 *
 * @return void
 */
void SDI_SetWideBus(cyg_bool wide)
{
    SDI_State.wide_bus = wide;
}

/**
 * Get the current frequency of the interface
 *
 * @return frequency
 */
cyg_uint32 SDI_GetClockFreq(void)
{
    cyg_uint32 clkdiv;
    cyg_uint32 speed;

    clkdiv = BIT_GET(SDI_REGISTER(SDI_REG_SDIPRE), SDI_SDIPRE_VALUE);
    assert(clkdiv);
    speed = SDI_MCLK_SPEED/(clkdiv+1);

    return speed;
}

/**
 * Send data to the SDcard
 *
 * @param data data to send
 * @param length in bytes
 * @param data_mode block or stream mode
 *
 * @return SDI_eError
 */
SDI_eError SDI_SendData(const cyg_uint8 * data, cyg_uint32 length, SDI_eDataMode data_mode)
{
    cyg_uint32 speed;
    cyg_uint32 reg = 0;
    SDI_eError ret = SDI_ERROR_NONE;

    assert(length <= SDI_MAX_TRANSFER_SIZE && length != 0);

    cyg_mutex_lock(&sdcard_data_mutex);

    if ((cyg_uint32)data & 0x3) {
        memcpy(sdcard_dma_buffer, data, length);
        SDI_State.data = sdcard_dma_buffer;
    } else {
        SDI_State.data = (cyg_uint8*) data;
    }

    SDI_State.eDataState = SDI_STATE_DATA_SENDING;
    SDI_State.length = length;

    speed = SDI_GetClockFreq();
    // 250ms of delay before the timeout
    SDI_REGISTER(SDI_REG_SDIDTIMER) = 5*(speed/20);

    SDI_REGISTER(SDI_REG_SDIBSIZE) = SDI_State.BlockLength;

    BIT_SET(reg, SDI_SDIDATCON_BLKNUM, length/SDI_State.BlockLength);
    BIT_SET(reg, SDI_SDIDATCON_DTST, 0x1); // data start
    BIT_SET(reg, SDI_SDIDATCON_DATMODE, 0x3); // data transmit start
    if (SDI_State.wide_bus) {
        BIT_SET(reg, SDI_SDIDATCON_WIDEBUS, 1);
    }
    BIT_SET(reg, SDI_SDIDATCON_BLKMODE, 1);
    BIT_SET(reg, SDI_SDIDATCON_RACMD, 1);
    BIT_SET(reg, SDI_SDIDATCON_ENDMA, 1);
    BIT_SET(reg, SDI_SDIDATCON_DATASIZE, 0x2); // word
    sdi_dma_devconfig(SDI_State.data, SDI_State.length, false);
    SDI_REGISTER(SDI_REG_SDIDATCON) = reg;
    SDI_SET_DATA_CMD_INT_MASK();

    // wait for all done
    cyg_semaphore_wait(&sdcard_sem_data_done);
    if (sdi_dma_finish()) {
        ret = SDI_ERROR_TIMEOUT;
    } else {
        ret = SDI_State.eLastDataError;
    }

    cyg_mutex_unlock(&sdcard_data_mutex);

    return ret;
}

/**
 * Wait for some data with a timeout
 *
 * @param data
 * @param length in bytes
 * @param data_mode SDI_DATAMODE_BLOCK or SDI_DATAMODE_STREAM
 *
 * @return SDI_eError
 */
SDI_eError SDI_OpenReceiveWindow(cyg_uint8 * data, cyg_uint32 length, SDI_eDataMode data_mode)
{
    cyg_uint32 speed;
    cyg_uint32 reg = 0;

    assert((length != 0) && (length%SDI_State.BlockLength == 0));
    assert(length <= SDI_MAX_TRANSFER_SIZE);
    assert(data);

    cyg_mutex_lock(&sdcard_data_mutex); // unlock by SDI_CloseReceiveWindow

    SDI_State.eDataState = SDI_STATE_DATA_RECEIVING;
    SDI_State.data = data;
    SDI_State.length = length;

    speed = SDI_GetClockFreq();
    // 100ms of delay before the timeout
    SDI_REGISTER(SDI_REG_SDIDTIMER) = speed/10;

    SDI_REGISTER(SDI_REG_SDIBSIZE) = SDI_State.BlockLength;

    BIT_SET(reg, SDI_SDIDATCON_BLKNUM, length/SDI_State.BlockLength);
    BIT_SET(reg, SDI_SDIDATCON_DTST, 0x1); // data start
    BIT_SET(reg, SDI_SDIDATCON_DATMODE, 0x2); // data receive start
    if (SDI_State.wide_bus) {
        BIT_SET(reg, SDI_SDIDATCON_WIDEBUS, 1);
    }
    BIT_SET(reg, SDI_SDIDATCON_BLKMODE, 1);
    BIT_SET(reg, SDI_SDIDATCON_RACMD, 1);
    BIT_SET(reg, SDI_SDIDATCON_ENDMA, 1);
    BIT_SET(reg, SDI_SDIDATCON_DATASIZE, 0x2); // word
    sdi_dma_devconfig(sdcard_dma_buffer, SDI_State.length, true);
    SDI_REGISTER(SDI_REG_SDIDATCON) = reg;
    SDI_SET_DATA_CMD_INT_MASK();

    return SDI_ERROR_NONE;
}

/**
 * Wait for the transfer done or a timeout
 *
 * @return SDI_eError
 */
SDI_eError SDI_CloseReceiveWindow(void)
{
    SDI_eError ret = SDI_ERROR_NONE;

    cyg_semaphore_wait(&sdcard_sem_data_done);
    if (sdi_dma_finish())
        ret = SDI_ERROR_TIMEOUT;

    if (!ret) {
        arm_mmu_invalidate_cache_range((CYG_ADDRESS)sdcard_dma_buffer, SDI_State.length);
        memcpy(SDI_State.data, sdcard_dma_buffer, SDI_State.length);
        ret = SDI_State.eLastDataError;
    }
    cyg_mutex_unlock(&sdcard_data_mutex);

    return ret;
}
