//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================

/**
**************************************************************
* @file parrot6_i2c.c
* @brief I2C master driver for the Parrot6. <br>
* Theses low-level functions are called by the Ecos I2C generic package.
*
* Copyright (C) 2006 Parrot S.A.
*
* @author   gautier.seidel@parrot.fr
* @date     2006-02-22
*
**************************************************************
*/

#include <cyg/infra/cyg_ass.h>
#include <cyg/infra/diag.h>         // diagnostic output
#include <cyg/io/i2c.h>

#include <pkgconf/devs_io_i2c_arm_parrot6.h>
#include <cyg/io/i2c_parrot6.h>

////////////////////////////////////////////////////
// Parrot proprietary I2C_MASTER registers
////////////////////////////////////////////////////
#define I2CM_TRANSMIT       0x000       // I2C_MASTER Transmit Register
#define I2CM_RECEIVE        0x000       // I2C_MASTER Receive Register
#define I2CM_ITEN           0x004       // I2C_MASTER Interrupt Enable Register
#define I2CM_COMMAND        0x008       // I2C_MASTER Command Register
#define I2CM_STATUS         0x008       // I2C_MASTER Status Register
#define I2CM_PRESCALE       0x00C       // I2C_MASTER Prescale Register

// Registers bitwise definitions

// Command Register
#define I2CM_COMMAND_ITACK      (1 << 0)     // Clears a pending interrupt
#define I2CM_COMMAND_ACK        (0 << 1)    // ACK bit command
#define I2CM_COMMAND_NACK       (1 << 1)    // NACK bit command
#define I2CM_COMMAND_WR         (1 << 2)    // Write bit command
#define I2CM_COMMAND_RD         (1 << 3)    // Read bit command
#define I2CM_COMMAND_STO        (1 << 4)    // Stop bit command
#define I2CM_COMMAND_STA        (1 << 5)    // Start bit command

// Interrupt Enable register
#define I2CM_ITEN_ITEN                1  // Interrupt enabled

// Transmit register
#define I2CM_TRANSMIT_WRIT            0
#define I2CM_TRANSMIT_READ            1

//! write to an i2c master register
#define i2cm_write_reg(_base_, _reg_, _value_) \
 (*((volatile cyg_uint32 *)((_base_)+(_reg_))) = (cyg_uint32)(_value_))

//! read an i2c master register
#define i2cm_read_reg(_base_, _reg_) \
 (*((volatile cyg_uint32 *)((_base_)+(_reg_))))

// i2c driver error codes
enum {
    I2C_ERR_OK = 0,
    I2C_ERR_AL,
    I2C_ERR_NACK,
    I2C_ERR_TIMEOUT,
};

// support for the debug
#if 1
# define DEBUG(_format_, ...)
#else
# define DEBUG(format, ...) diag_printf("%s(%d): " format, __FUNCTION__, __LINE__, ## __VA_ARGS__)
#endif

#if 1
# define DEBUG_ERR(_format_, ...)
#else
# define DEBUG_ERR(format, ...) diag_printf("%s(%d): " format, __FUNCTION__, __LINE__, ## __VA_ARGS__)
#endif

#define I2C_TIMEOUT          (1000000)
/* event timeout in ticks */
#define I2C_EVENT_TIMEOUT    (CYGNUM_DEVS_IO_I2C_ARM_PARROT6_EVENT_TIMEOUT)

#ifdef CYGNUM_DEVS_IO_I2C_ARM_PARROT6_USE_IRQS
static const int parrot6_i2cm_use_irqs = 1;
#else
static const int parrot6_i2cm_use_irqs = 0;
#endif

static
struct {
    size_t                      count;
    struct parrot6_i2c_data*    extras[CYGNUM_DEVS_IO_I2C_ARM_PARROT6_BUS_MAX];
} parrot6_i2c_drv = { 0 };


static void parrot6_i2c_prepare_completion(void)
{
#if 0
    // do not attempt to sleep in cyg_user_start() context
    if (parrot6_i2cm_use_irqs &&
        (cyg_thread_self() != cyg_thread_idle_thread())) {
        // enable interrupt
        i2cm_write_reg(I2CM_ITEN, I2CM_ITEN_ITEN);
    }
#endif
}

static cyg_uint32 parrot6_i2c_wait_completion(struct parrot6_i2c_data* extra)
{
    cyg_uint32          timeout = I2C_TIMEOUT;
    CYG_ADDRESS const   base = extra->base;

    if (parrot6_i2cm_use_irqs &&
        (cyg_thread_self() != cyg_thread_idle_thread())) {

        // sleep and wait for an event
        extra->status = (cyg_uint32)
                        cyg_flag_timed_wait(&extra->wait_flag,
                                            (cyg_flag_value_t) ~(0UL),
                                            CYG_FLAG_WAITMODE_OR |
                                            CYG_FLAG_WAITMODE_CLR,
                                            cyg_current_time() +
                                            I2C_EVENT_TIMEOUT);
#if 0
        // disable interrupt
        i2cm_write_reg(base, I2CM_ITEN, 0);
#endif

        if (! extra->status)
            timeout = 0;
    }
    else {
        do {
            extra->status = i2cm_read_reg(base, I2CM_STATUS);
            timeout--;
        } while ((extra->status & I2CM_STATUS_TIP) && timeout);
    }

    return timeout;
}


/** actually send a byte
 *
 * @param mash
 * @param dev The I2C device you want communicate with. dev is the name of the
 * device, not the address
 * @param data The byte to sent
 * @param start True if a start should be performed at the beginning
 * @param stop True if a stop should be performed at the end
 *
 * @return -I2C_ERR_AL if an arbitrage lost occured, -I2C_ERR_NACK for a NACK, I2C_ERR_OK for an ACK.
 *
 */
static int parrot6_i2c_send_byte ( cyg_i2c_bus* mash,
                                        const cyg_i2c_device* dev,
                                        cyg_uint8 data,
                                        cyg_bool send_start,
                                        cyg_bool send_stop)
{
    cyg_uint32                      cmd;
    struct parrot6_i2c_data* const  extra = (struct parrot6_i2c_data*)
                                            mash->i2c_extra;
    CYG_ADDRESS const               base = extra->base;

    DEBUG("dev=%08x addr=%02x send_start=%d data=%08x send_stop=%d\n",
          dev, dev->i2c_address, send_start, data, send_stop);

    // build the command:
    cmd = I2CM_COMMAND_WR;               // we at least want to write

    if (send_start) {
        cmd |= I2CM_COMMAND_STA ;          // start needed ?
    }

    if (send_stop) {
        cmd |= I2CM_COMMAND_STO;           // stop needed ?
    }

    // put the data to be send in the Tx register
    i2cm_write_reg(base, I2CM_TRANSMIT, data);

    // enable completion interrupt
    parrot6_i2c_prepare_completion();

    // send the command ( write + others )
    i2cm_write_reg(base, I2CM_COMMAND, cmd);

    // wait until the end of the tx
    if (! parrot6_i2c_wait_completion(extra))
        return -I2C_ERR_TIMEOUT;

    // return error code.
    if (extra->status & I2CM_STATUS_AL) {
        // return an arbitrage lost error.
        DEBUG_ERR("arbitration lost\n");
        return -I2C_ERR_AL;
    }

    if (extra->status & I2CM_STATUS_RXNACK) {
        // return nack.
        DEBUG_ERR("no acknoledge\n");
        return -I2C_ERR_NACK;
    }

    // no error.
    return I2C_ERR_OK;
}

/** actually read a byte
 *
 * @param mash
 * @param dev The I2C device you want communicate with. dev is the name of the
 * device, not the address
 * @param nak True if a nak should be performed at the end
 * @param stop True if a stop should be performed at the end
 *
 * @return The byte read or if an error occurred: -I2C_ERR_AL, -I2C_ERR_TIMEOUT.
 */
static int parrot6_i2c_read_byte(cyg_i2c_bus* mash,
                                 const cyg_i2c_device* dev,
                                 cyg_bool nak,
                                 cyg_bool send_stop)
{
    int                             ret;
    cyg_uint32                      cmd;
    struct parrot6_i2c_data* const  extra = (struct parrot6_i2c_data*)
                                            mash->i2c_extra;
    CYG_ADDRESS const               base = extra->base;

    DEBUG("dev=%08x nak=%d\n", dev, nak);

    // build the command:
    cmd = I2CM_COMMAND_RD;              // we at least want to read
    if (nak) {                          // nak needed ?
        cmd |= I2CM_COMMAND_NACK;
    }
    if (send_stop) {                     // stop needed ?
        cmd |= I2CM_COMMAND_STO;
    }

    // enable completion interrupt
    parrot6_i2c_prepare_completion();

    // start the reception
    i2cm_write_reg(base, I2CM_COMMAND, cmd);

    // wait until the end of the rx
    if (! parrot6_i2c_wait_completion(extra))
        return -I2C_ERR_TIMEOUT;

    // return error code.
    if (extra->status & I2CM_STATUS_AL) {
        // return an arbitrage lost error.
        DEBUG_ERR("arbitration lost\n");
        return -I2C_ERR_AL;
    }

    // no error, return byte.
    ret = i2cm_read_reg(base, I2CM_RECEIVE);

    // trace it if needed
    DEBUG("result %02x\n", ret);

    // return what we read
    return ret;
}

/** send a stop
 *
 * used for compatibility
 *
 * @param dev The device. Not used because the stop is performed for the whole
 * bus
 */
void parrot6_i2c_send_stop(cyg_i2c_bus* bus)
{
    struct parrot6_i2c_data* const  extra = (struct parrot6_i2c_data*)
                                            bus->i2c_extra;
    CYG_ADDRESS const               base = extra->base;

    DEBUG("parrot6 i2c send stop alone\n");

    // enable completion interrupt
    parrot6_i2c_prepare_completion();

    // start the reception
    i2cm_write_reg(base, I2CM_COMMAND, I2CM_COMMAND_STO);

    // wait until the end of the processing
    if (! parrot6_i2c_wait_completion(extra))
        DEBUG_ERR("timeout while sending stop!\n");

    if (extra->status  & I2CM_STATUS_AL)
        DEBUG_ERR("arbitration lost\n");
}


/** ************************************
 ** Now the api for upper layers      **
 ** ************************************
 */

//! I2C master interrupt service routine (ISR)
static cyg_uint32 parrot6_i2cm_isr(cyg_vector_t vector, cyg_addrword_t data)
{
    // simply request DSR
    cyg_drv_interrupt_mask(vector);
    return CYG_ISR_CALL_DSR;
}

//! I2C master interrupt deferred service routine (DSR)
static void parrot6_i2cm_dsr(cyg_vector_t vector, cyg_ucount32 count,
                             cyg_addrword_t data)
{
    unsigned int cnt;

    for (cnt = 0; cnt < parrot6_i2c_drv.count; cnt++) {
        struct parrot6_i2c_data* const  extra = parrot6_i2c_drv.extras[cnt];
        cyg_uint32 const                status = i2cm_read_reg(extra->base,
                                                               I2CM_STATUS);

        if (! (status & I2CM_STATUS_IF))
            continue;

        // acknowledge interrupt
        i2cm_write_reg(extra->base, I2CM_COMMAND, I2CM_COMMAND_ITACK);
#if 0
        // disable interrupt
        i2cm_write_reg(base, I2CM_ITEN, 0);
#endif

        // wake-up blocked thread
        cyg_flag_setbits(&extra->wait_flag, (cyg_flag_value_t) status);
    }

    cyg_drv_interrupt_unmask(vector);
}

/** i2c master initialization
 *
 * The prescale factor is fixed. Future evolution is to choose it in the
 * configtool
 *
 * @param bus The bus used. There is only one bus in the Parrot6, so not used.
 */
void parrot6_i2c_init( cyg_i2c_bus* bus)
{
    static int                      irq_attached = 0;
    struct parrot6_i2c_data* const  extra = (struct parrot6_i2c_data*)
                                            bus->i2c_extra;
    struct pclock const* const      clk = pclock_get(extra->clk_name);
    unsigned int                    prescale;

    DEBUG("parrot6 i2c init\n" );

    CYG_ASSERT(parrot6_i2c_drv.count < CYGNUM_DEVS_IO_I2C_ARM_PARROT6_BUS_MAX,
               "Too many registered buses");
    cyg_flag_init(&extra->wait_flag);
    parrot6_i2c_drv.extras[parrot6_i2c_drv.count++] = extra;

    // enable I2C master clock
    pclock_enable(clk);
    prescale = pclock_get_rate(clk) / extra->hz;
    prescale = ((prescale / 5) - 1) & 0xfff;
    i2cm_write_reg(extra->base, I2CM_PRESCALE, prescale);

    // install interrupt handler
    if (parrot6_i2cm_use_irqs) {
        // enable interrupt
        i2cm_write_reg(extra->base, I2CM_ITEN, I2CM_ITEN_ITEN);

        if (! irq_attached) {
            cyg_handle_t handle;
            static cyg_interrupt interrupt;

            cyg_drv_interrupt_create(CYGNUM_HAL_INTERRUPT_I2CM,
                                     4,
                                     (cyg_addrword_t)0,
                                     parrot6_i2cm_isr,
                                     parrot6_i2cm_dsr,
                                     &handle,
                                     &interrupt);
            cyg_drv_interrupt_attach(handle);
            irq_attached = 1;
            cyg_drv_interrupt_unmask(CYGNUM_HAL_INTERRUPT_I2CM);
        }
    }
}


/** perform an i2c transmission
 *
 * @param dev The device name
 * @param send_start True if a start should be performed at beginning. If true
 * tha I2C address of the device is also sended
 * @param tx_data A pointer to the data to send
 * @param count The number of bytes to send
 * @param send_stop True if a stop should be send at the end of the transmission
 *
 * @return The bytes written or if an error occurred: -I2C_ERR_AL, -I2C_ERR_NACK.
 */
cyg_uint32 parrot6_i2c_tx(const cyg_i2c_device* dev,
                          cyg_bool send_start,
                          const cyg_uint8* tx_data,
                          cyg_uint32 count,
                          cyg_bool send_stop)
{
    int status;
    cyg_uint32 result;

    CYG_CHECK_DATA_PTR(tx_data,
                       "i2c tx operation requested but no data supplied");
    CYG_PRECONDITION(count > 0, "at least one byte should be sent");

    DEBUG("dev=%08x addr=%02x send_start=%d buf=%08x count=%d "
          "send_stop=%d\n", dev, dev->i2c_address, send_start,
          tx_data, count, send_stop);

    // first send the start & the address if needed
    if (send_start) {

        // if we send the start, we also have to send the address shifted left,
        //  complete with 0 to indicate a write. No stop sended
        // the devide should respond an AK, otherwise we stop
        // last args: true for the start, false the stop
        status = parrot6_i2c_send_byte(dev->i2c_bus,
                                       dev,
                                       (dev->i2c_address << 1)|0,
                                       true,
                                       false);
        if (status < I2C_ERR_OK) {
            // we don't receive the ack. we stop the tx now
            DEBUG_ERR("an error (%d) occurred with device at address %02x\n",
                  status, dev->i2c_address);
            return (cyg_uint32)status;
        }
    }

    // now send the data, unless we receive a nak.
    for (result = 0; result < count; result++) {
        // two last args: false for the start, stop if needed
        status = parrot6_i2c_send_byte(dev->i2c_bus,
                                       dev,
                                       tx_data[result],
                                       false,
                                       ((result == count-1) && send_stop));

        if (status < I2C_ERR_OK) {
            // return an error code.
            DEBUG_ERR("an error (%d) occurred with device at address %02x\n",
                  status, dev->i2c_address);
            return (cyg_uint32)status;
        }
    }

    DEBUG("parrot6 i2c tx, %d bytes sent out of %d\n", result, count);

    // return the number of bytes sended.
    return result;
}


/** perform an rx from a device (= a read action)
 *
 * @param dev The device name
 * @param send_start True if a start should be performed at beginning. If
 * true the I2C address of the device is also sended.
 * @param rx_data A pointer to the buffer where data should be stored
 * @param count The number of bytes to read
 * @param send_nak True if a nack should be send at the end of the transmission.
 * @param send_stop True if a stop should be send at the end of the
 * transmission.
 *
 * @return The number of readed bytes or if an error occurred: -I2C_ERR_AL, -I2C_ERR_NACK, -I2C_ERR_TIMEOUT
 *
 */
cyg_uint32 parrot6_i2c_rx(const cyg_i2c_device* dev,
                          cyg_bool send_start,
                          cyg_uint8* rx_data, cyg_uint32 count,
                          cyg_bool send_nak,
                          cyg_bool send_stop)
{
    int status, byte;
    cyg_uint32 result  = 0;

    DEBUG("dev %08x, addr %02x, send_start %d, buf %08x, "
          "count %d, send_nak %d, send_stop %d\n",
          dev, dev->i2c_address, send_start, rx_data, count, send_nak,
          send_stop);

    CYG_CHECK_DATA_PTR(rx_data,
                       "i2c rx operation requested but no data supplied");
    CYG_PRECONDITION(count > 0, "at least one byte should be sent");
    CYG_PRECONDITION(send_nak || !send_stop,
                     "a stop can only be generated after rx has been nak'd");

    // first send the start & the address if needed
    if (send_start) {

        // if we have to send the start, we also have to send the address
        // shifted left,
        //  complete with 1 to indicate a read.
        // the device should respond an AK, otherwise we stop
        status = parrot6_i2c_send_byte(dev->i2c_bus,
                                       dev,
                                       (dev->i2c_address << 1)|0x01,
                                       true,
                                       false);
        if (status < I2C_ERR_OK) {
            // we don't receive the ack or an arbitrage lost occurred. we stop the tx now.
            DEBUG_ERR("an error (%d) occurred with device at address %02x\n",
                  status, dev->i2c_address);
            return (cyg_uint32)status;
        }
        // The device has ack'd so we can continue sending
    }

    // The device has ack'd. Read the number of bytes specified. The
    // device cannot stop sending, although it may clock-stretch. For
    // the last byte we usually want to NAK
    // but if the data stream is variable length with e.g.
    // the first byte indicating the length then we want to support
    // partial reads.
    for (result = 0; result < count; result++) {
        byte = parrot6_i2c_read_byte(dev->i2c_bus, dev,
                                     (send_nak && (result == (count-1))),
                                     (send_stop && (result == (count-1))));
        if (byte < 0) {
            // return error code.
            return (cyg_uint32)byte;
        }
        rx_data[result] = (cyg_uint8)byte;
    }

    DEBUG("parrot6 i2c rx, read %d bytes\n", result);
    return result;
}

/** set the bus in hte right state, outside tx or rx
 *
 * @param dev The device name, to perform the stop on the right bus.
 */
void parrot6_i2c_stop(const cyg_i2c_device* dev)
{
    parrot6_i2c_send_stop(dev->i2c_bus);
}
