/**
 ****************************************************************************
 * Driver for Cirrus Logic CS42XX stereo audio codec.
 *
 * @file   cs42xx_hal.h
 * @brief  Hardware Abstraction Layer for CS42XX driver.
 *
 * Code is based on the following datasheets:
 * http://www.cirrus.com/en/pubs/proDatasheet/CS4245_F1.pdf
 * http://www.cirrus.com/en/pubs/proDatasheet/CS4265_F1.pdf
 * http://www.cirrus.com/en/pubs/proDatasheet/CS5345_F1.pdf
 *  
 * This code has been mainly inspired from the GPIO button driver
 * (by Ivan Djelic, Parrot S.A.).
 *  
 * @author David Tylski
 * @date   Jan. 2007
 * 
 * Copyright (C) 2007, Parrot S.A. - www.parrot.biz
 ****************************************************************************
 */
#ifndef _CS42XX_HAL_H_
#define _CS42XX_HAL_H_

#include <cyg/hal/drv_api.h>
#include <cyg/io/gpio.h>


/** Structure for hardware configuration.
 *
 * Limitations on the device driver, when using multiple CS42XX devices
 * (implying constraints on hardware design):
 * - All devices shall be connected to the same reset pin.
 * - All devices shall share the same interrupt pin (they will trigger
 *   the same signal, e.g. falling edge).
 * - Functionally speaking, the way to trigger interrupt after having
 *   detected an event, shall be the same for all devices and all events
 *   (no support for variety of event triggering conditions), e.g.
 *   "interrupt is triggered only when any event begins to occur".
 *   [Cf. "Interrupts and Overflow" paragraph in the appropriate audio
 *    codec documentations].
 * - According to how the interrupt line is connected between devices,
 *   one may choose between active-low (advised) or active-high line.
 *
 * @typedef CS42XX_HALCFG
 */
typedef struct CS42XX_HALCFG_TAG CS42XX_HALCFG;
struct CS42XX_HALCFG_TAG
{
	struct
	{
		GPIO_PORT_PIN pin;    ///< Reset pin (GPIOxx).
	}
		reset;

	struct
	{
		// Common to both devices.
		cyg_bool        activehigh; ///< true if interrupt is active high.

		// ASIC side.
		cyg_vector_t    vector;     ///< Interrupt vector.
		GPIO_PORT_PIN   pin;        ///< Pin triggering the interrupt (GPIOxx).

		// Audio codec side.
		CS42XX_TRIGMODE trigmode; ///< Triggering mode when event occurs.
	}
		intr;
};

/** Prototype of function that can be called back after an interrupt
 *  occurred.
 */
typedef void (*CS42XX_HAL_PROCESSINTR_CB)(void);



/*
 * Exported functions.
 */


// Get the configuration of the whole set of devices.
extern Cyg_ErrNo
CS42XX_HAL_GetConfig(CS42XX_HALCFG *);

// Set the configuration for all devices.
extern Cyg_ErrNo
CS42XX_HAL_SetConfig(CS42XX_HALCFG *);

// Hardware-reset all devices connected to the reset pin.
extern Cyg_ErrNo
CS42XX_HAL_Reset(void);


#ifdef CYGPKG_DEVICES_CS42XX_USEINTR

// Install the interrupt handler according to configuration previously set.
extern Cyg_ErrNo
CS42XX_HAL_InstallIntr(CS42XX_HAL_PROCESSINTR_CB);

#endif // CYGPKG_DEVICES_CS42XX_USEINTR



#endif // _CS42XX_HAL_H_

