//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================

/**
 ****************************************************************************
 * Driver for Cirrus Logic CS42XX stereo audio codec.
 *
 * @file   cs42xx_regs.c
 * @brief  Access to CS42XX registers, with optional cache.
 * 
 * For the complete description of registers, please refer to:
 * http://www.cirrus.com/en/pubs/proDatasheet/CS4245_F1.pdf
 * http://www.cirrus.com/en/pubs/proDatasheet/CS4265_F1.pdf
 * http://www.cirrus.com/en/pubs/proDatasheet/CS5345_F1.pdf
 *
 * @author David Tylski
 * @date   Jan. 2007
 * 
 * Copyright (C) 2007, Parrot S.A. - www.parrot.biz
 ****************************************************************************
 */

#include <cyg/hal/drv_api.h>
#include <cyg/devs/cs42xx_regs.h>
#include <cyg/devs/cs42xx_supp.h>
#include <cyg/devs/cs42xx_i2c.h>


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                             REGISTER CACHE                              *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/**
 * @addtogroup cs42xx_regcache  Register cache
 * @{
 * 
 * This cache has two main advantages:
 * - when a "static" register has to be read multiple times, one has just to
 *   read once from device (init.), then continue reading in cache.
 * - when only updating flags of a configuration register, the value of other
 *   flags is available in the cache (avoids a useless READ operation).
 * 
 * However, it is NOT SUITABLE to access status and interrupt registers,
 * since their content is expected often to change. Prefer using direct
 * access to those registers.
 *
 * @warning  Register CS42XX@0x00 (CS42XX_REG_NULL) does not exist;
 *           register CS42XX@0x01 is cached in CS42XXRegcache[devid][0];
 *           etc.
 */
#ifdef CS42XX_USECACHE


/** Cached value of register (read/write access).
 *
 * @def   CS42XX_REG_CACHED
 * @param dev  Symbolic device Id
 * @param reg  Register given as CS42XX_REG_XXX
 */
#define CS42XX_REG_CACHED(dev,reg) \
		(CS42XXRegcache[dev][reg - CS42XX_REG_FIRST])

/** "Alteration" state of cached register: true if altered, false if
 *  synchronized (read/write access).
 *
 * @def   CS42XX_REG_ALTERED
 * @param dev  Symbolic device Id
 * @param reg  Register given as CS42XX_REG_XXX
 */
#define CS42XX_REG_ALTERED(dev,reg) \
		(CS42XXRegcacheAlt[dev][reg - CS42XX_REG_FIRST])


/** Cached values of device registers.
 * Register cache shall be as big as the maximal number of registers in
 * device.
 */
static CYG_BYTE CS42XXRegcache   [CS42XX_NBDEVICES][CS42XX_REG_NBMAX];

/** Flags to identify registers not yet committed. */
static cyg_bool CS42XXRegcacheAlt[CS42XX_NBDEVICES][CS42XX_REG_NBMAX];

/** Mutex to protect access to register cache. */
static cyg_drv_mutex_t CS42XXRegcacheMutex;



/** Initialize the register cache (no access to CS42XX devices).
 * Must be called before any other function here.
 *
 * @retval ENOERR  Success
 ****************************************************************************/
Cyg_ErrNo
CS42XX_RegCache_Init(void)
{
	CS42XX_DEVID devid;
	CS42XX_REG   reg;

	// Initialize mutex
	cyg_drv_mutex_init(&CS42XXRegcacheMutex);

	// Initialize cache for all possible devices
	for (devid = 0; devid < CS42XX_NBDEVICES; devid++)
	{
		for (reg = CS42XX_REG_FIRST; reg <= CS42XX_REG_LASTMAX; reg++)
		{
			// No special init. value, shall not be committed
			CS42XX_REG_CACHED (devid, reg) = 0x00;
			CS42XX_REG_ALTERED(devid, reg) = false;
		}
	}

	return ENOERR;
}


/** Load the register cache using data available on given device.
 * CS42XX_RegCache_Init() shall be called prior.
 *
 * @param devid  Symbolic device Id
 * 
 * @retval  ENOERR  Successful reading in device registers
 * @retval -ENODEV  No such device
 * @retval -EIO     Error during transmission
 ****************************************************************************/
Cyg_ErrNo
CS42XX_RegCache_Load(CS42XX_DEVID devid)
{
	Cyg_ErrNo  err;
	CS42XX_REG reg;

	if (devid >= CS42XX_NBDEVICES)
	{
		// No such device
		err = -ENODEV;
	}
	else
	{
		cyg_drv_mutex_lock(&CS42XXRegcacheMutex);

		// Read the first register (chip Id) so that we know the number of
		// registers in that device.
		err = CS42XX_I2C_Read(devid, CS42XX_REG_CHIPID,
					&CS42XX_REG_CACHED(devid, CS42XX_REG_CHIPID), 1);
		if (ENOERR == err)
		{
			// Find all properties of device according to this register
			// (in particular, the number of registers).
			err = CS42XX_Supp_Init(devid);
			if (ENOERR == err)
			{
				// Now, read all device memory (starting from CS42XX_REG_FIRST)
				// and store it into the register cache.
				// (For simplicity we read also the chip Id again.)
				err = CS42XX_I2C_Read(devid, CS42XX_REG_FIRST,
							&CS42XX_REG_CACHED(devid, CS42XX_REG_FIRST),
							CS42XX_REG_NBOF(devid));
				if (ENOERR == err)
				{
					// All cache is synchronized (i.e. not altered)
					for(reg = CS42XX_REG_FIRST;
						reg <= CS42XX_REG_LASTOF(devid);
						reg++)
					{
						CS42XX_REG_ALTERED(devid, reg) = false;
					}
				}
				// else reading of all registers failed.
			}
			// else device seems unknown (see details in err).
		}
		// else reading of chip Id register failed.

		cyg_drv_mutex_unlock(&CS42XXRegcacheMutex);
	}
	
	return err;
}


/** Read the cached value of register.
 * 
 * If perr points to a Cyg_ErrNo variable, it will receive the status of
 * operation:
 * - ( ENOERR)  Success
 * - (-EINVAL)  Invalid parameters
 * - (-ENODEV)  No such device
 *
 * @warning  If value changed on device register (like for interrupt
 *           registers), the value is not refreshed. So, reserve the use of
 *           this function to registers which are not modified by the device
 *           itself, e.g. configuration registers.
 *
 * @param devid   Symbolic device Id
 * @param addr    Address of register to read (from cache)
 * @param perr    Pointer to Cyg_ErrNo variable (NULL if not required)
 * 
 * @return   Value read in cache
 ****************************************************************************/
CYG_BYTE
CS42XX_RegCache_Peek(CS42XX_DEVID devid, CS42XX_REG reg, Cyg_ErrNo *perr)
{
	CYG_BYTE  val = 0x00;
	Cyg_ErrNo err = -ENODEV;

	if (!CS42XX_REG_ISDEVVALID(devid))
	{
		// No such device
		err = -ENODEV;
	}
	else if (!CS42XX_REG_ISVALID(devid, reg))
	{
		// Invalid register number
		err = -EINVAL;
	}
	else
	{
		// Get the cached content of register
		val = CS42XX_REG_CACHED(devid, reg);
		err = ENOERR;
	}

	// Report the error status.
	if (perr != NULL)
	{
		*perr = err;
	}
	// else no return of error code.

	return val;
}


/** Set the value of register in cache (commit will be necessary).
 *
 * @param devid  Symbolic device Id
 * @param addr   Address of register to read
 * @param value  Value to be stored in register
 *
 * @retval  ENOERR  Success
 * @retval -EINVAL  Invalid parameters
 * @retval -ENODEV  No such device
 ****************************************************************************/
Cyg_ErrNo
CS42XX_RegCache_Poke(CS42XX_DEVID devid, CS42XX_REG reg, CYG_BYTE val)
{
	Cyg_ErrNo err;

	if (!CS42XX_REG_ISDEVVALID(devid))
	{
		// No such device
		err = -ENODEV;
	}
	else if (!CS42XX_REG_ISVALID(devid, reg))
	{
		// Invalid register
		err = -EINVAL;
	}
	else
	{
		// Record the new content of register, and set alteration flag.
		cyg_drv_mutex_lock(&CS42XXRegcacheMutex);
		CS42XX_REG_CACHED (devid, reg) = val;
		CS42XX_REG_ALTERED(devid, reg) = true;
		cyg_drv_mutex_unlock(&CS42XXRegcacheMutex);

		err = ENOERR;
	}
	
	return err;
}


/** Commit register cache in the given range of addresses.
 * 
 * This function SHALL stay local to this submodule, and be called only by
 * CS42XX_RegCache_CommitAll().
 *
 * @param devid  Symbolic device Id
 * @param reg1   First register to be committed
 * @param reg2   Last register to be committed (included)
 *
 * @retval  ENOERR  Success
 * @retval -EINVAL  Invalid parameters
 * @retval -ENODEV  No such device
 * @retval -EIO     Error during transmission
 ****************************************************************************/
static Cyg_ErrNo
CS42XX_RegCache_Commit(CS42XX_DEVID devid, CS42XX_REG reg1, CS42XX_REG reg2)
{
	Cyg_ErrNo  err;
	CS42XX_REG reg;

	// Warning: do not attempt to lock mutex, since it has been taken by
	// caller CS42XX_RegCache_CommitAll()!

	if (!CS42XX_REG_ISDEVVALID(devid))
	{
		// No such device
		err = -ENODEV;
	}
	else if (!CS42XX_REG_ISVALID(devid, reg1) ||
			 !CS42XX_REG_ISVALID(devid, reg2) ||
			 (reg1 > reg2))
	{
		// Invalid range of registers
		err = -EINVAL;
	}
	else
	{
		// Commit registers reg1 to reg2 (included)
		err = CS42XX_I2C_Write(devid, reg1,
					&CS42XX_REG_CACHED(devid, reg1),
					reg2 - reg1 + 1);

		// If successful operation, set "alteration flags" to false.
		if (ENOERR == err)
		{
			for (reg = reg1; reg <= reg2; reg++)
			{
				CS42XX_REG_ALTERED(devid, reg) = false;
			}
		}
		// else error occurred during writing to registers (so it is
		// not synchronized properly).
	}

	return err;
}


/** Commit changes on register cache to device.
 *
 * For performance purposes, only registers marked as "altered" are
 * committed using the automatic MAP-increment feature of audio codec:
 * here we look for the altered registers grouped in block (a block can be
 * only 1 register), and send this block to the device.
 *
 * @param devid  Symbolic device Id
 * 
 * @retval  ENOERR  Cache committed successfully
 * @retval -EINVAL  Invalid parameters
 * @retval -ENODEV  No such device
 * @retval -EIO     Error during transmission
 ****************************************************************************/
Cyg_ErrNo
CS42XX_RegCache_CommitAll(CS42XX_DEVID devid)
{
	Cyg_ErrNo  err;
	CS42XX_REG reg, firstalt = CS42XX_REG_NULL;

	if (!CS42XX_REG_ISDEVVALID(devid))
	{
		// Invalid device Id.
		err = -ENODEV;
	}
	else
	{
		cyg_drv_mutex_lock(&CS42XXRegcacheMutex);

		// Look for altered registers and send new values to the device
		for ((err  = ENOERR),   (reg =  CS42XX_REG_FIRST);
			 (ENOERR == err) && (reg <= CS42XX_REG_LASTOF(devid));
			 reg++)
		{
			if (CS42XX_REG_NULL == firstalt)
			{
				// Look for the first altered register
				if (CS42XX_REG_ALTERED(devid, reg))
				{
					firstalt = reg;
				}
				// else this register is not altered.
			}
			else
			{
				// Look for the last altered register and commit
				// the range of registers when last has been found
				if (!CS42XX_REG_ALTERED(devid, reg))
				{
					err = CS42XX_RegCache_Commit(devid, firstalt, reg-1);
					firstalt = CS42XX_REG_NULL;
				}
				// else this register is altered too.
			}
		}

		if ((ENOERR == err) && (CS42XX_REG_NULL != firstalt))
		{
			// Last register has been altered too
			err = CS42XX_RegCache_Commit(devid,
						firstalt, CS42XX_REG_LASTOF(devid));
		}
		// else an error occurred (failure) or all registers have been
		// committed (success).

		cyg_drv_mutex_unlock(&CS42XXRegcacheMutex);
	}

	return err;
}


#endif // CS42XX_USECACHE
// !cs42xx_regcache
/** @} */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                     DIRECT ACCESS TO CODEC REGISTERS                    *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


/** Peek the value stored in register.
 * When cache is enabled, this function shall be called to update cache at
 * that address (warning, value in cache is then lost).
 * 
 * If perr points to a Cyg_ErrNo variable, it will receive the status of
 * operation:
 * - ( ENOERR)  Success
 * - (-EINVAL)  Invalid parameters
 * - (-ENODEV)  No such device
 * - (-EIO)     Error during transmission
 *
 * @param devid  Symbolic device Id
 * @param reg    Address of register to read (CS42XX_REG_XXX)
 * @param perr   Pointer to Cyg_ErrNo variable (NULL if not required)
 * 
 * @return  Value read
 ****************************************************************************/
CYG_BYTE
CS42XX_Peek(CS42XX_DEVID devid, CS42XX_REG reg, Cyg_ErrNo *perr)
{
	CYG_BYTE  val = 0;
	Cyg_ErrNo err;

	if (!CS42XX_REG_ISDEVVALID(devid))
	{
		// No such device
		err = -ENODEV;
	}
	else if (!CS42XX_REG_ISVALID(devid, reg))
	{
		// Invalid register number
		err = -EINVAL;
	}
	else
	{
		// Read value from register	
		err = CS42XX_I2C_Read(devid, reg, &val, 1);

#ifdef CS42XX_USECACHE
		// Update register cache. The previous value is overwritten.
		if (ENOERR == err)
		{
			cyg_drv_mutex_lock(&CS42XXRegcacheMutex);
			CS42XX_REG_CACHED (devid, reg) = val;
			CS42XX_REG_ALTERED(devid, reg) = false;
			cyg_drv_mutex_unlock(&CS42XXRegcacheMutex);
		}
		// else an error occurred during reading of register.
#endif
	}

	// If requested, report err.
	if (NULL != perr)
	{
		*perr = err;
	}
	// else no report requested.

	return val;
}


/** Store immediately the given value into the device register.
 * When cache is enabled, the value stored in cache is overwritten.
 * 
 * @param devid  Symbolic device Id
 * @param reg    Address of register to write (CS42XX_REG_XXX)
 * @param val    Value to be stored in register
 *
 * @retval  ENOERR  Success
 * @retval -ENODEV  No such device
 * @retval -EINVAL  Invalid parameters
 * @retval -EIO     Error during transmission
 ****************************************************************************/
Cyg_ErrNo
CS42XX_Poke(CS42XX_DEVID devid, CS42XX_REG reg, CYG_BYTE val)
{
	Cyg_ErrNo err;
	
	if (!CS42XX_REG_ISDEVVALID(devid))
	{
		// No such device
		err = -ENODEV;
	}
	else if (!CS42XX_REG_ISVALID(devid, reg))
	{
		// Invalid register number
		err = -EINVAL;
	}
	else
	{
		// Write value to device register
		err = CS42XX_I2C_Write(devid, reg, &val, 1);

#ifdef CS42XX_USECACHE
		// Update register cache
		if (ENOERR == err)
		{
			cyg_drv_mutex_lock(&CS42XXRegcacheMutex);
			CS42XX_REG_CACHED (devid, reg) = val;
			CS42XX_REG_ALTERED(devid, reg) = false;
			cyg_drv_mutex_unlock(&CS42XXRegcacheMutex);
		}
		// else error during writing (not synchronized properly).
#endif
	}
	
	return err;
}

