//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================

/**
 ****************************************************************************
 * Driver for Cirrus Logic CS42XX stereo audio codec.
 *
 * @file   cs42xx_testaudio.c
 * @brief  Tests applicable on CS42XX driver - Audio part.
 *
 * This code is afforded so that audio can be manipulated on the device
 * by the driver. It can be used either to test the hardware around audio
 * codec, or to help the CS42XX driver validation process. It does not
 * directly test the driver.
 *
 * @author David Tylski
 * @date   Mar. 2007
 *
 * Copyright (C) 2007, Parrot S.A. - www.parrot.biz
 ****************************************************************************
 */
#include <cyg/io/aai.h>
#include <cyg/hal/hal_diag.h>
#include <cyg/devs/cs42xx_api.h>

#include <cyg/devs/test/cs42xx_test.h>
#include <cyg/devs/test/cs42xx_testhmi.h>
#include <cyg/devs/test/cs42xx_testaudio.h>

#include <math.h> //for sin

// Audio testing software is conditioned by CS42XX_TEST_SUPPORT_AUDIO.
#ifdef CS42XX_TEST_SUPPORT_AUDIO



/*
 * Defines.
 ***********/

#define CS42XX_TESTAUDIO_THRPRIORITY		4
#define CS42XX_TESTAUDIO_STACKSIZE			4096

#define CS42XX_OUTPUT_PATH_NAME_RAM			"/out_test.raw"
#define CS42XX_SENT_RAW_FILE_HEADER			"\r\nTest audio start - RAW FILE"
#define CS42XX_SENT_RAW_FILE_FOOTER			"RAW FILE - Test audio stop\r\n"
#define CS42XX_MAX_ACQ_LEN					10485760

#define CS42XX_TESTAUDIO_SAMPLE_FREQUENCY_8K	8000
#define CS42XX_TESTAUDIO_SAMPLE_FREQUENCY_44K	44100
#define CS42XX_TESTAUDIO_MAX_SAMPLE_FREQUENCY	CS42XX_TESTAUDIO_SAMPLE_FREQUENCY_44K

#define CS42XX_TESTAUDIO_DEFAULT_AMPLITUDE		0x7FFF	///< Max amplitude (maximum for 16 bits samples = 0x7FFF)
#define CS42XX_TESTAUDIO_DEFAULT_FUNC_FREQ		1000	///< Frequency in Hertz.

#define CS42XX_TESTAUDIO_WRITE_BUFFER_LEN		128		///< Size in words (16 bits).

typedef enum CS42XX_TESTAUDIO_FUNCTION
{
	CS42XX_TESTAUDIO_NOFUNCT,
	CS42XX_TESTAUDIO_SINUS,
} CS42XX_TESTAUDIO_FUNCTION;

 /*
 * Types.
 *********/

/** Structure for the storage of data related to audio testing.
 * @typedef CS42XX_TESTAUDIO_DATA
 */
typedef struct CS42XX_TESTAUDIO_DATA_TAG CS42XX_TESTAUDIO_DATA;
struct CS42XX_TESTAUDIO_DATA_TAG
{
	CS42XX_TEST_ID testcurr;  ///< Id of test currently running.
	CS42XX_TEST_ID testreq;   ///< Id of future test.
	cyg_mutex_t    testmutex; ///< Mutex for testcurr and testreq.

	cyg_thread     threadobj;                         ///< Thread object.
	cyg_handle_t   thread;                            ///< Handle of thread.
	char           stack[CS42XX_TESTAUDIO_STACKSIZE]; ///< Stack.

	cyg_mbox       mailboxobj; ///< Mailbox object.
	cyg_handle_t   mailbox;    ///< Handle of mailbox.

	CS42XX_TESTAUDIO_FUNCTION	functcurr;		///< Id of function type currently running.
	CS42XX_TESTAUDIO_FUNCTION	functreq;		///< Id of future function type.
	cyg_int32							functfrequency;   ///< Frequency of function currently running.
};



/*
 * Constants.
 *************/

/** Silence. */
static const cyg_int16 CS42XXTestAudioSilence[2][CS42XX_TESTAUDIO_WRITE_BUFFER_LEN] __attribute__ ((aligned (4))) =
{
	{
		/* A, B */
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	},
	{
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	}
};

/** Highest squared frequency, stereo.
 * With s = 44100 Hz:
 * f = 22050 Hz on both channels, phased.
 */
static const cyg_int16 CS42XXTestAudioHF1[2][CS42XX_TESTAUDIO_WRITE_BUFFER_LEN] __attribute__ ((aligned (4))) =
{
	{
		/* A, B */
		0x7FFF, 0x7FFF,  0x8000, 0x8000,  0x7FFF, 0x7FFF,  0x8000, 0x8000,
		0x7FFF, 0x7FFF,  0x8000, 0x8000,  0x7FFF, 0x7FFF,  0x8000, 0x8000,
		0x7FFF, 0x7FFF,  0x8000, 0x8000,  0x7FFF, 0x7FFF,  0x8000, 0x8000,
		0x7FFF, 0x7FFF,  0x8000, 0x8000,  0x7FFF, 0x7FFF,  0x8000, 0x8000,
		0x7FFF, 0x7FFF,  0x8000, 0x8000,  0x7FFF, 0x7FFF,  0x8000, 0x8000,
		0x7FFF, 0x7FFF,  0x8000, 0x8000,  0x7FFF, 0x7FFF,  0x8000, 0x8000,
		0x7FFF, 0x7FFF,  0x8000, 0x8000,  0x7FFF, 0x7FFF,  0x8000, 0x8000,
		0x7FFF, 0x7FFF,  0x8000, 0x8000,  0x7FFF, 0x7FFF,  0x8000, 0x8000,
		0x7FFF, 0x7FFF,  0x8000, 0x8000,  0x7FFF, 0x7FFF,  0x8000, 0x8000,
		0x7FFF, 0x7FFF,  0x8000, 0x8000,  0x7FFF, 0x7FFF,  0x8000, 0x8000,
		0x7FFF, 0x7FFF,  0x8000, 0x8000,  0x7FFF, 0x7FFF,  0x8000, 0x8000,
		0x7FFF, 0x7FFF,  0x8000, 0x8000,  0x7FFF, 0x7FFF,  0x8000, 0x8000,
		0x7FFF, 0x7FFF,  0x8000, 0x8000,  0x7FFF, 0x7FFF,  0x8000, 0x8000,
		0x7FFF, 0x7FFF,  0x8000, 0x8000,  0x7FFF, 0x7FFF,  0x8000, 0x8000,
		0x7FFF, 0x7FFF,  0x8000, 0x8000,  0x7FFF, 0x7FFF,  0x8000, 0x8000,
		0x7FFF, 0x7FFF,  0x8000, 0x8000,  0x7FFF, 0x7FFF,  0x8000, 0x8000,
	},
	{
		0x7FFF, 0x7FFF,  0x8000, 0x8000,  0x7FFF, 0x7FFF,  0x8000, 0x8000,
		0x7FFF, 0x7FFF,  0x8000, 0x8000,  0x7FFF, 0x7FFF,  0x8000, 0x8000,
		0x7FFF, 0x7FFF,  0x8000, 0x8000,  0x7FFF, 0x7FFF,  0x8000, 0x8000,
		0x7FFF, 0x7FFF,  0x8000, 0x8000,  0x7FFF, 0x7FFF,  0x8000, 0x8000,
		0x7FFF, 0x7FFF,  0x8000, 0x8000,  0x7FFF, 0x7FFF,  0x8000, 0x8000,
		0x7FFF, 0x7FFF,  0x8000, 0x8000,  0x7FFF, 0x7FFF,  0x8000, 0x8000,
		0x7FFF, 0x7FFF,  0x8000, 0x8000,  0x7FFF, 0x7FFF,  0x8000, 0x8000,
		0x7FFF, 0x7FFF,  0x8000, 0x8000,  0x7FFF, 0x7FFF,  0x8000, 0x8000,
		0x7FFF, 0x7FFF,  0x8000, 0x8000,  0x7FFF, 0x7FFF,  0x8000, 0x8000,
		0x7FFF, 0x7FFF,  0x8000, 0x8000,  0x7FFF, 0x7FFF,  0x8000, 0x8000,
		0x7FFF, 0x7FFF,  0x8000, 0x8000,  0x7FFF, 0x7FFF,  0x8000, 0x8000,
		0x7FFF, 0x7FFF,  0x8000, 0x8000,  0x7FFF, 0x7FFF,  0x8000, 0x8000,
		0x7FFF, 0x7FFF,  0x8000, 0x8000,  0x7FFF, 0x7FFF,  0x8000, 0x8000,
		0x7FFF, 0x7FFF,  0x8000, 0x8000,  0x7FFF, 0x7FFF,  0x8000, 0x8000,
		0x7FFF, 0x7FFF,  0x8000, 0x8000,  0x7FFF, 0x7FFF,  0x8000, 0x8000,
		0x7FFF, 0x7FFF,  0x8000, 0x8000,  0x7FFF, 0x7FFF,  0x8000, 0x8000,
	}
};

/** Highest squared frequency, opposite stereo channels.
 * With s = 44100 Hz:
 * f = 22050 Hz on both channels, 180�-phased.
 */
static const cyg_int16 CS42XXTestAudioHF1Opp[2][CS42XX_TESTAUDIO_WRITE_BUFFER_LEN] __attribute__ ((aligned (4))) =
{
	{
		/* A, B */
		0x7FFF, 0x8000,  0x8000, 0x7FFF,  0x7FFF, 0x8000,  0x8000, 0x7FFF,
		0x7FFF, 0x8000,  0x8000, 0x7FFF,  0x7FFF, 0x8000,  0x8000, 0x7FFF,
		0x7FFF, 0x8000,  0x8000, 0x7FFF,  0x7FFF, 0x8000,  0x8000, 0x7FFF,
		0x7FFF, 0x8000,  0x8000, 0x7FFF,  0x7FFF, 0x8000,  0x8000, 0x7FFF,
		0x7FFF, 0x8000,  0x8000, 0x7FFF,  0x7FFF, 0x8000,  0x8000, 0x7FFF,
		0x7FFF, 0x8000,  0x8000, 0x7FFF,  0x7FFF, 0x8000,  0x8000, 0x7FFF,
		0x7FFF, 0x8000,  0x8000, 0x7FFF,  0x7FFF, 0x8000,  0x8000, 0x7FFF,
		0x7FFF, 0x8000,  0x8000, 0x7FFF,  0x7FFF, 0x8000,  0x8000, 0x7FFF,
		0x7FFF, 0x8000,  0x8000, 0x7FFF,  0x7FFF, 0x8000,  0x8000, 0x7FFF,
		0x7FFF, 0x8000,  0x8000, 0x7FFF,  0x7FFF, 0x8000,  0x8000, 0x7FFF,
		0x7FFF, 0x8000,  0x8000, 0x7FFF,  0x7FFF, 0x8000,  0x8000, 0x7FFF,
		0x7FFF, 0x8000,  0x8000, 0x7FFF,  0x7FFF, 0x8000,  0x8000, 0x7FFF,
		0x7FFF, 0x8000,  0x8000, 0x7FFF,  0x7FFF, 0x8000,  0x8000, 0x7FFF,
		0x7FFF, 0x8000,  0x8000, 0x7FFF,  0x7FFF, 0x8000,  0x8000, 0x7FFF,
		0x7FFF, 0x8000,  0x8000, 0x7FFF,  0x7FFF, 0x8000,  0x8000, 0x7FFF,
		0x7FFF, 0x8000,  0x8000, 0x7FFF,  0x7FFF, 0x8000,  0x8000, 0x7FFF,
	},
	{
		0x7FFF, 0x8000,  0x8000, 0x7FFF,  0x7FFF, 0x8000,  0x8000, 0x7FFF,
		0x7FFF, 0x8000,  0x8000, 0x7FFF,  0x7FFF, 0x8000,  0x8000, 0x7FFF,
		0x7FFF, 0x8000,  0x8000, 0x7FFF,  0x7FFF, 0x8000,  0x8000, 0x7FFF,
		0x7FFF, 0x8000,  0x8000, 0x7FFF,  0x7FFF, 0x8000,  0x8000, 0x7FFF,
		0x7FFF, 0x8000,  0x8000, 0x7FFF,  0x7FFF, 0x8000,  0x8000, 0x7FFF,
		0x7FFF, 0x8000,  0x8000, 0x7FFF,  0x7FFF, 0x8000,  0x8000, 0x7FFF,
		0x7FFF, 0x8000,  0x8000, 0x7FFF,  0x7FFF, 0x8000,  0x8000, 0x7FFF,
		0x7FFF, 0x8000,  0x8000, 0x7FFF,  0x7FFF, 0x8000,  0x8000, 0x7FFF,
		0x7FFF, 0x8000,  0x8000, 0x7FFF,  0x7FFF, 0x8000,  0x8000, 0x7FFF,
		0x7FFF, 0x8000,  0x8000, 0x7FFF,  0x7FFF, 0x8000,  0x8000, 0x7FFF,
		0x7FFF, 0x8000,  0x8000, 0x7FFF,  0x7FFF, 0x8000,  0x8000, 0x7FFF,
		0x7FFF, 0x8000,  0x8000, 0x7FFF,  0x7FFF, 0x8000,  0x8000, 0x7FFF,
		0x7FFF, 0x8000,  0x8000, 0x7FFF,  0x7FFF, 0x8000,  0x8000, 0x7FFF,
		0x7FFF, 0x8000,  0x8000, 0x7FFF,  0x7FFF, 0x8000,  0x8000, 0x7FFF,
		0x7FFF, 0x8000,  0x8000, 0x7FFF,  0x7FFF, 0x8000,  0x8000, 0x7FFF,
		0x7FFF, 0x8000,  0x8000, 0x7FFF,  0x7FFF, 0x8000,  0x8000, 0x7FFF,
	}

};

/** Low squared frequency, stereo.
 * With s = 44100 Hz:
 * f  =  344.5 Hz on both channels, phased.
 * f2 = 1033.6 Hz = 3*f
 * f3 = 1722.7 Hz = 5*f
 * f4 = 2411.7 Hz = 7*f
 * ...
 */
static const cyg_int16 CS42XXTestAudioLF1[2][CS42XX_TESTAUDIO_WRITE_BUFFER_LEN] __attribute__ ((aligned (4))) =
{
	{
		/* A, B */
		0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,
		0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,
		0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,
		0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,
		0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,
		0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,
		0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,
		0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,
		0x8000, 0x8000,  0x8000, 0x8000,  0x8000, 0x8000,  0x8000, 0x8000,
		0x8000, 0x8000,  0x8000, 0x8000,  0x8000, 0x8000,  0x8000, 0x8000,
		0x8000, 0x8000,  0x8000, 0x8000,  0x8000, 0x8000,  0x8000, 0x8000,
		0x8000, 0x8000,  0x8000, 0x8000,  0x8000, 0x8000,  0x8000, 0x8000,
		0x8000, 0x8000,  0x8000, 0x8000,  0x8000, 0x8000,  0x8000, 0x8000,
		0x8000, 0x8000,  0x8000, 0x8000,  0x8000, 0x8000,  0x8000, 0x8000,
		0x8000, 0x8000,  0x8000, 0x8000,  0x8000, 0x8000,  0x8000, 0x8000,
		0x8000, 0x8000,  0x8000, 0x8000,  0x8000, 0x8000,  0x8000, 0x8000,
	},
	{
		0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,
		0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,
		0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,
		0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,
		0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,
		0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,
		0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,
		0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,  0x7FFF, 0x7FFF,
		0x8000, 0x8000,  0x8000, 0x8000,  0x8000, 0x8000,  0x8000, 0x8000,
		0x8000, 0x8000,  0x8000, 0x8000,  0x8000, 0x8000,  0x8000, 0x8000,
		0x8000, 0x8000,  0x8000, 0x8000,  0x8000, 0x8000,  0x8000, 0x8000,
		0x8000, 0x8000,  0x8000, 0x8000,  0x8000, 0x8000,  0x8000, 0x8000,
		0x8000, 0x8000,  0x8000, 0x8000,  0x8000, 0x8000,  0x8000, 0x8000,
		0x8000, 0x8000,  0x8000, 0x8000,  0x8000, 0x8000,  0x8000, 0x8000,
		0x8000, 0x8000,  0x8000, 0x8000,  0x8000, 0x8000,  0x8000, 0x8000,
		0x8000, 0x8000,  0x8000, 0x8000,  0x8000, 0x8000,  0x8000, 0x8000,
	}
};

/** Constant sinusoid.
 * With s = 44100 Hz:
 * f = 344.5 Hz on both channels, phased.
 */
static const cyg_int16 CS42XXTestAudioSinus[2][CS42XX_TESTAUDIO_WRITE_BUFFER_LEN] __attribute__ ((aligned (4))) =
{
	{
		/* A, B */
		 0x0000,  0x0000,   0x0C8B,  0x0C8B,   0x18F5,  0x18F5,   0x251E,  0x251E,
		 0x3105,  0x3105,   0x3C49,  0x3C49,   0x472A,  0x472A,   0x5126,  0x5126,
		 0x5A7E,  0x5A7E,   0x62F0,  0x62F0,   0x6A5D,  0x6A5D,   0x70E4,  0x70E4,
		 0x7644,  0x7644,   0x7A7E,  0x7A7E,   0x7D90,  0x7D90,   0x7F5B,  0x7F5B,
		 0x7FFF,  0x7FFF,   0x7F5B,  0x7F5B,   0x7D90,  0x7D90,   0x7A7E,  0x7A7E,
		 0x7644,  0x7644,   0x70E4,  0x70E4,   0x6A5D,  0x6A5D,   0x62F0,  0x62F0,
		 0x5A7E,  0x5A7E,   0x5126,  0x5126,   0x472A,  0x472A,   0x3C49,  0x3C49,
		 0x3105,  0x3105,   0x251E,  0x251E,   0x18F5,  0x18F5,   0x0C8B,  0x0C8B,
		 0x0000,  0x0000,  -0x0C8B, -0x0C8B,  -0x18F5, -0x18F5,  -0x251E, -0x251E,
		-0x3105, -0x3105,  -0x3C49, -0x3C49,  -0x472A, -0x472A,  -0x5126, -0x5126,
		-0x5A7E, -0x5A7E,  -0x62F0, -0x62F0,  -0x6A5D, -0x6A5D,  -0x70E4, -0x70E4,
		-0x7644, -0x7644,  -0x7A7E, -0x7A7E,  -0x7D90, -0x7D90,  -0x7F5B, -0x7F5B,
		-0x7FFF, -0x7FFF,  -0x7F5B, -0x7F5B,  -0x7D90, -0x7D90,  -0x7A7E, -0x7A7E,
		-0x7644, -0x7644,  -0x70E4, -0x70E4,  -0x6A5D, -0x6A5D,  -0x62F0, -0x62F0,
		-0x5A7E, -0x5A7E,  -0x5126, -0x5126,  -0x472A, -0x472A,  -0x3C49, -0x3C49,
		-0x3105, -0x3105,  -0x251E, -0x251E,  -0x18F5, -0x18F5,  -0x0C8B, -0x0C8B,
	},
	{
		 0x0000,  0x0000,   0x0C8B,  0x0C8B,   0x18F5,  0x18F5,   0x251E,  0x251E,
		 0x3105,  0x3105,   0x3C49,  0x3C49,   0x472A,  0x472A,   0x5126,  0x5126,
		 0x5A7E,  0x5A7E,   0x62F0,  0x62F0,   0x6A5D,  0x6A5D,   0x70E4,  0x70E4,
		 0x7644,  0x7644,   0x7A7E,  0x7A7E,   0x7D90,  0x7D90,   0x7F5B,  0x7F5B,
		 0x7FFF,  0x7FFF,   0x7F5B,  0x7F5B,   0x7D90,  0x7D90,   0x7A7E,  0x7A7E,
		 0x7644,  0x7644,   0x70E4,  0x70E4,   0x6A5D,  0x6A5D,   0x62F0,  0x62F0,
		 0x5A7E,  0x5A7E,   0x5126,  0x5126,   0x472A,  0x472A,   0x3C49,  0x3C49,
		 0x3105,  0x3105,   0x251E,  0x251E,   0x18F5,  0x18F5,   0x0C8B,  0x0C8B,
		 0x0000,  0x0000,  -0x0C8B, -0x0C8B,  -0x18F5, -0x18F5,  -0x251E, -0x251E,
		-0x3105, -0x3105,  -0x3C49, -0x3C49,  -0x472A, -0x472A,  -0x5126, -0x5126,
		-0x5A7E, -0x5A7E,  -0x62F0, -0x62F0,  -0x6A5D, -0x6A5D,  -0x70E4, -0x70E4,
		-0x7644, -0x7644,  -0x7A7E, -0x7A7E,  -0x7D90, -0x7D90,  -0x7F5B, -0x7F5B,
		-0x7FFF, -0x7FFF,  -0x7F5B, -0x7F5B,  -0x7D90, -0x7D90,  -0x7A7E, -0x7A7E,
		-0x7644, -0x7644,  -0x70E4, -0x70E4,  -0x6A5D, -0x6A5D,  -0x62F0, -0x62F0,
		-0x5A7E, -0x5A7E,  -0x5126, -0x5126,  -0x472A, -0x472A,  -0x3C49, -0x3C49,
		-0x3105, -0x3105,  -0x251E, -0x251E,  -0x18F5, -0x18F5,  -0x0C8B, -0x0C8B,
	}
};



/*
 * Variables.
 *************/

/** Data related to audio testing. */
static CS42XX_TESTAUDIO_DATA CS42XXTestAudioData;

/** Buffer for audio tests. */
static cyg_int16 CS42XXTestAudioBuffer[2][CS42XX_TESTAUDIO_WRITE_BUFFER_LEN]  __attribute__ ((aligned(4)));
static cyg_int16 CS42XXTESTSampleBufferXHz[CS42XX_TESTAUDIO_MAX_SAMPLE_FREQUENCY];

// Handles for devices
static cyg_io_handle_t CS42XXTestAudioAAIOut;
static cyg_io_handle_t CS42XXTestAudioAAIIn;



/*
 * Declaration of local functions.
 **********************************/

static cyg_bool CS42XX_TestAudio_TestMailbox(void);



/*
 * Definition of functions.
 ***************************/



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                               AUDIO TESTS                               *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

static Cyg_ErrNo
CS42XX_TestAudio_LoadFunction(CS42XX_TESTAUDIO_FUNCTION funct, cyg_int32 samplefreq)
{
	Cyg_ErrNo err = ENOERR;
	cyg_int32 frequency = CS42XXTestAudioData.functfrequency;
	int amplitude = CS42XX_TESTAUDIO_DEFAULT_AMPLITUDE;
	cyg_int32 i;

	if (CS42XXTestAudioData.functreq != CS42XXTestAudioData.functcurr)
	{
		switch(funct)
		{
		case CS42XX_TESTAUDIO_SINUS:
			// Sinus
			diag_printf("\r\nLoading sample buffer... please wait");
			for(i=0;i<(samplefreq/frequency);i++)
				CS42XXTESTSampleBufferXHz[i]=(cyg_int16)(amplitude * sin(2*M_PI*frequency*i/samplefreq));
		break;
		default:
		break;
		}

	cyg_mutex_lock(&CS42XXTestAudioData.testmutex);
	CS42XXTestAudioData.functcurr = CS42XXTestAudioData.functreq;
	cyg_mutex_unlock(&CS42XXTestAudioData.testmutex);
	}

	return err;
}



/** Play silence on AAI output, until a new audio command is posted.
 ****************************************************************************/
static Cyg_test_code
CS42XX_TestAudio_PlaySilence(void)
{
	Cyg_ErrNo err = ENOERR;
	cyg_uint32 len = CS42XX_TESTAUDIO_WRITE_BUFFER_LEN * sizeof(cyg_int16);
	cyg_uint8  k = 0;

	// Loop until a new command is posted
	while(!CS42XX_TestAudio_TestMailbox())
	{
		err = cyg_io_write(CS42XXTestAudioAAIOut,
				(void *)CS42XXTestAudioSilence[k], &len);
		if (ENOERR != err)
		{
			diag_printf(">");
		}
		// else no write error encountered.

		k = (k+1)%2;
	}

	// End of test.
	return CYGNUM_TEST_PASS;
}


/** Play a high-frequency squared signal on AAI output, until a new audio
 *  command is posted.
 ****************************************************************************/
static Cyg_test_code
CS42XX_TestAudio_PlaySquaredHF1(void)
{
	Cyg_ErrNo err = ENOERR;
	cyg_uint32 len = CS42XX_TESTAUDIO_WRITE_BUFFER_LEN * sizeof(cyg_int16);
	cyg_uint8  k = 0;

	// Loop until a new command is posted
	while(!CS42XX_TestAudio_TestMailbox())
	{
		err = cyg_io_write(CS42XXTestAudioAAIOut,
				(void *)CS42XXTestAudioHF1[k], &len);
		if (ENOERR != err)
		{
			diag_printf(">");
		}
		// else no write error encountered.

		k = (k+1)%2;
	}

	// End of test.
	return CYGNUM_TEST_PASS;
}


/** Play a high-frequency squared signal with opposite channels on AAI
 *  output, until a new audio command is posted.
 ****************************************************************************/
static Cyg_test_code
CS42XX_TestAudio_PlaySquaredHF1Opp(void)
{
	Cyg_ErrNo err = ENOERR;
	cyg_uint32 len = CS42XX_TESTAUDIO_WRITE_BUFFER_LEN * sizeof(cyg_int16);
	cyg_uint8  k = 0;

	// Loop until a new command is posted
	while(!CS42XX_TestAudio_TestMailbox())
	{
		err = cyg_io_write(CS42XXTestAudioAAIOut,
				(void *)CS42XXTestAudioHF1Opp[k], &len);
		if (ENOERR != err)
		{
			diag_printf(">");
		}
		// else no write error encountered.

		k = (k+1)%2;
	}

	// End of test.
	return CYGNUM_TEST_PASS;
}


/** Play a low-frequency squared signal on AAI output, until a new audio
 *  command is posted.
 ****************************************************************************/
static Cyg_test_code
CS42XX_TestAudio_PlaySquaredLF1(void)
{
	Cyg_ErrNo err = ENOERR;
	cyg_uint32 len = CS42XX_TESTAUDIO_WRITE_BUFFER_LEN * sizeof(cyg_int16);
	cyg_uint8  k = 0;

	// Loop until a new command is posted
	while(!CS42XX_TestAudio_TestMailbox())
	{
		err = cyg_io_write(CS42XXTestAudioAAIOut,
				(void *)CS42XXTestAudioLF1[k], &len);
		if (ENOERR != err)
		{
			diag_printf(">");
		}
		// else no write error encountered.

		k = (k+1)%2;
	}

	// End of test.
	return CYGNUM_TEST_PASS;
}


/** Play a sinusoid on AAI output, until a new audio command is posted.
 ****************************************************************************/
static Cyg_test_code
CS42XX_TestAudio_PlaySinus(void)
{
	Cyg_ErrNo err = ENOERR;
	cyg_uint32 len = CS42XX_TESTAUDIO_WRITE_BUFFER_LEN * sizeof(cyg_int16);
	int i, j = 0;
	cyg_int32 samplefreq = CS42XX_TESTAUDIO_SAMPLE_FREQUENCY_44K;
	cyg_uint32 nbsamples;

	cyg_mutex_lock(&CS42XXTestAudioData.testmutex);
	CS42XXTestAudioData.functreq = CS42XX_TESTAUDIO_SINUS;
	CS42XXTestAudioData.functfrequency = CS42XX_TESTAUDIO_DEFAULT_FUNC_FREQ;
	cyg_mutex_unlock(&CS42XXTestAudioData.testmutex);

	CS42XX_TestAudio_LoadFunction(CS42XX_TESTAUDIO_SINUS, samplefreq);

	nbsamples = samplefreq/CS42XXTestAudioData.functfrequency;

	// Loop until a new command is posted
	while(!CS42XX_TestAudio_TestMailbox())
	{
		for(i=0; i<(CS42XX_TESTAUDIO_WRITE_BUFFER_LEN/2); i++)
		{
			//same value for left and right channels
			CS42XXTestAudioBuffer[0][2*i] = CS42XXTESTSampleBufferXHz[j];
			CS42XXTestAudioBuffer[0][2*i+1] = CS42XXTESTSampleBufferXHz[j];
			j = (j+1) % nbsamples;
		}

		err = cyg_io_write(CS42XXTestAudioAAIOut,
				(void *)CS42XXTestAudioBuffer[0], &len);
		if (ENOERR != err)
		{
			diag_printf(">");
		}
		// else no write error encountered.
	}

	// End of test.
	return CYGNUM_TEST_PASS;
}


/** Acquire audio and do some statistics on input data.
 ****************************************************************************/
static Cyg_test_code
CS42XX_TestAudio_Acquire(cyg_bool restitute)
{
	Cyg_ErrNo  err = ENOERR;
	unsigned int  size = 0;
	cyg_uint32 k = 0;
	cyg_uint32 len = CS42XX_TESTAUDIO_WRITE_BUFFER_LEN*sizeof(cyg_int16);
	cyg_uint32 lentotal = 0;
	FILE* outfile = NULL;

	outfile = fopen(CS42XX_OUTPUT_PATH_NAME_RAM, "wb");
	if (!outfile)
	{
		diag_printf("error while opening file %s (%d)", CS42XX_OUTPUT_PATH_NAME_RAM, errno);
	}
	else
	{
		len = CS42XX_TESTAUDIO_WRITE_BUFFER_LEN*sizeof(cyg_int16);
		cyg_io_read(CS42XXTestAudioAAIIn,
				(void *)CS42XXTestAudioBuffer[1], &len);
		if (ENOERR != err)
		{
			diag_printf("<");
		}
		// else no write error encountered.

		// Loop until a new command is posted
		while (!CS42XX_TestAudio_TestMailbox() && lentotal<CS42XX_MAX_ACQ_LEN)
		{
			err = cyg_io_read(CS42XXTestAudioAAIIn,
					(void *)CS42XXTestAudioBuffer[k], &len);
			if (ENOERR != err)
			{
				diag_printf("<");
			}
			// else no read error encountered.

			k = (k+1)%2; // swap buffers

			len = CS42XX_TESTAUDIO_WRITE_BUFFER_LEN*sizeof(cyg_int16);
			size = fwrite((void *)CS42XXTestAudioBuffer[k], 1, len, outfile);
			if (size == 0)
			{
				diag_printf("-");
			}
			// else no write error encountered.

			lentotal += len;

			if (restitute)
			{
				err = cyg_io_write(CS42XXTestAudioAAIOut,
						(void *)CS42XXTestAudioBuffer[k], &len);
				if (ENOERR != err)
				{
					diag_printf(">");
				}
				// else no write error encountered.
			}
		}
		fclose(outfile);
		diag_printf("\r\nRaw file size (bytes): %d\r\n", lentotal);
	}

	// End of test.
	return CYGNUM_TEST_PASS;
}


/** Play on AAI output the signal present on AAI input, until a new audio
 *  command is posted.
 *
 * Errors cannot be detected here, so result is always CYGNUM_TEST_PASS.
 ****************************************************************************/
static Cyg_test_code
CS42XX_TestAudio_SendRawFile(void)
{
	cyg_uint32 j;
	cyg_uint32 lentotal = 0;
	FILE* outfile = NULL;
	cyg_uint8 buffer[256];
	unsigned int size;

	if(!CS42XX_TestAudio_TestMailbox())
	{
		outfile = fopen(CS42XX_OUTPUT_PATH_NAME_RAM, "rb");
		if (!outfile)
		{
			diag_printf("error while opening file %s (%d)", CS42XX_OUTPUT_PATH_NAME_RAM, errno);
		}
		else
		{
			memset(buffer,0,256);
			diag_printf(CS42XX_SENT_RAW_FILE_HEADER);
			while ((size = (fread(buffer, 1, 256, outfile))) != 0)
			{
				for (j=0; j<256; j++)
				{
					HAL_DIAG_WRITE_CHAR(buffer[j]);
				}
				lentotal += 256;
				memset(buffer,0,256);
			}
			diag_printf(CS42XX_SENT_RAW_FILE_FOOTER);
			diag_printf("\r\nRaw file size (bytes): %d\r\n", lentotal);
		}
	}

	// End of test.
	return CYGNUM_TEST_PASS;
}

/** Play on AAI output the signal present on AAI input, until a new audio
 *  command is posted.
 *
 * Errors cannot be detected here, so result is always CYGNUM_TEST_PASS.
 ****************************************************************************/
static Cyg_test_code
CS42XX_TestAudio_LoopbackInOut(void)
{
	Cyg_ErrNo  err = ENOERR;
	cyg_uint32 k = 0;
	cyg_uint32 len = 0;

	len = CS42XX_TESTAUDIO_WRITE_BUFFER_LEN*sizeof(cyg_int16);
	cyg_io_read(CS42XXTestAudioAAIIn,
			(void *)CS42XXTestAudioBuffer[1], &len);
	if (ENOERR != err)
	{
		diag_printf("<");
	}
	// else no write error encountered.

	// Loop until a new command is posted
	while (!CS42XX_TestAudio_TestMailbox())
	{
		err = cyg_io_read(CS42XXTestAudioAAIIn,
				(void *)CS42XXTestAudioBuffer[k], &len);
		if (ENOERR != err)
		{
			diag_printf("<");
		}
		// else no read error encountered.

		len = CS42XX_TESTAUDIO_WRITE_BUFFER_LEN*sizeof(cyg_int16);
		err = cyg_io_write(CS42XXTestAudioAAIOut,
				(void *)CS42XXTestAudioBuffer[k], &len);
		if (ENOERR != err)
		{
			diag_printf(">");
		}

		k = (k+1)%2; // swap buffers
		// else no write error encountered.
	}

	// End of test.
	return CYGNUM_TEST_PASS;
}



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                        MANAGEMENT OF AUDIO TESTS                        *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


/** Thread managing audio for tests of CS42XX audio codec.
 *
 * According to its variable configuration, this thread starts and stops
 * audio tests.
 *
 * @param data  (Not used)
 ****************************************************************************/
static void
CS42XX_TestAudio_Thread(CYG_ADDRWORD data)
{
	Cyg_test_code testresult;

	while(1)
	{
		switch(CS42XXTestAudioData.testcurr)
		{

		case CS42XX_TEST_AUDIO_PLAYSILENCE:
			testresult = CS42XX_TestAudio_PlaySilence();
			break;

		case CS42XX_TEST_AUDIO_PLAYSQ_HF1:
			testresult = CS42XX_TestAudio_PlaySquaredHF1();
			break;

		case CS42XX_TEST_AUDIO_PLAYSQ_HF1OPP:
			testresult = CS42XX_TestAudio_PlaySquaredHF1Opp();
			break;

		case CS42XX_TEST_AUDIO_PLAYSQ_LF1:
			testresult = CS42XX_TestAudio_PlaySquaredLF1();
			break;

		case CS42XX_TEST_AUDIO_PLAYSINUS:
			testresult = CS42XX_TestAudio_PlaySinus();
			break;

		case CS42XX_TEST_AUDIO_ACQUIRE:
			testresult = CS42XX_TestAudio_Acquire(false);
			break;

		case CS42XX_TEST_AUDIO_ACQUIRE_RESTIT:
			testresult = CS42XX_TestAudio_Acquire(true);
			break;

		case CS42XX_TEST_AUDIO_LOOPINOUT:
			testresult = CS42XX_TestAudio_LoopbackInOut();
			break;

		case CS42XX_TEST_AUDIO_SENDRAWFILE:
			{
				void *ptr;

				testresult = CS42XX_TestAudio_SendRawFile();

				ptr = cyg_mbox_get(CS42XXTestAudioData.mailbox);
				cyg_mutex_lock(&CS42XXTestAudioData.testmutex);
				CS42XXTestAudioData.testcurr = *(CS42XX_TEST_ID *)ptr;
				cyg_mutex_unlock(&CS42XXTestAudioData.testmutex);

				CS42XX_TestHMI_PostResult(testresult);
				testresult = CYGNUM_TEST_NA;
			}
			break;

		//case CS42XX_TEST_NONE:
		default:
			// No test: block until a new test is asked
			{
				void *ptr;
				ptr = cyg_mbox_get(CS42XXTestAudioData.mailbox);

				cyg_mutex_lock(&CS42XXTestAudioData.testmutex);
				CS42XXTestAudioData.testcurr = *(CS42XX_TEST_ID *)ptr;
				cyg_mutex_unlock(&CS42XXTestAudioData.testmutex);
			}
			testresult = CYGNUM_TEST_NA;
			break;
		}

		// Post test results
		if (CYGNUM_TEST_NA != testresult)
		{
			CS42XX_TestHMI_PostResult(testresult);
		}
	}
	// Loop infinitely.
}


/** Look for pending messages in mailbox; if any, take the first message
 *  and store it as the new test Id.
 *
 * This function is not blocking.
 *
 * @retval true   A message has been found and stored as new test Id
 * @retval false  No message pending
 ****************************************************************************/
static cyg_bool
CS42XX_TestAudio_TestMailbox(void)
{
	cyg_bool newmsg = false;
	void *ptr;

	ptr = cyg_mbox_tryget(CS42XXTestAudioData.mailbox);
	if (NULL != ptr)
	{
		// New message available
		newmsg = true;

		// Store it as a new test Id
		cyg_mutex_lock(&CS42XXTestAudioData.testmutex);
		CS42XXTestAudioData.testcurr = *(CS42XX_TEST_ID *)ptr;
		cyg_mutex_unlock(&CS42XXTestAudioData.testmutex);
	}

	return newmsg;
}


/** Select the audio test to be run.
 *
 * @param[in] testid  Audio test to be run
 ****************************************************************************/
 void
CS42XX_TestAudio_Run(CS42XX_TEST_ID testid)
{
	// Post a new command to the thread
	CS42XXTestAudioData.testreq = testid;
	cyg_mbox_put(CS42XXTestAudioData.mailbox,
				&CS42XXTestAudioData.testreq);
}


/** Configure digital output of AAI for CS42XX testing purposes.
 *
 * @retval  ENOERR   Success
 * @retval -EINVAL   No such device
 ****************************************************************************/
static Cyg_ErrNo
CS42XX_TestAudio_InitOut(void)
{
	Cyg_ErrNo err;

	diag_printf("\r\nInit audio Out (%s)...",
			CS42XX_TEST_AAIDEVICE_OUT);

	err = cyg_io_lookup(CS42XX_TEST_AAIDEVICE_OUT,
			&CS42XXTestAudioAAIOut);
	if (ENOERR == err)
	{
		// Set AAI mode
		err = cyg_io_set_config_word(CS42XXTestAudioAAIOut,
					CYG_IO_SET_CONFIG_AAI_MODE, AAI_DSP);
		if (ENOERR != err)
		{
			CS42XX_TEST_FAIL_ERRNO("AAI mode", err);
		}

		// Set appropriate MCLK: f(MCLK) = 256*f(AC_SYNC)
		err = cyg_io_set_config_word(CS42XXTestAudioAAIOut,
					CYG_IO_SET_CONFIG_AAI_MCLK_FREQ, AAI_MCLK_FREQ_256);
		if (ENOERR != err)
		{
			CS42XX_TEST_FAIL_ERRNO("MCLK", err);
		}

		// No channel drop
		err = cyg_io_set_config(CS42XXTestAudioAAIOut,
					CYG_IO_SET_CONFIG_AAI_CHANNEL_DROP, NULL, NULL);
		if (ENOERR != err)
		{
			CS42XX_TEST_FAIL_ERRNO("channel drop", err);
		}

		// Set synchronization threshold
		err = cyg_io_set_config_word(CS42XXTestAudioAAIOut,
				CYG_IO_SET_CONFIG_AAI_CHANNEL_SYNC_THRESHOLD, 0);
		if (ENOERR != err)
		{
			CS42XX_TEST_FAIL_ERRNO("synchronization threshold", err);
		}

		// Set blocking mode
		err = cyg_io_set_config_word(CS42XXTestAudioAAIOut,
				CYG_IO_SET_CONFIG_AAI_CHANNEL_BLOCKING_MODE, true);
		if (ENOERR != err)
		{
			CS42XX_TEST_FAIL_ERRNO("blocking mode", err);
		}
	}
	else
	{
		diag_printf("\r\n*** %d: no such device",
				CS42XX_TEST_AAIDEVICE_OUT);
	}

	return err;
}

/** Configure digital input of AAI for CS42XX testing purposes.
 *
 * @retval  ENOERR   Success
 * @retval -EINVAL   No such device
 ****************************************************************************/
static Cyg_ErrNo
CS42XX_TestAudio_InitIn(void)
{
	Cyg_ErrNo err;
	cyg_uint32 aaicfg;
	cyg_io_handle_t aaiout = CS42XXTestAudioAAIOut;
	cyg_uint32 len = sizeof(CS42XXTestAudioAAIOut);
	cyg_uint32 enablemic = 1;

	diag_printf("\r\nInit audio In (%s)...",
			CS42XX_TEST_AAIDEVICE_IN);

	err = cyg_io_lookup(CS42XX_TEST_AAIDEVICE_IN,
			&CS42XXTestAudioAAIIn);
	if (ENOERR == err)
	{

		// Set bit micro src to 1
		err  = cyg_io_set_config(CS42XXTestAudioAAIIn,
					CYG_IO_SET_CONFIG_AAI_MICRO_SRC, &enablemic, &len);
		if (ENOERR != err)
		{
			 diag_printf("Micro src mode : %s (%d)\n", strerror(err), err);
		}

		// Set AAI mode
		err = cyg_io_set_config_word(CS42XXTestAudioAAIIn,
					CYG_IO_SET_CONFIG_AAI_MODE, AAI_DSP);
		if (ENOERR != err)
		{
			CS42XX_TEST_FAIL_ERRNO("AAI mode", err);
		}

		// Set appropriate MCLK: f(MCLK) = 256*f(AC_SYNC)
		err = cyg_io_set_config_word(CS42XXTestAudioAAIIn,
					CYG_IO_SET_CONFIG_AAI_MCLK_FREQ, AAI_MCLK_FREQ_256);
		if (ENOERR != err)
		{
			CS42XX_TEST_FAIL_ERRNO("MCLK", err);
		}

		// Disable Parrot5 internal ADC and select I�S input AC_IN
		err = cyg_io_get_config_word(CS42XXTestAudioAAIIn,
				CYG_IO_GET_CONFIG_AAI_REG_CFG, &aaicfg);
		if (ENOERR != err)
		{
			CS42XX_TEST_FAIL_ERRNO("internal ADC", err);
		}

		aaicfg &= ~AAI_CFG_RUN_ADC;   // Disable internal ADC
		aaicfg |=  AAI_CFG_MIC34_PAD; // Select AC_IN
		err = cyg_io_set_config_word(CS42XXTestAudioAAIIn,
				CYG_IO_SET_CONFIG_AAI_REG_CFG, aaicfg);
		if (ENOERR != err)
		{
			CS42XX_TEST_FAIL_ERRNO("I�S input", err);
		}

		// Set related channel
/*		err = cyg_io_set_config(CS42XXTestAudioAAIIn,
				CYG_IO_SET_CONFIG_AAI_CHANNEL_DC_MODE,
				&aaiout, &len);
		if (ENOERR != err)
		{
			CS42XX_TEST_FAIL_ERRNO("related channel", err);
		}
*/

		// No channel drop
		err = cyg_io_set_config(CS42XXTestAudioAAIIn,
					CYG_IO_SET_CONFIG_AAI_CHANNEL_DROP, NULL, NULL);
		if (ENOERR != err)
		{
			CS42XX_TEST_FAIL_ERRNO("channel drop", err);
		}

		// Set synchronization threshold
		err = cyg_io_set_config_word(CS42XXTestAudioAAIIn,
				CYG_IO_SET_CONFIG_AAI_CHANNEL_SYNC_THRESHOLD, 0);
		if (ENOERR != err)
		{
			CS42XX_TEST_FAIL_ERRNO("synchronization threshold", err);
		}

		// Set blocking mode
		err = cyg_io_set_config_word(CS42XXTestAudioAAIIn,
				CYG_IO_SET_CONFIG_AAI_CHANNEL_BLOCKING_MODE, true);
		if (ENOERR != err)
		{
			CS42XX_TEST_FAIL_ERRNO("blocking mode", err);
		}
	}
	else
	{
		diag_printf("\r\n*** %d: no such device",
				CS42XX_TEST_AAIDEVICE_IN);
	}

	return err;
}


/** Configure audio part of CS42XX driver testing software.
 *
 * @warning  This part is platform-specific.
 *
 * @retval true   Initialization complete, thread launched
 * @retval false  Failure during initialization, no thread launched
 ****************************************************************************/
 cyg_bool
CS42XX_TestAudio_Init(void)
{
	Cyg_ErrNo err;

	// Initialize global variables
	CS42XXTestAudioData.testcurr = CS42XX_TEST_NONE;
	CS42XXTestAudioData.testreq  = CS42XX_TEST_NONE;
	CS42XXTestAudioData.functcurr = CS42XX_TESTAUDIO_NOFUNCT;
	CS42XXTestAudioData.functreq  = CS42XX_TESTAUDIO_NOFUNCT;
	cyg_mutex_init(&CS42XXTestAudioData.testmutex);

	// Create a mailbox to command audio testing thread
	cyg_mbox_create(&CS42XXTestAudioData.mailbox,
					&CS42XXTestAudioData.mailboxobj);

	CS42XXTestAudioAAIOut = NULL;
	CS42XXTestAudioAAIIn  = NULL;

	// Configure AAI output
	err = CS42XX_TestAudio_InitOut();
	if (ENOERR == err)
	{
		// Configure AAI input
		err = CS42XX_TestAudio_InitIn();
	}
	// else output configuration failed.

	if (ENOERR == err)
	{
		// Start the audio testing thread
		cyg_thread_create(CS42XX_TESTAUDIO_THRPRIORITY,
				CS42XX_TestAudio_Thread, 0, "CS42XX audio testing",
				&CS42XXTestAudioData.stack[0], CS42XX_TESTAUDIO_STACKSIZE,
				&CS42XXTestAudioData.thread, &CS42XXTestAudioData.threadobj);
		cyg_thread_resume(CS42XXTestAudioData.thread);
	}

	return (ENOERR == err);
}



#endif // CS42XX_TEST_SUPPORT_AUDIO
