//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================

/**
 ***************************************************************************
 * @file sta326.h
 *
 * @brief STA326 digital amplifier driver.
 *
 * Copyright (C) 2006 Parrot S.A.
 *
 * @author	Franois Ferrand
 * @date	04/2006
 ***************************************************************************
 */

#include <cyg/infra/cyg_ass.h>
#include <cyg/devs/sta326.h>

/** Digital amplifier I2C device.
 */
CYG_I2C_DEVICE(STA326_i2c_device,		\
			   NULL,					\
			   0x34>>1,					\
			   0x00,					\
			   CYG_I2C_DEFAULT_DELAY);

cyg_uint8 STA326_mute;			///< Mute state
cyg_mutex_t STA326_mute_mutex;	///< Mutex to access STA326_mute


/** Initialize the amplifier.
 * This function should be used on startup, to configure the amplifier. It
 * does not perform a reset, so the caller should take care of that before
 * calling the function, to ensure the parameters are properly set.
 *
 * Once the function has been called, the other APIs can be used to perform
 * other commands (setup, volume change...).
 *
 * @param config A structure containing the configuration of the amplifier.
 * This structure is used only during the call, it does not need to be valid
 * afterward.
 */
void STA326_Init(cyg_i2c_bus* bus, cyg_uint16 addr, const STA326_CONFIG * config)
{
	cyg_uint8 cfgbuf[]= {
		config->confA,
		config->confB,
		config->confC,
		config->confD,
		config->confE,
		config->confF,
		STA326_MUTE_MASTER,	//master mute is set by default
		config->volumeMaster,
		config->volumeChan1,
		config->volumeChan2,
		config->volumeChan3
	};

	CYG_ASSERT(bus, "NULL pointer arguments: bus");
	CYG_ASSERT(config, "NULL pointer arguments: config");

	cyg_mutex_init(&STA326_mute_mutex);
	STA326_mute = STA326_MUTE_MASTER;
	
	STA326_i2c_device.i2c_bus = bus;
	STA326_i2c_device.i2c_address = (addr >> 1);
	STA326_Write(STA326_REG_ConfA, cfgbuf, sizeof(cfgbuf));
}

/** Mute any of the channels.
 * @param mute A combination of the following flags:
 * 			- STA326_MUTE_MASTER mute all channels
 *			- STA326_MUTE_CHAN1 mute channel 1
 * 			- STA326_MUTE_CHAN2 mute channel 2
 * 			- STA326_MUTE_CHAN3 mute channel 3
 *
 * To unmute, call the STA326_UnMute() function.
 * A channel is actually muted if STA326_MUTE_CHAN<x> or STA326_MUTE_MASTER
 * is set. To unmute it, both flags must be passed to the unmute function.
 */
void STA326_Mute(cyg_uint8 mute)
{
	cyg_mutex_lock(&STA326_mute_mutex);
	if ((STA326_mute | mute) != STA326_mute)
	{
		STA326_mute |= mute;
		STA326_Write(STA326_REG_Mute, &STA326_mute, 1);
	}
	cyg_mutex_unlock(&STA326_mute_mutex);
}

/** Unmute any of the channels.
 * @param mute A combination of the following flags:
 * 			- STA326_MUTE_MASTER unmute all channels
 *			- STA326_MUTE_CHAN1 unmute channel 1
 * 			- STA326_MUTE_CHAN2 unmute channel 2
 * 			- STA326_MUTE_CHAN3 unmute channel 3
 *
 * A channel is actually muted if STA326_MUTE_CHAN<x> or STA326_MUTE_MASTER
 * is set. To unmute it, both flags must be passed to the function.
 */
void STA326_UnMute(cyg_uint8 mute)
{
	cyg_mutex_lock(&STA326_mute_mutex);
	if (STA326_mute & mute)
	{
		STA326_mute &= ~mute;
		STA326_Write(STA326_REG_Mute, &STA326_mute, 1);
	}
	cyg_mutex_unlock(&STA326_mute_mutex);
}

void STA326_Write(cyg_uint16 addr, const cyg_uint8 * buf, cyg_uint8 len)
{
	cyg_uint8 tx_data[1];

	CYG_ASSERT(STA326_i2c_device.i2c_bus, "Call STA326_Init() first");
	CYG_ASSERT(buf, "NULL pointer arguments: buf");

	// address:
	tx_data[0] = addr;

	// lock
	cyg_i2c_transaction_begin(&STA326_i2c_device);

	// send the address: start, no stop
	cyg_i2c_transaction_tx(&STA326_i2c_device, true, tx_data, 1, false);
	// send the data: start, stop
	cyg_i2c_transaction_tx(&STA326_i2c_device, false, buf, len, true);

	// unlock
	cyg_i2c_transaction_end(&STA326_i2c_device);
}

void STA326_Read(cyg_uint16 addr, cyg_uint8 * buf , cyg_uint8 len )
{
	cyg_uint8 tx_data[1];

	CYG_ASSERT(STA326_i2c_device.i2c_bus, "Call STA326_Init() first");
	CYG_ASSERT(buf, "NULL pointer arguments: buf");

	// address:
	tx_data[0] = addr ;

	// lock
	cyg_i2c_transaction_begin(&STA326_i2c_device);

	// send the address: start, no stop
	cyg_i2c_transaction_tx(&STA326_i2c_device, true, tx_data, 1, true);
	// receive the data: start, stop
	cyg_i2c_transaction_rx(&STA326_i2c_device, true, buf, len, true, true);

	// unlock
	cyg_i2c_transaction_end(&STA326_i2c_device);
}

void STA326_Coeff_Write(STA326_Coeff coeff, int value)
{
	unsigned char command[20];
	int pos = 0;

	CYG_ASSERT(STA326_i2c_device.i2c_bus, "Call STA326_Init() first");

	//lock
	cyg_i2c_transaction_begin(&STA326_i2c_device);

	// First, send the coefficient
	command[0] = STA326_REG_Cfaddr2;
	cyg_i2c_transaction_tx(&STA326_i2c_device, true, command, 1, false);

	command[pos++] = coeff;
	CYG_ASSERT(!(value >> 24) || !(~(value >> 24)), "coeffs are 24bit.");
	command[pos++] = (value >> 16) & 0xff;
	command[pos++] = (value >> 8) & 0xff;
	command[pos++] = (value) & 0xff;
	cyg_i2c_transaction_tx(&STA326_i2c_device, false, command, pos, true);

	// Then, send write command
	command[0] = STA326_REG_Cfud;
	cyg_i2c_transaction_tx(&STA326_i2c_device, true, command, 1, false);

	command[0] = STA326_Coeff_WriteOne;
	cyg_i2c_transaction_tx(&STA326_i2c_device, false, command, 1, true);

	// unlock
	cyg_i2c_transaction_end(&STA326_i2c_device);
}

void STA326_Coeff_Read(STA326_Coeff coeff, int *value)
{
	unsigned char command[20];
	
	CYG_ASSERT(STA326_i2c_device.i2c_bus, "Call STA326_Init() first");

	//lock
	cyg_i2c_transaction_begin(&STA326_i2c_device);

	// First write the address to STA326_REG_Cfaddr2
	command[0] = STA326_REG_Cfaddr2;
	command[1] = coeff;
	cyg_i2c_transaction_tx(&STA326_i2c_device, true, command, 2, true);

	// Then, write R1 to STA326_REG_Cfud
	command[0] = STA326_REG_Cfud;
	command[1] = STA326_Coeff_ReadOne;
	cyg_i2c_transaction_tx(&STA326_i2c_device, true, command, 2, true);

	// Then read coeff
	command[0] = STA326_REG_B1cf1;
	cyg_i2c_transaction_tx(&STA326_i2c_device, true, command, 1, false);
	cyg_i2c_transaction_rx(&STA326_i2c_device, true, command, 3, true, true);

	*value = (command[0] << 16) | (command[1] << 8) | command[2];
	
	// unlock
	cyg_i2c_transaction_end(&STA326_i2c_device);
	
}

void STA326_Coeff_WriteBiquad(STA326_Coeff_BIQUAD biquad, int b1, int b2, int a1, int a2, int b0)
{
	unsigned char command[20];
	int coeffs[5] = { b1, b2, a1, a2, b0 };
	int pos = 0;
	int i;

	command[pos++] = biquad;
	for(i=0; i<5; i++)
	{
		CYG_ASSERT(!(coeffs[i] >> 24) || !(~(coeffs[i] >> 24)), "coeffs are 24bit.");
		command[pos++] = (coeffs[i] >> 16) & 0xff;
		command[pos++] = (coeffs[i] >> 8) & 0xff;
		command[pos++] = (coeffs[i]) & 0xff;
	}
	command[pos++] = STA326_Coeff_WriteAll;

	STA326_Write(STA326_REG_Cfaddr2, command, pos);
}

