//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================

/**
 ****************************************************************************
 * Driver for Wolfson WM8731 stereo audio codec.
 *
 * @file   WM8731_i2c.c
 * @brief  Interface with IC driver.
 *
 * Code is based on the WM8731.pdf
 *
 * @author Marc Flambard
 * @date   2007-05-24
 *
 * Copyright (C) 2007, Parrot S.A. - www.parrot.biz
 ****************************************************************************
 */

#include <cyg/infra/cyg_ass.h>
#include <cyg/io/i2c.h>
#include <cyg/devs/wm8731.h>


/*
 * Macros.
 **********/

/** Declare a WM8731 device. */
#define WM8731_I2C_DEVICE()						\
	{											\
		.i2c_bus     = NULL,					\
		.i2c_address = 0x00,					\
		.i2c_flags   = 0x00,					\
		.i2c_delay   = CYG_I2C_DEFAULT_DELAY	\
	}

/** Get the IC address of a WM8731 device on IC bus from ad0.
 *
 * @def   WM8731_I2C_ADDR
 * @param ad0  State of AD0 pin on device (boolean)
 ****************************************************************************/
#define WM8731_I2C_ADDR(ad0)		(0x1A | (ad0))


/*
 * Variables.
 *************/

/** List of WM8731 devices.
 *
 * @var WM8731_I2C_Device
 */
cyg_i2c_device WM8731_I2C_Device[WM8731_NBDEVICES] = {
#if WM8731_NBDEVICES < 1
#error "Invalid number of WM8731 devices configured."
#endif
		WM8731_I2C_DEVICE(),
#if WM8731_NBDEVICES > 1
		WM8731_I2C_DEVICE(),
#endif
#if WM8731_NBDEVICES > 2
#error "Too many WM8731 devices configured; please update driver."
#endif
};


/*
 * Definition of functions.
 ***************************/


/** Get the IC device associated to a symbolic device Id.
 *
 * @param devid  Symbolic device Id
 *
 * @return       Pointer to an IC device
 * @retval NULL  Invalid device Id
 ****************************************************************************/
static inline cyg_i2c_device *
WM8731_I2C_GetDevice(const WM8731_DEVID devid)
{
	return ((devid < WM8731_NBDEVICES)?
		&WM8731_I2C_Device[devid] : NULL);
}


/** Attach a symbolic device to a bus.
 * This function "fills a slot" by setting address of device and bus to
 * which it is connected.
 *
 * @param devid  Symbolic device Id
 * @param bus    Bus to attach device to
 * @param ad0    State of ad0 pin on audio device
 *
 * @retval  ENOERR  Operation successful
 * @retval -EINVAL  Invalid parameters
 * @retval -ENODEV  No such device
 ****************************************************************************/
Cyg_ErrNo
WM8731_I2C_AttachBus(const WM8731_DEVID devid, const cyg_i2c_bus *bus,
		const cyg_bool ad0)
{
	Cyg_ErrNo       err = -EINVAL;
	cyg_i2c_device *dev;

	if (NULL != bus)
	{
		err = -ENODEV;
		dev = WM8731_I2C_GetDevice(devid);
		if (dev)
		{
			dev->i2c_bus     = (cyg_i2c_bus *)bus;
			dev->i2c_address = WM8731_I2C_ADDR(ad0);
			err = ENOERR;
		}
		// else invalid device Id.
	}
	// else invalid parameter.
	
	return err;
}


/** Write one or successive bytes in device memory, at given location.
 *
 * @param devid    IC device to access to
 * @param addr     Address where to locate the data
 * @param data     data to be sent
 *
 * @retval  ENOERR   Operation successful
 * @retval -ENODEV   No such device
 * @retval -EIO      Error during transmission
 ****************************************************************************/
Cyg_ErrNo
WM8731_I2C_Write(const WM8731_DEVID devid,
		cyg_uint8 addr, const cyg_uint16 data)
{
	Cyg_ErrNo       err = -ENODEV;
	cyg_i2c_device *dev;
	cyg_uint32      cnt;
	cyg_uint8       TxData[2];
	
	TxData[0]=(addr<<1)|((data>>8)&1);
	TxData[1]=data&0xFF;
	
	dev = WM8731_I2C_GetDevice(devid);

	if (dev && dev->i2c_address)
	{
		CYG_ASSERT(dev->i2c_bus, "Call WM8731_I2C_AttachBus first");
		
		// Lock IC
		cyg_i2c_transaction_begin(dev);
		
		// START(, devaddr), address, data, STOP
		cnt = cyg_i2c_transaction_tx(dev, true, TxData, 2, true);
		if(cnt==2) err = ENOERR;
		
		// Unlock IC
		cyg_i2c_transaction_end(dev);
	}
	// else no such device (no slot or wrong IC address).

	return err;
}

