//==========================================================================
//
//      devs/watchdog/arm/parrot5/watchdog_parrot5.cxx
//
//      Watchdog implementation for the Parrot5 ASIC
//
//==========================================================================
// Author(s):    ivan.djelic@parrot.fr
// Contributors:
// Date:         2006-04-18
// Purpose:      Watchdog class implementation
// Description:  Contains an implementation of the Watchdog class for use
//               with the Parrot5.
//==========================================================================

#include <pkgconf/kernel.h>
#include <pkgconf/infra.h>
#include <pkgconf/kernel.h>
#include <pkgconf/watchdog.h>
#include <pkgconf/devs_watchdog_parrot5.h>

#include <cyg/infra/cyg_type.h>
#include <cyg/infra/cyg_ass.h>
#include <cyg/infra/cyg_trac.h>
#include <cyg/hal/hal_io.h>
#include <cyg/hal/hal_diag.h>

#include <cyg/hal/p5dev.h>

#include <cyg/io/watchdog.hxx>

#if !defined(CYGSEM_WATCHDOG_RESETS_ON_TIMEOUT)
#include <cyg/hal/hal_platform_ints.h>
#include <cyg/kernel/intr.hxx>
#endif

//==========================================================================

#define RESOLUTION  ((cyg_uint64)(CYGNUM_DEVS_WATCHDOG_PARROT5_DESIRED_TIMEOUT_MS)*1000000)
#define TICKDIV     ((CYGNUM_DEVS_WATCHDOG_PARROT5_DESIRED_TIMEOUT_MS)*32/10-1)

//==========================================================================

static cyg_uint32 wdogctl = SYS_WDOGCTL_TICKEN;

#if defined(CYGSEM_WATCHDOG_RESETS_ON_TIMEOUT)

void
Cyg_Watchdog::init_hw(void)
{
  CYG_REPORT_FUNCTION();
  CYG_REPORT_FUNCARGVOID();
  resolution = RESOLUTION;
  /* enable hardware watchdog reset */
  wdogctl |= SYS_WDOGCTL_WDOGEN;
  CYG_REPORT_RETURN();
}

#else /* defined(CYGSEM_WATCHDOG_RESETS_ON_TIMEOUT) */

//==========================================================================

static Cyg_Watchdog *wd;

//==========================================================================

static cyg_uint32
isr(cyg_vector vector, CYG_ADDRWORD data)
{
  CYG_REPORT_FUNCTION();
  CYG_REPORT_FUNCARG2XV(vector, data);

  // clear interrupt flag
  HAL_WRITE_UINT32( PARROT5_SYS+_SYS_FLAGS, SYS_FLAGS_TICK );

  wd->trigger();
  Cyg_Interrupt::acknowledge_interrupt(CYGNUM_HAL_INTERRUPT_TICK);
  CYG_REPORT_RETVAL(Cyg_Interrupt::HANDLED);
  return Cyg_Interrupt::HANDLED;
}

//==========================================================================

static Cyg_Interrupt CYGBLD_ATTRIB_INIT_PRI( CYG_INIT_INTERRUPTS ) wdint(
    CYGNUM_HAL_INTERRUPT_TICK,
    4,
    0,
    isr,
    NULL
  );

//==========================================================================

void
Cyg_Watchdog::init_hw(void)
{
  CYG_REPORT_FUNCTION();
  CYG_REPORT_FUNCARGVOID();
  resolution = RESOLUTION;

#if defined(CYGSEM_DEVS_WATCHDOG_PARROT5_FORCE_WDOGRESET)
  {
	  /* if DEBUG is disabled, do not enable IRQ and force WDOG reset */
	  cyg_uint32 reg;
	  HAL_READ_UINT32(PARROT5_SYS+_SYS_REMAP, reg);
	  if ((reg & SYS_REMAP_DEBUGEN) == 0) {
		  wdogctl |= SYS_WDOGCTL_WDOGEN;
		  return;
	  }
  }
#endif /* CYGSEM_DEVS_WATCHDOG_PARROT5_FORCE_WDOGRESET */

  wd = this;
  wdint.configure_interrupt(CYGNUM_HAL_INTERRUPT_TICK, false, true);
  wdint.attach();
  wdint.acknowledge_interrupt(CYGNUM_HAL_INTERRUPT_TICK);
  wdint.unmask_interrupt(CYGNUM_HAL_INTERRUPT_TICK);
  CYG_REPORT_RETURN();
}

#endif	/* defined(CYGSEM_WATCHDOG_RESETS_ON_TIMEOUT) */

//==========================================================================
/*
 * Reset watchdog timer. This needs to be called regularly to prevent
 * the watchdog from firing.
 */

void
Cyg_Watchdog::reset(void)
{
  CYG_REPORT_FUNCTION();
  CYG_REPORT_FUNCARGVOID();

  // clear interrupt flag
  HAL_WRITE_UINT32( PARROT5_SYS+_SYS_FLAGS, SYS_FLAGS_TICK );
  // stop Watchdog
  HAL_WRITE_UINT32( PARROT5_SYS+_SYS_WDOGCTL, 0 );
  // and restart
  HAL_WRITE_UINT32( PARROT5_SYS+_SYS_WDOGCTL, TICKDIV|wdogctl );

  CYG_REPORT_RETURN();
}

//==========================================================================
/*
 * Start watchdog to generate a hardware reset
 * or interrupt when expiring.
 */

void
Cyg_Watchdog::start(void)
{
    CYG_REPORT_FUNCTION();
    CYG_REPORT_FUNCARGVOID();

    // clear interrupt flag
    HAL_WRITE_UINT32( PARROT5_SYS+_SYS_FLAGS, SYS_FLAGS_TICK );
    // stop Watchdog
    HAL_WRITE_UINT32( PARROT5_SYS+_SYS_WDOGCTL, 0 );
    // start tick counter
    HAL_WRITE_UINT32( PARROT5_SYS+_SYS_WDOGCTL, TICKDIV|wdogctl );

    CYG_REPORT_RETURN();
}

//==========================================================================
// End of watchdog_parrot5.cxx
