#include "fatfs_supp.h"


/** Fill stavfs structure.
 *
 * @param disk A pointer to the disk
 * @param info A pointer to the statvfs structure
 * @return ENOERR in any case
 */
int fatfs_statvfs(fatfs_disk_t *disk,
                  statvfs_t    *info)
{
   cyg_uint8  data[0x5A];
   cyg_uint32 len;
#ifdef CYGCFG_FS_FAT_WINFREESPC
   cyg_uint32 clean;
#endif
   cyg_uint32 fbr_sec_per_fat;  // 0x0B
   int err = ENOERR;

   len = 0x5A;
   err = disk_read(disk, (void*)data, &len, 0);
   if (err != ENOERR)
      return err;

   GET_WORD(data,  fbr_sec_per_fat,    0x16);

   // This is a quick check for FAT12/16 or FAT32 boot record.
   // The sec_per_fat field must be 0 on FAT32, since this
   // field plays a crucial role in detection of the FAT type
   // (12,16,32) it is quite safe to make this assumption.
   if ( (0 == fbr_sec_per_fat) && 
        (disk->fat_type == FATFS_FAT32) )
   {
#ifdef CYGCFG_FS_FAT_FASTSTAT
       disk->read_fsinfo = true;
#endif

#ifdef CYGCFG_FS_FAT_WINFREESPC
      //check windows clean freecount_clust info
      GET_BYTE(data, clean,0x41); 

      // if disk is marked as clean or if freespace has already been computed
      if( ((clean&0x1) == 0) || (disk->read_fsinfo == true) )
#else
      // if freespace has already been computed  
      if(disk->read_fsinfo == true) 
#endif
      {
          fatfs_get_fat32_fsinfo(disk, 
                                     &(info->f_blocks),
                                     &(info->f_bfree));
      }
      else
      {
          err = fatfs_get_disk_usage(disk,
                                     &(info->f_blocks),
                                     &(info->f_bfree));
          if (err == ENOERR){
            err = fatfs_write_fat32_fsinfo(disk, info->f_bfree, 0);
            disk->read_fsinfo = true;
          }
      }
   }
   else
   {// FAT12/FAT16
      err = fatfs_get_disk_usage(disk,
                                 &(info->f_blocks),
                                 &(info->f_bfree));
   }

   if(err == ENOERR)
   {
      info->f_bsize   = disk->cluster_size;
      info->f_frsize  = disk->cluster_size;
      info->f_bavail  = info->f_bfree;
      // f_files is not supported, but could easily be added if needed...
      info->f_files   = 0;
      info->f_ffree   = 0;
      info->f_favail  = 0;
      info->f_fsid = (cyg_uint32)disk;
      // this is PFFS specific...
      info->f_flag    = 0;
#ifdef CYGCFG_FS_FAT_USE_VFAT
      info->f_namemax = LNAME_MAX_CHAR;
#else
      info->f_namemax = SNAME_MAX_CHAR;
#endif
   }

   return err;
}

/** Browse root dir to find the volume label entry.
 *
 * @param disk        A pointer to the disk
 * @param raw_cdentry A pointer to the volume label entry
 * @param pos        A pointer to the entry position
 *
 * @return EEOF if no volume label is set
 */
int fatfs_find_label(fatfs_disk_t        *disk,
                     fat_raw_dir_entry_t *raw_cdentry,
                     fatfs_data_pos_t    *pos)
{
   int err = ENOERR;

   pos->cluster      = disk->fat_root_dir_cluster;
   pos->cluster_snum = 0;
   pos->cluster_pos  = 0;

   // Browse root dir entries
   while (true)
   {
        // FAT12/16 root dir check
        if (0 == pos->cluster) {
            if (pos->cluster_pos >= disk->fat_root_dir_size){
                FAT_INFO("Volume label not found: FAT16, end of root dir reached\n");
                err = ENOSPC;
            }
        }
        else {  // Change cluster if needed
            if (pos->cluster_pos >= disk->cluster_size){
                err = get_next_cluster(disk, pos, CO_NONE);
                if (err != ENOERR)
                    break;

                FAT_INFO("searching in next root cluster %d\n, pos->cluster");                
            }
        }//end if (0 == pos->cluster)

        err = read_raw_dentry(disk, pos, raw_cdentry);
        if (err != ENOERR){   
            break;
        }

        if (DENTRY_IS_ZERO(raw_cdentry))
        {   // If we get a ZERO dir entry, we assume that
            // there are no more entries in current dir
            FAT_INFO("Volume label not found: FAT16, no more entries 0x%x\n", pos->cluster_pos);
            err = EEOF;
            break;
        }
        else if ( DENTRY_IS_VOLUME(raw_cdentry) && !DENTRY_IS_DELETED(raw_cdentry) )
        {   // We found the volume id
            err = ENOERR;
            break;
        }

#ifdef CYGCFG_FS_FAT_USE_VFAT
        if ( DENTRY_IS_DELETED(raw_cdentry) ) {
            pos->cluster_pos += DENTRY_SIZE;
        }else {
            err = go_to_next_raw_entry(disk, pos, raw_cdentry, CO_NONE);
            if (err != ENOERR)
               return err;
        }
#else
        pos->cluster_pos += DENTRY_SIZE;
#endif
    } //while

   return err;
}

void
init_fatfs_fds(fatfs_disk_t *disk)
{
    int i;

    for (i = 0; i < CYGNUM_FILEIO_NFD; i++)
    {
        disk->fatfs_fds_pool[i] = &disk->fatfs_fds_base[i];
    }
    disk->fatfs_fds_free_cnt = i;
}

fatfs_fd_t *
alloc_fatfs_fd(fatfs_disk_t *disk, fatfs_node_t *node)
{
    fatfs_fd_t *fd = NULL;

    if (disk->fatfs_fds_free_cnt > 0)
    {
        fd = disk->fatfs_fds_pool[--disk->fatfs_fds_free_cnt];

        fd->node = node;
        fatfs_initpos(disk, &node->dentry, &fd->pos);
    }

    return fd;
}

void
free_fatfs_fd(fatfs_disk_t *disk, fatfs_fd_t *fd)
{
    disk->fatfs_fds_pool[disk->fatfs_fds_free_cnt++] = fd;
}

void
init_dirsearch(fatfs_dirsearch_t *ds,
               fatfs_disk_t      *disk,
               fatfs_node_t      *dir,
               const char        *name)
{
    ds->disk = disk;

    if (NULL == dir)
        ds->dir = disk->root;
    else
        ds->dir = dir;

    ds->path    = name;
    ds->node    = ds->dir;
    ds->namelen = 0;
    ds->last    = false;
}


static int
find_direntry(fatfs_dirsearch_t *ds)
{
    fatfs_dir_entry_t  dentry;
    fatfs_data_pos_t   pos;
    int                err;

    CYG_TRACE1(TFS, "searching for dir entry '%s'", ds->name);

    memset(&dentry, 0, sizeof(fatfs_dir_entry_t));
    // First check the cache

    ds->node = fatfs_node_find(ds->disk,
                               ds->name,
                               ds->namelen,
                               ds->dir->dentry.cluster);

    if (ds->node != NULL)
    {
        // Dir entry found in cache
        CYG_TRACE0(TFS, "dir entry found in cache");

        fatfs_node_touch(ds->disk, ds->node);

        //GE-patch to increase listing speed adding next entry to the cache
        // Read next dir entry
        pos = ds->node->dentry.disk_pos;

        err = fatfs_read_next_dir_entry(ds->disk, &ds->dir->dentry, &pos, &dentry);
        // add this entry to the cache if it's valid
        if (err == ENOERR)
        {
            if(NULL == fatfs_node_find(ds->disk,
                                       dentry.filename,
                                       strlen(dentry.filename),
                                       ds->dir->dentry.cluster) )
            {
               fatfs_node_alloc(ds->disk, &dentry);
            }
        }
        //GE-patch to increase listing speed adding next entry to cache

        return ENOERR;
    }

    // Dir entry not in cache - search the current dir
    fatfs_initpos(ds->disk, &ds->dir->dentry, &pos);

    while (true)
    {
        // Read next dir entry

        err = fatfs_read_dir_entry(ds->disk, &ds->dir->dentry, &pos, &dentry);
        if (err != ENOERR)
            return (err == EEOF ? ENOERR : err);

        // Compare filenames

        if ('\0' == dentry.filename[ds->namelen] &&
               0 == strncasecmp(dentry.filename, ds->name, ds->namelen))
        {
            // Dir entry found - allocate new node and return

            CYG_TRACE0(TFS, "dir entry found");

            ds->node = fatfs_node_alloc(ds->disk, &dentry);
            if (NULL == ds->node)
                return EMFILE;

            //GE-patch to increase listing speed adding next enntry to the cache
            else
            {
               // Read next dir entry
               err = fatfs_read_dir_entry(ds->disk, &ds->dir->dentry, &pos, &dentry);
               if (err == ENOERR)
               {
                     if(NULL == fatfs_node_find(ds->disk,
                                                dentry.filename,
                                                strlen(dentry.filename),
                                                ds->dir->dentry.cluster) )
                     {
                        fatfs_node_alloc(ds->disk, &dentry);
                     }
               }
            }
            //GE-patch to increase listing speed adding next enntry to the cache

            return ENOERR;
        }
    }
}

static int
find_entry(fatfs_dirsearch_t *ds)
{
    const char  *name     = ds->path;
    const char  *n        = name;
    char         namelen  = 0;
    int          err;

    if( !S_ISDIR(ds->dir->dentry.mode) )
    {
        CYG_TRACE1(TFS, "entry '%s' not dir", ds->dir->dentry.filename);
        return ENOTDIR;
    }

    // Isolate the next element of the path name
    while (*n != '\0' && *n != '/')
        n++, namelen++;

    // If we terminated on a NUL, set last flag
    if (*n == '\0')
        ds->last = true;

    // Update name in dirsearch object
    ds->name    = name;
    ds->namelen = namelen;

    err = find_direntry(ds);
    if (err != ENOERR)
        return err;

    CYG_TRACE2(TFS, "entry '%s' %s", name, (ds->node ? "found" : "not found"));

    if (ds->node != NULL)
       return ENOERR;
    else
       return ENOENT;
}

int
fatfs_find(fatfs_dirsearch_t *ds)
{
    int err;

    CYG_TRACE1(TFS, "find path='%s'", ds->path);

    // Iterate down directory tree until we find the object we want
    for(;;)
    {
start:
        // Short circuit empty paths or special case with "."
        if (ds->path[0] == '\0' || (ds->path[0] == '.' && ds->path[1] == '\0'))
            return ENOERR;

        // Special case "./"
        if (ds->path[0] == '.' && ds->path[1] == '/') {
            ds->path+=2;
            goto start;
        }
        // Skip dirname separators
        if (*(ds->path) == '/') {
            ds->path++;
            goto start;
        }

        err = find_entry(ds);

        if (err != ENOERR)
            return err;

        if (ds->last)
        {
            CYG_TRACE0(TFS, "entry found");
            return ENOERR;
        }

        // Update dirsearch object to search next directory
        ds->dir   = ds->node;
        ds->path += ds->namelen;

        // Skip dirname separators
        while (*(ds->path) == '/') ds->path++;

        CYG_TRACE1(TFS, "find path to go='%s'", ds->path);
    }
}



// -------------------------------------------------------------------------
// disk_write()
// Writes data to disk.

int
disk_write(fatfs_disk_t *disk,
           void         *buf,
           cyg_uint32   *len,
           cyg_uint32    pos)
{
    int ret = 0;

    if (disk->bcache_mem) {
         CYG_ASSERT(disk->disk_write, "NULL pointer!");
         ret = disk->disk_write(&disk->blib, buf, len, 0, pos);
    } else {
         CYG_ASSERT(0, "write during boot!");
         ret = EINVAL;
    }

    return ret;
}

// -------------------------------------------------------------------------
// write_tentry()
// Writes FAT table entry to disk (to all copies of FAT).

int
write_tentry(fatfs_disk_t *disk, cyg_uint32 num, cyg_uint32 *entry)
{
    cyg_uint8  data[4];
    cyg_uint32 pos=0, num3;
    cyg_uint32 e;
    cyg_uint32 len;
    int err;
    int i;

    switch (disk->fat_type)
    {
        case FATFS_FAT12:
            num3 = num * 3;
            pos  = disk->fat_tbl_pos + (num3 >> 1);
            len  = 2;

            err = disk_read(disk, (void*)data, &len, pos);
            if (err != ENOERR)
                return err;

            GET_WORD(data, e, 0x00);

            if (0 == (num3 & 1)) e = (e & 0xF000) | (*entry & 0x0FFF);
            else                 e = (e & 0x000F) | ((*entry & 0x0FFF) << 4);

            SET_WORD(data, e, 0x00);

            break;

        case FATFS_FAT16:
            pos = disk->fat_tbl_pos + (num << 1);
            len = 2;

            e = *entry;
            SET_WORD(data, e, 0x00);

            break;

        case FATFS_FAT32:
            pos = disk->fat_tbl_pos + (num << 2);
            len = 4;

            err = disk_read(disk, (void*)data, &len, pos);
            if (err != ENOERR)
                return err;

            GET_DWORD(data, e, 0x00);

            e = (e & 0xF0000000) | *entry;

            SET_DWORD(data, e, 0x00);

            break;

        default:
            CYG_ASSERT(false, "Unknown FAT type");
    }

#ifdef CYGCFG_FS_FAT_ROBUST
    if(disk->robust == true)
    {
        err = disk_write(disk, (void*)data, &len, pos);
        if (err != ENOERR)
            return err;
    }
    else
#endif // CYGCFG_FS_FAT_ROBUST
    {
        for (i = 0; i < disk->fat_tbls_num; i++)
        {
            err = disk_write(disk, (void*)data, &len, pos);
            if (err != ENOERR)
                return err;

            pos += disk->fat_tbl_size;
        }
    }


    return ENOERR;
}

// -------------------------------------------------------------------------
// set_tentry_next_cluster()
// Sets the the next cluster number to FAT table entry.

static __inline__ void
set_tentry_next_cluster(fatfs_disk_t *disk,
                        cyg_uint32   *entry,
                        cyg_uint32    next_cluster)
{
    *entry = next_cluster;
}


// -------------------------------------------------------------------------
// link_cluster()
// Links two clusters.

static int
link_cluster(fatfs_disk_t *disk, cyg_uint32 cluster1, cyg_uint32 cluster2)
{
    cyg_uint32 tentry;

    set_tentry_next_cluster(disk, &tentry, cluster2);
    return write_tentry(disk, cluster1, &tentry);
}

// -------------------------------------------------------------------------
// find_and_append_cluster()
// Finds a free cluster on disk and appends it to the given cluster.
// New cluster is marked as LAST.

int
find_and_append_cluster(fatfs_disk_t   *disk,
                        cyg_uint32      cluster,
                        cyg_uint32     *new_cluster,
                        cluster_opts_t  opts)
{
    cyg_uint32 free_cluster;
    int        err;

    err = find_next_free_cluster(disk, cluster,
        &free_cluster, opts | CO_MARK_LAST | CO_ERASE_NEW);

    if (err != ENOERR)
        return err;

    err = link_cluster(disk, cluster, free_cluster);
    if (err != ENOERR)
        return err;

    *new_cluster = free_cluster;

    CYG_TRACE2(TCL, "appending new cluster=%d to cluster=%d",
                    free_cluster, cluster);

    return ENOERR;
}
