//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================

/**
 * ============================================================================
 * @file  exception.c
 * @brief Exception handlers
 *
 * Memory protection exception handlers.
 *
 * Copyright (C) 2004,2005 Parrot S.A.
 *
 * @author     ivan.djelic@parrot.fr
 * @date       2004-05-04
 * ============================================================================
 */

#include <cyg/kernel/kapi.h>
#include <cyg/infra/diag.h>
#include <cyg/infra/cyg_ass.h>
#include <cyg/infra/cyg_type.h>
#include <cyg/hal/hal_arch.h>
#include <pkgconf/memprot.h>

#include "mmu.h"
#include "diag.h"

//! declaration of the swi_handler routine from swi_handler.S
cyg_VSR_t swi_handler;

/** Invalid SWI (i.e. called from a user code section) handler.
 *
 * This function is called from swi_handler.S in SYSTEM mode, whenever a
 * SWI instruction was executed from a user code section (which is forbidden).
 * It prints a short error message before terminating the current thread.
 *
 * @param vaddr The faulting SWI instruction address
 */
void cyg_memprot_invalid_swi(CYG_ADDRESS vaddr) {
  cyg_thread * self = (cyg_thread *) cyg_thread_self();
  INFO("unvalid SWI call @ 0x%08x\n", vaddr );
  INFO("killing thread \"%s\" (handle = 0x%08x)\n", self->name,
       (unsigned int) self );
  cyg_thread_exit();
}

/** Stack overflow handler.
 *
 * This function is called from swi_handler.S in SYSTEM mode, whenever a
 * stack overflow was incidentally detected during a SWI call handling.
 * It prints a short error message before terminating the current thread.
 *
 * @param vaddr The SWI instruction address
 */
void cyg_memprot_stack_overflow(CYG_ADDRESS vaddr) {
  cyg_thread * self = (cyg_thread *) cyg_thread_self();
  INFO("stack overflow @ 0x%08x\n", vaddr );
  INFO("killing thread \"%s\" (handle = 0x%08x)\n", self->name,
	(unsigned int) self );
  cyg_thread_exit();
}

/** SWI handler installer.
 *
 * This function is called in SYSTEM mode during memory protection setup to
 * redirect SWI interrupt handling to swi_handler.S.
 */
void cyg_memprot_install_swi_handler(void) {
  DEBUG("installing SWI handler\n");
  cyg_interrupt_set_vsr( CYGNUM_HAL_VECTOR_SOFTWARE_INTERRUPT, swi_handler );
}

/** Thread last stop before exit.
 *
 * This function is called before killing a faulting thread, you can put a
 * permanent breakpoint in it for debugging.
 */
void cyg_memprot_breakme(void) {
  cyg_thread_exit();
}

/** The exception handler.
 *
 * This function is called in SYSTEM mode when an exception occurs.
 * It performs exception decoding and prints an adequate error message.
 * If the exception occured in
 * user mode, the current thread is terminated; otherwise the system is halted.
 * @param data This parameter is simply ignored
 * @param exception A code indicating which type of exception occured
 * @param info A pointer to a frame with all saved pre-exception registers
 */
void
cyg_memprot_exception_handler( cyg_addrword_t data, cyg_code_t exception,
			       cyg_addrword_t info ) {
  HAL_SavedRegisters *frame = (HAL_SavedRegisters *)info;
  CYG_ADDRESS vaddr;
  char *msg;
  int mode, domain, thread_domain;
  switch ( exception ) {
  case CYGNUM_HAL_EXCEPTION_ILLEGAL_INSTRUCTION:
    INFO("undefined instruction @ pc = 0x%08x\n", frame->pc );
    break;
  case CYGNUM_HAL_EXCEPTION_INTERRUPT:
    INFO("software interrupt @ pc = 0x%08x\n", frame->pc );
    break;
  case CYGNUM_HAL_EXCEPTION_DATA_ACCESS:
    memprot_mmu_get_data_abort_info( &vaddr, &domain, &msg );
    INFO("data abort @ pc = 0x%08x (%s)\n", frame->pc-4, msg );
    if ( domain >= 0 ) {
      INFO("[trying to access 0x%08x in domain %d]\n", vaddr, domain );
    }
    else {
      INFO("[trying to access 0x%08x]\n", vaddr );
    }
    break;
  case CYGNUM_HAL_EXCEPTION_CODE_ACCESS:
    INFO("prefetch abort @ pc = 0x%08x\n", frame->pc );
    break;
  default:
    INFO("unknown exception (%d) @ pc = 0x%08x\n", exception, frame->pc );
    break;
  }
  /* in which mode did the exception occur ? */
  mode = frame->cpsr & CPSR_MODE_BITS;
  if ( mode == CPSR_USER_MODE ) {
    cyg_thread * handle = (cyg_thread *) cyg_thread_self();
    /* we can recover from the exception by terminating the thread */
    if (( frame->sp > handle->stack_base + handle->stack_size )||
	( frame->sp < handle->stack_base + 128 ))
      // restore a good stack pointer
      frame->sp = handle->stack_base + handle->stack_size;
    // switch the faulting thread to system mode before calling exit()
    frame->cpsr &= ~CPSR_MODE_BITS;
    frame->cpsr |= CPSR_SYSTEM_MODE;
    if ( exception != CYGNUM_HAL_EXCEPTION_CODE_ACCESS )
      frame->lr = frame->pc;
    // get thread domain
    thread_domain = memprot_mmu_get_domain(frame->sp);
    // kill thread
    frame->pc = (cyg_uint32) cyg_memprot_breakme;
    INFO("killing thread \"%s\" (handle = 0x%08x, domain = %d)\n",
	 handle->name, (cyg_uint32)handle, thread_domain );
    return;
  }
  CYG_FAIL("\nException occured in a privileged mode ! Halting...");
}

/** Exception handler installer.
 *
 * This function is called in SYSTEM mode during memory protection setup to
 * install our own exception handler.
 */
void
cyg_memprot_install_exception_handler(void) {
  DEBUG("Installing exception handler\n");
  cyg_exception_set_handler( CYGNUM_HAL_EXCEPTION_DATA_ACCESS,
			     cyg_memprot_exception_handler, 0, NULL, NULL );
}
