//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================

/**
 * ============================================================================
 * @file  memprot.c
 * @brief Memory protection package set-up functions
 *
 * Copyright (C) 2004,2005 Parrot S.A.
 *
 * @author     ivan.djelic@parrot.fr
 * @date       2004-05-04
 * ============================================================================
 */

#include <string.h>
#include <errno.h>

#include <cyg/kernel/kapi.h>
#include <cyg/hal/hal_intr.h>
#include <cyg/hal/hal_arch.h>
#include <cyg/infra/diag.h>
#include <cyg/infra/cyg_ass.h>

#include <pkgconf/hal.h>
#include <pkgconf/hal_arm.h>

#include <pkgconf/memprot.h>

#include "exception.h"
#include "mmu.h"
#include "vp_layout.h"
#include "diag.h"
#include "memalloc.h"

// get memory layout
#include <pkgconf/system.h>
#include CYGHWR_MEMORY_LAYOUT_H

#include "ram_alloc.h"
#include "mem_layout.h"
#include "memprot.h"

//! The virtual process memory layouts
cygmem_vp_layout_t cygmem_vp_layouts[16] = VP_LAYOUTS;

//! Our own MMU translation table
unsigned int *memprot_ttb = 0;

//! Index in per-thread data
int memprot_data_index;

//! Dumps a virtual process memory layout
void
memprot_dump_vp( cygmem_vp_layout_t * layout, int id ) {
  INFO( "------------- virtual process %02d ---------------\n", id );
  INFO( "  vm   : 0x%08x - 0x%08x\n", layout->vm_start, layout->vm_end );
  INFO( "  ro   : 0x%08x - 0x%08x ( rom : 0x%08x - 0x%08x )\n",
        layout->ro_start, layout->ro_end, layout->rom_ro_start,
        layout->rom_ro_end );
  INFO( "  rw   : 0x%08x - 0x%08x\n", layout->rw_start, layout->rw_end );
  INFO( "  data : 0x%08x - 0x%08x ( rom : 0x%08x - 0x%08x )\n",
        layout->data_start, layout->data_end, layout->rom_data_start,
        layout->rom_data_end );
  INFO( "  bss  : 0x%08x - 0x%08x\n", layout->bss_start, layout->bss_end );
  INFO( "  heap : 0x%08x - 0x%08x  init size = %08x\n",
        layout->heap_start, layout->heap_end,
        layout->heap_top-layout->heap_start );
  INFO( "stacks : 0x%08x - 0x%08x\n", layout->stacks_start,
        layout->stacks_end );
}

// ===========================================================================

/** Contiguous RAM allocator.
 *
 * This allocator is used at startup to guarantee that all data and bss
 * sections are mapped to contiguous physical RAM (which may be useful when
 * DMA transfers are used).
 */

static CYG_ADDRESS memprot_allocator_base;
static CYG_ADDRESS memprot_allocator_top;

#define IS_ALIGNED(_P,_S)  (((_P)&((_S)-1)) == 0 )
#define ALIGNED(_P,_S)     ((_P)&~((_S)-1))

void
memprot_allocator_init( CYG_ADDRESS base, CYG_ADDRESS top ) {
  memprot_allocator_base = ALIGNED( base + MMU_PAGE_SIZE-1, MMU_PAGE_SIZE );
  memprot_allocator_top = ALIGNED( top, MMU_PAGE_SIZE );
  CYG_ASSERT( memprot_allocator_base + MMU_PAGE_SIZE <= memprot_allocator_top,
              "RAM must contain at least a 4 KB page !" );
}

static CYG_ADDRESS
memprot_allocator_allocate(size_t size) {
  CYG_ASSERT(( size == MMU_SECTION_SIZE )||
             ( size == MMU_LARGE_PAGE_SIZE )||
             ( size == MMU_PAGE_SIZE ), "Bad allocation size !" );
  // allocate ascending from memory base
  if ( IS_ALIGNED( memprot_allocator_base, size )&&
       ( memprot_allocator_base + size <= memprot_allocator_top )) {
    memprot_allocator_base += size;
    return memprot_allocator_base-size;
  }
  return 0;
}

static CYG_ADDRESS
memprot_allocator_allocate_pt(size_t size) {
  CYG_ASSERT( size == MMU_PAGE_SIZE, "Bad allocation size !" );
  // allocate descending from memory top
  if ( memprot_allocator_top - size >= memprot_allocator_base ) {
    memprot_allocator_top -= size;
    return memprot_allocator_top;
  }
  return 0;
}

static void
memprot_allocator_free( CYG_ADDRESS addr, size_t size ) {
  CYG_FAIL( "Contiguous RAM allocator does not support deallocation !" );
}

void memprot_allocator_get_limits(CYG_ADDRESS *base, CYG_ADDRESS *top) {
  *base = memprot_allocator_base;
  *top = memprot_allocator_top;
}

// ===========================================================================

//! Map an address range of virtual memory (without memory allocation).
void
memprot_map( CYG_ADDRESS va, CYG_ADDRESS pa, size_t size, int prot, int cache,
             int domain ) {
  int status;
  status = memprot_mmu_map(va,pa,size,prot,cache,domain,MMU_NO_ALLOCATE,0);
  CYG_ASSERT(( status == ENOERR ), "memory map error" );
}

//! Map and allocate an address range of virtual memory.
void
memprot_map_alloc( CYG_ADDRESS va,size_t size,int prot, int cache,int domain) {
  int status;
  status = memprot_mmu_map(va, 0, size, prot, cache, domain, MMU_ALLOCATE, 0 );
  CYG_ASSERT(( status == ENOERR ), "memory map error" );
}

#ifdef CYGOPT_MEMPROT_RW_CODE
// make code sections writable to allow breakpoints in GDB
#  define AP_SYSTEM_CODE (AP_RW_RO)
#  define AP_USER_CODE   (AP_RW_RW)
#else
// normal access rights
#  define AP_SYSTEM_CODE (AP_RO_RO)
#  define AP_USER_CODE   (AP_RO_RO)
#endif

//! Convert RAM virtual address to RAM physical address
#define V2P(_vaddr)      ((_vaddr)-CYGMEM_REGION_ram+CYGMEM_REGION_ram_PHYS)

/** Memory protection setup.
 *
 * This function is called at startup (after kernel initialisation and before
 * cyg_user_start() to setup the memory map, install handlers, create mempools,
 * etc.
 */
void memprot_setup(void) {
  int i;
  CYG_ADDRESS ram_base, ram_top;
  cygmem_vp_layout_t *layout = &cygmem_vp_layouts[0];

  DEBUG("memory protection setup\n");

  // get a slot in thread data for thread management
  memprot_data_index = cyg_thread_new_data_index();

  // install exception handlers
  cyg_memprot_install_swi_handler();
  cyg_memprot_install_exception_handler();

  ram_base = V2P((CYG_ADDRESS)&__free_ram_start);
  ram_top = V2P(CYGMEM_REGION_ram + CYGMEM_REGION_ram_SIZE);

  // map our own TTB at the end of RAM
  memprot_ttb = (unsigned int *)(ram_top - 0x4000);
  memset( memprot_ttb, 0, 0x4000);

  // use our own contiguous RAM allocator
  memprot_allocator_init( ram_base, ram_top-0x4000 );
  memprot_mmu_configure( memprot_ttb,
                         memprot_allocator_allocate,
                         memprot_allocator_free,
                         memprot_allocator_allocate_pt,
                         memprot_allocator_free );

  // we now build our own TTB
  // eCos normally maps RAM at address 0 (except for P5dev board)

  // we don't have to map the rom_vectors section, it has been copied already

  // map eCos RO sections (cacheable and bufferable)
  memprot_map( layout->ro_start, V2P(layout->ro_start),
               layout->ro_end-layout->ro_start, AP_SYSTEM_CODE, 1, 0 );

  // map eCos RW data (cacheable and bufferable)
  memprot_map( layout->rw_start, V2P(layout->rw_start),
               layout->rw_end-layout->rw_start, AP_RW_RO, 1, 0 );

  // allocate and map eCos virtual heap
  //mmu_map( layout->heap_start, 0, layout->heap_top-layout->heap_start,
  //	   AP_RW_RO, 0, 0, MMU_ALLOCATE );

  // platform specific mapping (VSRs, FLASH, I/O, etc.)
#ifdef CYGPKG_HAL_ARM_ARM9_VPB926EJS
  memprot_vpb926ejs_setup( ram_base, ram_top );
#endif

#ifdef CYGPKG_HAL_ARM_ARM9_SMDK2410
  memprot_smdk2410_setup();
#endif

#ifdef CYGPKG_HAL_ARM_ARM9_P5DEV
  memprot_p5dev_setup();
#endif

  // now we may switch to our own TTB
  CYG_ASSERT( memprot_ttb, "Memory protection TTB was not initialized !");

  // Before we switch to our own TTB, we flush the data cache. This is required
  // for at least 2 reasons:
  //
  // 1) Even if dirty lines in the cache point to areas which we will mapped at
  // the same addresses in both tables, if C/B bits change then data will not
  // be flushed later (e.g. ARM 920T).
  // 2) Our own TTB may itself reside on a cached area before the switch, so it
  // needs to be committed.
  //
  // The code executed *after* the flush and *before* the switch should access
  // memory which is mapped identically (same addresses, same C/B bits).
  // Here we use flat-mapped eCos code/stack/data/bss so it should be OK.

  FLUSH_CACHE();

  memprot_mmu_switch_ttb(memprot_ttb);
  // setup DACR with client access for all domains
  memprot_mmu_set_DACR( 0x55555555, CYGOPT_MEMPROT_DACR_MASK);

  // setup virtual processes
  for ( i = 1; i < 16; i++ ) {
    layout = &cygmem_vp_layouts[i];
    if ( layout->ro_start < layout->ro_end ) {
      // This virtual process is not empty (i.e. has code)
      //memprot_dump_vp( layout, i );
      // map read-only sections
      memprot_map( layout->ro_start, V2P(layout->rom_ro_start),
                   layout->rom_ro_end - layout->rom_ro_start,AP_USER_CODE,1,i);

      // allocate and map data and bss sections
      if ( layout->rw_start < layout->rw_end ) {
        // if we use CYG_STARTUP == RAM, we can avoid relocation if LMA
        // addresses of .dataXX sections are MMU aligned
        // map .data
        if ( layout->rom_data_end > layout->rom_data_start ) {
          memprot_map( layout->data_start, V2P(layout->rom_data_start),
                       layout->data_end-layout->data_start, AP_RW_RW, 1, i );
        }
        // map and allocate .bss
        if ( layout->bss_end > layout->bss_start )
          memprot_map_alloc( layout->bss_start,
                             layout->bss_end-layout->bss_start, AP_RW_RW, 1,i);

        // relocate (copy) data if required
        /*
          memprot_map_alloc( layout->rw_start, layout->rw_end-layout->rw_start,
          AP_RW_RW, 1, i );
          // relocate (copy) data if required
          if ( layout->rom_data_start < layout->rom_data_end )
          memcpy( (void *)layout->data_start, (void *)layout->rom_data_start,
		  layout->rom_data_end-layout->rom_data_start );
        */

        // zero bss section
        if ( layout->bss_start < layout->bss_end )
          memset( (void *)layout->bss_start, 0,
                  layout->bss_end-layout->bss_start );
      }
    }
  }
  //memprot_mmu_dump_memory_map();

  // switch to regular MMU RAM allocator
  memprot_allocator_get_limits( &ram_base, &ram_top );
  ram_init( ram_base, ram_top );
  memprot_mmu_configure( memprot_ttb, 0, 0, 0, 0 );

  // create memory pools
#ifdef CYGOPT_MEMPROT_USER_MALLOC
  cyg_memprot_create_mempools();
#endif
}

//! This function overrides cyg_package_start()
void cyg_package_start(void) {
#ifdef CYGSEM_START_UITRON_COMPATIBILITY
  cyg_uitron_start();
#endif
  memprot_setup();
}
