//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================

/**
*******************************************************************************
* @file  zcip_prot.c
* @brief ZeroConf protocol implementation
*            RFC3927 ZeroConf IPv4 Link-Local addressing
*
* Copyright (C) 2008 Parrot S.A.
*
* @author     ana.donezar@parrot.com
* @date       2008-01-22
*******************************************************************************
*/

#include <zcip.h>

#include <pkgconf/system.h>
#include <pkgconf/net.h>

/* #include <sys/param.h> */

#include <network.h>
#include <errno.h>

#include <cyg/infra/cyg_ass.h>

/* #ifdef INET6 */
/* #include <net/if_var.h> */
/* #include <netinet6/in6_var.h> */
/* #endif */

/* #include <sys/queue.h> */
/* #include <sys/sysctl.h> */
/* #include <sys/socket.h> */

/* #include <net/if.h> */
/* #include <net/if_dl.h> */
/* #include <net/if_types.h> */
/* #include <net/route.h> */
/* #include <net/netisr.h> */
/* #include <net/if_llc.h> */
/* #ifdef BRIDGE */
/* #include <net/ethernet.h> */
/* #include <net/bridge.h> */
/* #endif */

#include <netinet/in.h>
#include <netinet/if_ether.h>

#include <stdlib.h>
#include <net/iso88025.h>

#include <stdio.h>
#include <cyg/infra/diag.h>
#include <cyg/utils/clock.h>


/** Table to store all known interfaces */
struct zcip_ifentry ZcipIfTable[ZCIP_IFTABLE_MAXENTRIES];

/** Socket used by the whole zconf procedure. */
int ZcipSock;

/** Structure used internally when resolving arp conflicts */
struct bootp ZcipBootp;

/** 
 * Pick a random IP address on 169.254/16
 * excluding 169.254.0.x and 169.254.255.x
 * 
 * @param ip Pointer to chosen ip address - return value
 */
static void pick_ip(struct in_addr *ip)
{
	unsigned tmp;
	unsigned int sec, usec;
	clock_us(&sec, &usec);
	do {
		tmp = rand_r(&usec) & IN_CLASSB_HOST;
	} while (tmp > (IN_CLASSB_HOST - 0x0200));

	//(tmp + 0x0100) being 0xXXYY completes the address 169.254.x.y
	ip->s_addr = htonl(INADDR_ZCIP_GROUP + 0x0100 + tmp);
}

/** 
 * Store a known interface in the table, 
 * to be able to retrieve intf name for ip or mac addresses.
 * 
 * @param intf Interface name (exemple "wl0")
 * @param ip protocol address
 * @param mac physical address
 */
static void store_if(const char * intf, struct in_addr ip, struct ether_addr mac)
{
	static int ifentry = 0;

	// store the entry for further use
	ifentry = ifentry % ZCIP_IFTABLE_MAXENTRIES;
	strncpy(ZcipIfTable[ifentry].intf, intf, ZCIP_INTF_MAXLEN - 1);
	ZcipIfTable[ifentry].intf[ZCIP_INTF_MAXLEN - 1] = '\0';
	memcpy(&ZcipIfTable[ifentry].mac_addr, &mac, ETHER_ADDR_LEN);
	memcpy(&ZcipIfTable[ifentry].ip_addr, &ip, IP_ADDR_LEN);

	return;
}

/**  
 * Retrieve local Hardware address
 *
 * @param  intf Interface name (exemple "wl0")
 * @param mac pointer to physical address - return value
 * @retval 0 on success, -1 or errno on failure
 */
static int get_hwaddr(const char *intf, 	struct ether_addr *mac)
{
	struct ifreq ifr;

	// get the interface's ethernet address
	if ((ZcipSock = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		//perror("socket");
		return errno;
	}
	strncpy(ifr.ifr_name, intf, sizeof(ifr.ifr_name) - 1);
	ifr.ifr_name[sizeof(ifr.ifr_name) - 1] = '\0';
	if ( ioctl(ZcipSock, SIOCGIFHWADDR, &ifr) != 0) {
		perror("MAC address");
		close(ZcipSock);
		return -1;
	}
	memcpy(mac, &ifr.ifr_hwaddr.sa_data, ETHER_ADDR_LEN);

	close(ZcipSock);
	return 0;
}

/**
 * Retrieve local IP address
 *
 * @param intf Interface name (exemple "wl0")
 * @param ip pointer to ip address - return value
 * @retval 0 on success, -1 or errno on failure
 */
static int get_ipaddr(const char *intf, 	struct in_addr *ip)
{
	struct ifreq ifr;
	struct sockaddr_in *in_addr;

	// get the interface's ethernet address
	if ((ZcipSock = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		//perror("socket");
		return errno;
	}
	strncpy(ifr.ifr_name, intf, sizeof(ifr.ifr_name) - 1);
	ifr.ifr_name[sizeof(ifr.ifr_name) - 1] = '\0';
	if ( ioctl(ZcipSock, SIOCGIFADDR, &ifr) != 0) {
		perror("IP address");
		close(ZcipSock);
		return -1;
	}
	in_addr = (struct sockaddr_in *)&(ifr.ifr_addr);
	memcpy(ip, &in_addr->sin_addr, IP_ADDR_LEN);
	close(ZcipSock);
	return 0;
}

/**
 * Put the interface in up state if needed and set it 
 * in promiscuous mode. Done at the begining of do_zcip
 * 
 * @param  intf Interface name (exemple "wl0")
 * @param ip ip address to use on init
 */
static void arp_interface_up(const char *intf, struct in_addr ip)
{
	struct ifreq ifr;
	struct bootp *tmp = &ZcipBootp;
	int s;

	s = socket(AF_INET, SOCK_DGRAM, 0);
	if (s < 0) {
		perror("socket");
		return;
	}
	strcpy(ifr.ifr_name, intf);

	// Init interface if needed
	if (ioctl(s, SIOCGIFFLAGS , &ifr)) {
		perror("SIOCGIFFLAGS");
	}
	if ( ( ifr.ifr_flags & (IFF_UP | IFF_RUNNING) ) !=  (IFF_UP | IFF_RUNNING) ) {
		build_bootp_record(tmp,intf,inet_ntoa(ip),
						   "255.255.0.0",
						   "169.254.255.255",
						   "0.0.0.0",
						   "0.0.0.0");
		if (!init_net(intf, tmp)){
			NETWORK_PRINTF("Network initialization failed for %s\n", intf);
			close(s);
			return;
		}
	}

	// Set "running in promisc mode"
	ifr.ifr_flags = IFF_UP | IFF_RUNNING | IFF_PROMISC;
	if (ioctl(s, SIOCSIFFLAGS, &ifr)) {
		perror("SIOCSIFFLAGS");
	}
	close(s);
}

/**
 * Put the interface in down state. To be done before leaving do_zcip.
 * If not, a dummy interface open on a fake ip address remains open.
 *
 * @param  intf Interface name (exemple "wl0")
 */
static void arp_interface_down(const char *intf)
{
	struct ifreq ifr;
	int s;

	cyg_route_reinit();  // Force any existing routes to be forgotten

	s = socket(AF_INET, SOCK_DGRAM, 0);
	if (s < 0) {
		perror("socket");
		return;
	}

	strcpy(ifr.ifr_name, intf);
	if (ioctl(s, SIOCDIFADDR, &ifr)) { /* delete IF addr */
		perror("SIOCDIFADDR1");
	}

	// Shut down interface so it can be reinitialized
	ifr.ifr_flags &= ~(IFF_UP | IFF_RUNNING);
	if (ioctl(s, SIOCSIFFLAGS, &ifr)) { /* set ifnet flags */
		perror("SIOCSIFFLAGS down");
	}

	close(s);
}

/**
 * Open socket for arp exchanges
 *
 * @param  intf Interface name (exemple "wl0")
 * @retval 0 on succes, -1 on failure
 */
int open_arp(const char *intf)
{
	char tmp[16];
	int err;
	struct sockaddr_in addr;

    //Init the variable of this struct
    memset(&addr, 0, sizeof(addr));

	ZcipSock = socket(AF_RIF, SOCK_RAW,0);
	if (ZcipSock < 0) {
		return -1;
	}

	bzero(tmp,16);
	sprintf(tmp, "%s:%d", intf, 0);
	err = bind(ZcipSock, (struct sockaddr *)tmp, 0 );
	if (err < 0) {
		close(ZcipSock);
		return -1;
	}
	
	// we must be connected
	// since RIF doesn't care about the address
	// fake a connect so the stack feels good
	err = connect(ZcipSock, (struct sockaddr *)&addr, sizeof(addr));
	if (err < 0) {
		close(ZcipSock);
		return -1;
	}
    return 0;
}

/**
 * Send an ARP packet.
 * 
* @param op operation (request or reply)
* @param smac source hardware address
* @param sip source ip address
* @param tmac target hardware address
* @param tip target ip address
 */
void send_arp(short op,
	struct ether_addr *smac, struct in_addr sip,
	struct ether_addr *tmac, struct in_addr tip)
{
	unsigned char p[ARP_PACKET_LEN];
	struct ether_header *h = (struct ether_header *)p;
	struct ether_arp *a = (struct ether_arp *)(p+ETHER_HDR_LEN) ;
	int nr;

	memset(p, 0, ARP_PACKET_LEN );

	// ethernet header
	h->ether_type = htons(ETHERTYPE_ARP);
	memcpy(h->ether_shost, smac, ETHER_ADDR_LEN);
	memcpy(h->ether_dhost, tmac, ETHER_ADDR_LEN);

	// arp packet
	a->arp_hrd = htons(ARPHRD_ETHER);
	a->arp_pro = htons(ETHERTYPE_IP);
	a->arp_hln = ETHER_ADDR_LEN;
	a->arp_pln = IP_ADDR_LEN;
	a->arp_op = htons(op);
	memcpy(a->arp_sha, smac, ETHER_ADDR_LEN);
	memcpy(a->arp_spa, &sip, IP_ADDR_LEN);
	memcpy(a->arp_tha, tmac, ETHER_ADDR_LEN);
	memcpy(a->arp_tpa, &tip, IP_ADDR_LEN);

	// send packet down
	nr = write(ZcipSock, &p, ARP_PACKET_LEN);

	if (nr != ARP_PACKET_LEN) {
		perror("socket write");
	}
	return;
}

/**
 * Wait for an ARP packet.
 * 
 * @param p Buffer to fill with received arp packet
 * @param plen Lenght of buffer
 * @retval lenght read on success, -1 on error
 */
int read_arp(unsigned char *p, int plen)
{
	int length;
/* 	fd_set fdvar; */
/* 	struct timeval timeout; */

	if ( ( p == NULL ) || (plen == 0) )
		return -1;

/* 	timeout.tv_sec = 1; */
/* 	timeout.tv_usec = 0; */

/* 	FD_ZERO(&fdvar); */
/* 	FD_SET(sock, &fdvar); */
/* 	if(select(sock + 1, &fdvar, NULL, NULL, &timeout) < 0){ */
/* 		perror("l2 select recv failed\n"); */
/* 		return -1; */
/* 	}  */
/* 	if (FD_ISSET(sock, &fdvar)) { */
	length = read(ZcipSock, p, plen) ;
	if ( ( length == -1) && (errno ==  ETIMEDOUT) )
		length = 0;
/* 	} */

	return length;
}


/**
 * the Zeroconf state machine - this does all the work
 * FIXME: timing between different announcements
 * 
 * @param interface May be a string of type "wl0" or a IP address
 *                            when we want to resolve a conflict
 * @param res If NULL, the interface will be bring up before getting out
 *                   else, the user should init interface using "res"
 * @param defend Tells if the zcip procedure should start 
 *                         from the begining or just  try to keep an ip.
 * @retval 0 on success, -1 on error
 */
int
do_zcip(const char *interface, struct bootp *res, bool defend)
{
	int state = PROBE;
	const char *intf = NULL;
	static bool running = false;
	struct timeval tv;
	int len = 0, conflicts = 0, probes = 0, claims = 0;

	static char bdcast_addr[6] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff};
	struct ether_addr mymac;
	struct in_addr myip;
	struct in_addr null_ip;
	memset(&null_ip, 0, sizeof(null_ip));

	if (running){
		NETWORK_PRINTF("do_zcip: Error - already running\n");
		return -1;
	}
	running = true;

	if (defend){
		bool found = false;
		int i = 0;
		while ( (i< ZCIP_IFTABLE_MAXENTRIES) && (!found) ){
			if ( (memcmp(interface, &ZcipIfTable[i].ip_addr, IP_ADDR_LEN) == 0) ||
				 (memcmp(interface, &ZcipIfTable[i].mac_addr, ETHER_ADDR_LEN) == 0) ){
				intf = ZcipIfTable[i].intf;
				found = true;
			}
			i++;
		}
		if (!found){
			NETWORK_PRINTF("do_zcip: Error - unknown interface\n");
			return -1;
		}
	}
	else{
		intf = interface;
	}


	if (defend){
		// Get ip addr
		get_ipaddr(intf, &myip);
		// Start directly on the announce state
		state = ANNOUNCE;
	}
	else{
		// Pick a random ip to start with
		pick_ip(&myip);
#ifdef CYGDBG_NET_ZCIP_CHATTER
		diag_printf("do_zcip: %s try to pick %s\n", intf, inet_ntoa(myip));
#endif
	}
	get_hwaddr(intf, &mymac);

	// Init interface in "arp mode"
	arp_interface_up(intf, myip);

	// open and bind ARP socket
	open_arp(intf);

	// set recv timeout
	tv.tv_sec = 1;
	tv.tv_usec = 0;
	setsockopt(ZcipSock, SOL_SOCKET, SO_RCVTIMEO, &tv, sizeof(tv));

	while (1) {
		unsigned char p[ARP_PACKET_LEN];
		struct ether_header *h = (struct ether_header *)p;
		struct ether_arp *a = (struct ether_arp *)(p+ETHER_HDR_LEN) ;
		int ipconflict;
		int macconflict;

#ifdef CYGDBG_NET_ZCIP_CHATTER
		diag_printf("do_zcip: %s (state %d) probes=%u, claims=%u \n",
					intf, state, probes, claims);
#endif
		
		// timeout on socket, no data received (good news)
		if ( len ==  0) {
			switch (state) {
			case PROBE:
				// No conflicting ARP packets received 
				if (probes < PROBE_NUM) {
#ifdef CYGDBG_NET_ZCIP_CHATTER
					diag_printf("do_zcip: %s probe=%u on %s\n", intf, probes, inet_ntoa(myip));
#endif
					send_arp(ARPOP_REQUEST,&mymac, null_ip,
							 (struct ether_addr *)&bdcast_addr, myip);
					probes++;
				}
				else {
					// Move to announce state.
					state = ANNOUNCE;
#ifdef CYGDBG_NET_ZCIP_CHATTER
					diag_printf("do_zcip: %s announce=%u on %s\n", intf, claims, inet_ntoa(myip));
#endif
					send_arp(ARPOP_REQUEST, &mymac, myip, 
							 (struct ether_addr *)&bdcast_addr, myip);
					claims = 0;
				}
				break;
			case RATE_LIMIT_PROBE:
				// After many conflicts, probes have been disables, switch directly to announce state
				state = ANNOUNCE;
				claims = 0;
#ifdef CYGDBG_NET_ZCIP_CHATTER
				diag_printf("do_zcip: %s announce=%u on %s\n", intf, claims, inet_ntoa(myip));
#endif
				send_arp(ARPOP_REQUEST, &mymac, myip,
						 (struct ether_addr *)&bdcast_addr, myip);
				break;
			case ANNOUNCE:
				// No conflicting ARP packets received 
				if (claims < ANNOUNCE_NUM) {
#ifdef CYGDBG_NET_ZCIP_CHATTER
					diag_printf("do_zcip: %s announce=%u on %s\n", intf, claims, inet_ntoa(myip));
#endif
					send_arp(ARPOP_REQUEST, &mymac, myip, 
							 (struct ether_addr *)&bdcast_addr, myip);
					claims++;
				}
				else {
					// Move to monitor state.
					state = MONITOR;
					if ( !res ){
						struct bootp *tmp = &ZcipBootp;
						build_bootp_record( tmp,
										   intf,
										   inet_ntoa(myip),
										   "255.255.0.0",
										   "169.254.255.255",
										   "0.0.0.0",
										   "0.0.0.0");
						arp_interface_down(intf);
						show_bootp(intf, tmp);
						init_net(intf, tmp);
					}
					else{
						build_bootp_record( res,
										   intf,
										   inet_ntoa(myip),
										   "255.255.0.0",
										   "169.254.255.255",
										   "0.0.0.0",
										   "0.0.0.0");
						arp_interface_down(intf);
					}
					conflicts = 0;
					close(ZcipSock);
					store_if(intf, myip, mymac);
					running = false;
					return 0;
				}
				break;
			case DEFEND:
				// we have managed to keep our IP after a conflict. Go back to monitor.
				state = MONITOR;
				conflicts = 0;
				break;
			default:
#ifdef CYGDBG_NET_ZCIP_CHATTER
				diag_printf("do_zcip: %s invalid state(%d) \n", intf, state);
#endif
				// Bad behaviour - restart the whole thing
				state = PROBE;
				pick_ip(&myip);
				send_arp(ARPOP_REQUEST, &mymac, null_ip, 
						 (struct ether_addr *)&bdcast_addr, myip);
				probes = 0;
				claims = 0;
				break;
			} // switch (state)
		}
		// packets arriving
		else if ( len >0 ) {
			if (h->ether_type != htons(ETHERTYPE_ARP)){
				len = 0;
				continue;
			}
#ifdef CYGDBG_NET_ZCIP_CHATTER
			{
				unsigned char * sha = a->arp_sha;
				unsigned char * tha = a->arp_tha;
				struct in_addr spa;
				struct in_addr tpa;
				memcpy(&spa, a->arp_spa, sizeof(struct in_addr) );
				memcpy(&tpa, a->arp_tpa, sizeof(struct in_addr) );
				diag_printf("do_zcip: %s recv arp op=%d,\n",
					intf, ntohs(a->arp_op));
				diag_printf("\tsource=%02X:%02X:%02X:%02X:%02X:%02X ip=%s\n",
					sha[0],sha[1],sha[2],sha[3],sha[4],sha[5],
							inet_ntoa( spa ));
				diag_printf("\ttarget=%02X:%02X:%02X:%02X:%02X:%02X ip=%s\n",
					tha[0],tha[1],tha[2],tha[3],tha[4],tha[5],
					inet_ntoa( tpa ));
			}
#endif
			if (a->arp_op != htons(ARPOP_REQUEST)
				&& a->arp_op != htons(ARPOP_REPLY))
				continue;

			// Received packet from host using our ip address
			ipconflict = 0;
			// Received packet sent to my ip address with a mac address different than mine
			macconflict = 0; 

			if (memcmp(a->arp_spa, &myip.s_addr, sizeof(struct in_addr)) == 0
				&& memcmp(&a->arp_sha, &mymac, ETHER_ADDR_LEN) != 0
				) {
				ipconflict = 1;
			}
			if (a->arp_op == htons(ARPOP_REQUEST)
				&& memcmp(a->arp_tpa, &myip.s_addr, sizeof(struct in_addr)) == 0
				&& memcmp(&a->arp_tha, &mymac, ETHER_ADDR_LEN) != 0
				) {
				macconflict = 1;
			}
#ifdef CYGDBG_NET_ZCIP_CHATTER
			diag_printf("do_zcip: state (%d) ip conflict = %d, mac conflict = %d\n",
								state, ipconflict, macconflict);
#endif
			switch (state) {
			case PROBE:
			case ANNOUNCE:
				// Check for ip conflicts or incorrect ARP probes (mac conflicts)
				if (ipconflict || macconflict) {
					conflicts++;
					if (conflicts >= MAX_CONFLICTS) {
#ifdef CYGDBG_NET_ZCIP_CHATTER
						diag_printf("do_zcip: %s ratelimit\n", intf);
#endif
						state = RATE_LIMIT_PROBE;
					}

					// move out and restart protocol with a new ip
					pick_ip(&myip);
#ifdef CYGDBG_NET_ZCIP_CHATTER
					diag_printf("do_zcip: %s try to pick %s\n", intf, inet_ntoa(myip));
#endif
					send_arp(ARPOP_REQUEST, &mymac, null_ip, 
							 (struct ether_addr *)&bdcast_addr, myip);
					probes = 0;
					claims = 0;
				}
				break;
			case MONITOR:
				// If there is a conflict, try to defend with an ARP probe.
				if (ipconflict) {
#ifdef CYGDBG_NET_ZCIP_CHATTER
					diag_printf("do_zcip: %s monitor conflict -- defending\n", intf);
#endif
					state = DEFEND;
					send_arp(ARPOP_REQUEST, &mymac, myip, &mymac, myip);
				}
				break;
			case DEFEND:
				// Conflict remains, restart the protocol.
				if (ipconflict) {
					state = PROBE;
#ifdef CYGDBG_NET_ZCIP_CHATTER
					diag_printf("do_zcip: %s defend conflict -- restarting\n", intf);
#endif
					// move out and restart protocol with a new ip
					pick_ip(&myip);
					send_arp(ARPOP_REQUEST, &mymac, null_ip, 
							 (struct ether_addr *)&bdcast_addr, myip);
					probes = 0;
					claims = 0;
				}
				break;
			default:
#ifdef CYGDBG_NET_ZCIP_CHATTER
				diag_printf("do_zcip: %s invalid state(%d) \n", intf, state);
#endif
				// Bad behaviour - restart the whole thing
				state = PROBE;
				pick_ip(&myip);
				send_arp(ARPOP_REQUEST, &mymac, null_ip, 
						 (struct ether_addr *)&bdcast_addr, myip);
				probes = 0;
				claims = 0;
				break;
			} // switch state
		}
		// link went down or ARP socket was externally closed
		else {
			perror("zcip socket");
			close(ZcipSock);
			arp_interface_down(intf);
			running = false;
			return -1;
		}
		len = read_arp(p, sizeof(p)) ;
	} // while (1)
	close(ZcipSock);
	arp_interface_down(intf);
	running = false;
	return -1;
}
