/**
*******************************************************************************
* @file  common.h
* @brief Dynamo private functions and macros
*
* Copyright (C) 2007 Parrot S.A.
*
* @author     ivan.djelic@parrot.fr
* @date       2007-03-01
*******************************************************************************
*/
#ifndef CYGONCE_DYNAMO_DYNAMO_H
#define CYGONCE_DYNAMO_DYNAMO_H

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include <cyg/kernel/kapi.h>
#include <cyg/infra/diag.h>
#include <pkgconf/dynamo.h>
#include <string.h>

#include <cyg/utils/utils.h>
#include <cyg/dynamo/process.h>
#include <cyg/hal/hal_mmu.h>

#include "ram_alloc.h"
#include "vm_alloc.h"
#include "dlmallocuser.h"

#define DYNAMO_MAX_THREADS          (32)
#define DYNAMO_MAX_MUTEXES          (64)

// Default segment sizes
#define DYNAMO_PROC_RO_SIZE         (32 << 20) // 32 Mbytes
#define DYNAMO_PROC_RW_SIZE         (32 << 20) // 32 Mbytes
#define DYNAMO_PROC_HEAP_SIZE       (32 << 20) // 32 Mbytes

//! This structure contains private information for user threads
typedef struct dynamo_thread_data {
    CYG_ADDRESS swi_stack[8];                     //!< system call workspace
    cyg_uint32 dacr;                              //!< user-mode DACR
    cyg_thread_entry_t *entry;                    //!< thread entry point
    cyg_process *process;                         //!< parent process
    CYG_ADDRESS ustack_base;                      //!< user stack base
    CYG_ADDRESS ustack_top;                       //!< user stack top
    cyg_uint8 kstack[CYGOPT_DYNAMO_KSTACK_SIZE];  //!< kernel stack
} dynamo_thread_data_t;

typedef struct dynamo_thread {
    cyg_thread obj;
    dynamo_thread_data_t data;
} dynamo_thread_t;

//! A global lock for virtual memory access
extern cyg_mutex_t vm_mutex;

void dynamo_init(void);
void dynamo_hal_setup(void);
void dlfcn_init(void);

int system_copy(void *dst, const void *src, unsigned int size);

// exception handling
void dynamo_install_exception_handlers(void);

// Process management
void cyg_process_init(void);

CYG_ADDRESS cyg_process_sbrk( cyg_process *proc, ptrdiff_t incr);

void * cyg_process_malloc( size_t size );
void cyg_process_free( void *ptr );
void * cyg_process_calloc( size_t nmemb, size_t size );
void * cyg_process_realloc( void *ptr, size_t size );

cyg_process *cyg_process_self(void);
cyg_process *cyg_process_get(cyg_thread *this);
Cyg_ErrNo cyg_process_map( cyg_process *proc, CYG_ADDRESS vaddr,size_t size );
Cyg_ErrNo cyg_process_unmap( cyg_process *proc, CYG_ADDRESS vaddr,size_t size);
Cyg_ErrNo cyg_process_add_thread( cyg_process         *proc,
                                  cyg_addrword_t      sched_info,
                                  cyg_thread_entry_t  *entry,
                                  cyg_addrword_t      entry_data,
                                  char                *name,
                                  void                *stack_base,
                                  cyg_ucount32        stack_size,
                                  cyg_handle_t        *handle,
                                  cyg_thread          *thread );
void cyg_process_entry(cyg_addrword_t data);
void cyg_process_thread_entry(cyg_addrword_t entry_data);
void cyg_process_mem_lock(cyg_process *proc, CYG_ADDRESS vaddr, size_t size );
void cyg_process_mem_unlock(cyg_process *proc, CYG_ADDRESS vaddr, size_t size);

// system call check
enum {
    CHECK_RO,
    CHECK_RW,
};

int cyg_process_check_access(cyg_process *proc,
                             CYG_ADDRESS vaddr, size_t size, int mode);
int cyg_process_range_get(CYG_ADDRESS vaddr, size_t size, int mode);
void cyg_process_range_release(CYG_ADDRESS vaddr, size_t size);
void cyg_process_imb(void *base, unsigned int size);

#define RANGE_GET_RO(_addr,_size) \
  cyg_process_range_get((_addr), (_size), CHECK_RO)

#define RANGE_GET_RW(_addr,_size) \
  cyg_process_range_get((_addr), (_size), CHECK_RW)

#define RANGE_RELEASE(_addr,_size) \
  cyg_process_range_release((_addr), (_size))


#define IS_ALIGNED(_P,_S)  (((_P)&((_S)-1)) == 0 )
#define ALIGN_BASE(_P,_S)  ((_P)&~((_S)-1))
#define ALIGN_TOP(_P,_S)   (((_P)+(_S)-1)&~((_S)-1))

#define PAGE_CEIL(_P)      ALIGN_TOP(_P,MMU_PAGE_SIZE)

//! A private userland structure to manage heap
struct process_heap {
    CYG_ADDRESS base;                      //! Arena base address
    CYG_ADDRESS end;                       //!
    CYG_ADDRESS top;                       //! User controlled level
    cyg_mempool_dlmalloc dlpool;
    cyg_handle_t pool;
    cyg_mutex_t mutex;
};

// A few macros for declaring system calls for internal use

#define SYSCALL(_call)         __swi_ ## _call
#define DECL_SYSCALL(_call)    extern __typeof__(_call) SYSCALL(_call)

#define SYSWRAP(_call)         __wrap_ ## _call
#define DECL_SYSWRAP(_call)    extern __typeof__(_call) SYSWRAP(_call)

// A few calls we need internally

DECL_SYSCALL(cyg_mutex_init);
DECL_SYSCALL(cyg_mutex_lock);
DECL_SYSCALL(cyg_mutex_unlock);
DECL_SYSCALL(cyg_thread_exit);
DECL_SYSCALL(cyg_process_map);
DECL_SYSCALL(cyg_process_unmap);
DECL_SYSCALL(cyg_process_imb);
DECL_SYSCALL(open);
DECL_SYSCALL(read);
DECL_SYSCALL(lseek);
DECL_SYSCALL(close);
DECL_SYSCALL(diag_write_string);

DECL_SYSWRAP(diag_printf);

// Debugging and diagnostic macros

#ifndef PFX
#define PFX "dynamo: "
#endif

#if ( CYGOPT_DYNAMO_VERBOSITY_LEVEL >= 3 )
#  define DEBUG2( _fmt_, args... )   diag_printf( PFX _fmt_, ##args )
#  define UDEBUG2( _fmt_, args... )  SYSWRAP(diag_printf)( PFX _fmt_, ##args )
#else
#  define DEBUG2( _fmt_, args... )   {}
#  define UDEBUG2( _fmt_, args... )  {}
#endif

#if ( CYGOPT_DYNAMO_VERBOSITY_LEVEL >= 2 )
#  define DEBUG1( _fmt_, args... )   diag_printf( PFX _fmt_, ##args )
#  define UDEBUG1( _fmt_, args... )  SYSWRAP(diag_printf)( PFX _fmt_, ##args )
#else
#  define DEBUG1( _fmt_, args... )   {}
#  define UDEBUG1( _fmt_, args... )  {}
#endif

#if ( CYGOPT_DYNAMO_VERBOSITY_LEVEL >= 1 )
#  define DIAG( _fmt_, args... )     diag_printf( PFX _fmt_, ##args )
#  define UDIAG( _fmt_, args... )    SYSWRAP(diag_printf)( PFX _fmt_, ##args )
#else
#  define DIAG( _fmt_, args... )     {}
#  define UDIAG( _fmt_, args... )    {}
#endif

#endif
