//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================

/**
*******************************************************************************
* @file  process.c
* @brief Virtual process management
*
* Copyright (C) 2007 Parrot S.A.
*
* @author     ivan.djelic@parrot.fr
* @date       2007-03-01
*******************************************************************************
*/

#include <cyg/kernel/kapi.h>
#include <dlfcn/dlfcn.h>
#include "common.h"

//! A global thread pool for all processes
static dynamo_thread_t dynamo_thread_pool[DYNAMO_MAX_THREADS];

//! A global mutex to protect access
static cyg_mutex_t process_mutex;

void
cyg_process_init(void) {
    int i;

    cyg_mutex_init(&process_mutex);

    // initialize thread pool
    for ( i = 0; i < DYNAMO_MAX_THREADS; i++ ) {
        dynamo_thread_pool[i].data.process = NULL;
    }
}


/** Create a process.
 *
 * This is can only be called from a privileged mode, and should not be a
 * system call.
 */
Cyg_ErrNo
cyg_process_create( cyg_addrword_t sched_info,
                    const char *name,
                    const char *file,
                    const char *entry,
                    cyg_uint32 domain,
                    cyg_uint32 max_threads,
                    cyg_uint32 max_mem,
                    cyg_handle_t *handle,
                    cyg_process *this ) {

    Cyg_ErrNo ret = ENOERR;

    cyg_mutex_init(&this->mutex);

    this->name = name;
    this->file = file;
    this->entry = entry;
    this->domain = (domain & 0xF);
    this->sched_info = sched_info;
    this->max_threads = max_threads;
    this->nb_threads = 0;
    this->max_mem = ALIGN_BASE( max_mem, MMU_PAGE_SIZE );
    this->used_mem = 0;
    this->heap_lock_count = 0;

    // allocate virtual memory
    cyg_mutex_lock(&vm_mutex);

    // ro segment
    this->ro.base = vm_allocate(DYNAMO_PROC_RO_SIZE);
    this->ro.size = DYNAMO_PROC_RO_SIZE;
    if ( !this->ro.base ) {
        ret = -ENOMEM;
    }
    // rw segment
    if ( ret == ENOERR ) {
        this->rw.base = vm_allocate(DYNAMO_PROC_RW_SIZE);
        this->rw.size = DYNAMO_PROC_RW_SIZE;
        if ( !this->rw.base ) {
            ret = -ENOMEM;
        }
    }
    // heap segment
    if ( ret == ENOERR ) {
        this->heap.base = vm_allocate(DYNAMO_PROC_HEAP_SIZE);
        this->heap.size = DYNAMO_PROC_HEAP_SIZE;
        if ( !this->heap.base ) {
            ret = -ENOMEM;
        }
    }
    cyg_mutex_unlock(&vm_mutex);

    // allocate a stack for the initial thread
    if ( ret == ENOERR ) {
        ret = cyg_process_map( this, this->rw.base, MMU_PAGE_SIZE );
    }

    // create initial thread
    if ( ret == ENOERR ) {

        DEBUG1( "creating process '%s'...\n", this->name );

        ret = cyg_process_add_thread( this,
                                      sched_info,
                                      cyg_process_entry,
                                      0,
                                      (char *)this->name,
                                      (void *)this->rw.base,
                                      MMU_PAGE_SIZE,
                                      handle,
                                      NULL );
    }

    return ret;
}

void
cyg_process_resume(cyg_handle_t process) {
    cyg_thread_resume(process);
}

/** Return current thread parent process if it has one.
 *
 */
cyg_process *
cyg_process_self(void) {
    cyg_process *ret = NULL;
    dynamo_thread_t *thread;
    CYG_ADDRESS self = (CYG_ADDRESS)cyg_thread_self();

    // verify that the current thread is in the thread pool
    if (( self >= (CYG_ADDRESS)&dynamo_thread_pool[0] )&&
        ( self <= (CYG_ADDRESS)&dynamo_thread_pool[DYNAMO_MAX_THREADS-1] )) {

        thread = (dynamo_thread_t *)self;
        ret = thread->data.process;
    }

    return ret;
}

/** Return a thread parent process if it has one.
 *
 */
cyg_process *
cyg_process_get(cyg_thread *this) {
    int index;
    cyg_process *ret = NULL;
    dynamo_thread_t *thread;
    CYG_ADDRESS self = (CYG_ADDRESS)this;

    // verify that this thread is in the thread pool
    if (( self >= (CYG_ADDRESS)&dynamo_thread_pool[0] )&&
        ( self <= (CYG_ADDRESS)&dynamo_thread_pool[DYNAMO_MAX_THREADS-1] )) {

        // check alignment
        index = (self-(CYG_ADDRESS)dynamo_thread_pool)/sizeof(dynamo_thread_t);
        if ( self == (CYG_ADDRESS)&dynamo_thread_pool[index] ) {
            thread = (dynamo_thread_t *)self;
            ret = thread->data.process;
        }
    }

    return ret;
}

/** Create a process heap.
 *
 * This is called from user mode.
 */
static Cyg_ErrNo
cyg_process_create_heap( cyg_process *proc ) {

    int ret;
    struct process_heap *heap;

    // map a single page for a start in heap segment
    ret = SYSCALL(cyg_process_map)( proc, proc->heap.base, MMU_PAGE_SIZE );

    // create a memory pool
    if ( ret == ENOERR ) {

        heap = (struct process_heap *)proc->heap.base;
        heap->base = proc->heap.base + sizeof(struct process_heap);
        heap->end = proc->heap.base + proc->heap.size;
        heap->top = heap->base;
        SYSCALL(cyg_mutex_init)(&heap->mutex);

        // now create mempool
        cyg_mempool_dlmalloc_create((void *)heap->base,
                                    heap->end-heap->base,
                                    &heap->pool,
                                    &heap->dlpool,
                                    1 /* allow sbrk() */,
                                    proc /* attach pool to this process */ );
    }
    return ret;
}


/** Heap resizing.
 *
 * This function is called from DL malloc routines to automatically
 * extend/trim the heap area. It can result in memory pages being mapped
 * in/out.
 *
 * @param incr How much memory should be added on top of heap (can be negative)
 * @return Previous heap top address
 */
CYG_ADDRESS
cyg_process_sbrk( cyg_process *proc, ptrdiff_t incr) {

    int ret = 0;
    struct process_heap *heap;
    CYG_ADDRESS new_top, old_top;

    // get process heap structure address
    heap = (struct process_heap *)proc->heap.base;

    // there should be no need to lock the heap mutex because we are called
    // directly from DL routines

    old_top = PAGE_CEIL(heap->top);
    // compute new heap top
    new_top = PAGE_CEIL(heap->top + incr);

    // check new heap sanity
    if (( heap->top+incr < heap->base )||( new_top > heap->end )) {

        UDIAG( "sbrk: bad incr=%d in process 0x%08x\n", incr,
               (unsigned int)proc );
        ret = -EINVAL;
    }

    // free physical ram if required
    if (( ret == ENOERR )&&( new_top < old_top )) {

        ret = SYSCALL(cyg_process_unmap)( proc, new_top, old_top - new_top );
        if ( ret != ENOERR ) {
            UDIAG( "sbrk: unable to unmap 0x%08x-0x%08x in process 0x%08x !\n",
                   new_top, old_top, (unsigned int)proc );
        }
    }

    // allocate more physical ram if required
    if (( ret == ENOERR )&&( old_top < new_top )) {

        ret = SYSCALL(cyg_process_map)( proc, old_top, new_top - old_top );
        if ( ret == -EINVAL ) {
            UDIAG( "sbrk: unable to map 0x%08x-0x%08x in process 0x%08x !\n",
                   old_top, new_top, (unsigned int)proc );
        }
    }

    if ( ret == ENOERR ) {
        old_top = heap->top;
        heap->top += incr;
        UDEBUG1( "sbrk(%d) in process 0x%08x\n", incr, (unsigned int )proc );
    }

    return ( ret == ENOERR )? old_top : (CYG_ADDRESS)-1;
}

//=============================================================================
// Memory allocation external API
//=============================================================================

//! User mode malloc
void *
cyg_process_malloc( size_t size ) {
    void *data_ptr = NULL;
    struct process_heap *heap;
    cyg_process *proc = cyg_process_self();

    if ( proc ) {
        // get process heap structure address
        heap = (struct process_heap *)proc->heap.base;

        SYSCALL(cyg_mutex_lock)(&heap->mutex);
        data_ptr = cyg_mempool_dlmalloc_try_alloc( heap->pool, size );
        SYSCALL(cyg_mutex_unlock)(&heap->mutex);
    }
    return data_ptr;
}

//! User mode free
void
cyg_process_free( void *ptr ) {
    struct process_heap *heap;
    cyg_process *proc = cyg_process_self();

    if ( ptr && proc ) {
        // get process heap structure address
        heap = (struct process_heap *)proc->heap.base;

        SYSCALL(cyg_mutex_lock)(&heap->mutex);
        cyg_mempool_dlmalloc_free( heap->pool, (cyg_uint8 *) ptr );
        SYSCALL(cyg_mutex_unlock)(&heap->mutex);
    }
}

//! User mode calloc
void *
cyg_process_calloc( size_t nmemb, size_t size ) {
    cyg_ucount32 realsize;
    void *data_ptr;

    realsize = nmemb*size;
    data_ptr = cyg_process_malloc(realsize);

    // Fill with 0's if non-NULL
    if ( data_ptr != NULL ) {
        memset( data_ptr, 0, realsize );
    }
    return data_ptr;
}

//! User mode realloc
void *
cyg_process_realloc( void *ptr, size_t size ) {
    cyg_int32 oldsize;
    void *newptr = NULL;
    struct process_heap *heap;
    cyg_process *proc;

    // if pointer is NULL, we must malloc it
    if ( ptr == NULL ) {
        return cyg_process_malloc(size);
    }
    // if size is 0, we must free it
    if ( size == 0 ) {
        cyg_process_free(ptr);
        return NULL;
    }
    proc = cyg_process_self();

    if ( proc ) {
        // get process heap structure address
        heap = (struct process_heap *)proc->heap.base;

        SYSCALL(cyg_mutex_lock)(&heap->mutex);
        // otherwise try to resize allocation
        newptr = cyg_mempool_dlmalloc_resize_alloc( heap->pool,
                                                    (cyg_uint8 *)ptr,
                                                    size, &oldsize);
        SYSCALL(cyg_mutex_unlock)(&heap->mutex);
    }
    if ( NULL == newptr ) {
        newptr = cyg_process_malloc(size);
        if ( NULL != newptr ) {
            memcpy( newptr, ptr,
                    size < (size_t) oldsize ? size : (size_t) oldsize );
            cyg_process_free(ptr);
        }
    }
    return newptr;
}

/** Process entry point.
 *
 * This function is called in user mode when a new process is created. It is in
 * charge of creating a heap, loading shared libraries and executing code.
 */
void
cyg_process_entry(cyg_addrword_t data) {
    Cyg_ErrNo err;
    int dlerr;
    void *handle;
    const char *entry_symbol;
    cyg_thread_entry *entry;
    cyg_process *this = cyg_process_self();

    if ( this ) {
        // first we need a heap
        err = cyg_process_create_heap(this);

        if ( err == ENOERR ) {

            UDEBUG1( "loading '%s' in process '%s'\n", this->file, this->name);

            handle = dlopen( this->file, 0, &dlerr );
            if ( handle ) {
                entry_symbol = ( this->entry )? this->entry : "main";
                entry = (cyg_thread_entry *)dlsym( handle, entry_symbol );
                if ( entry ) {
                    // execute code
                    (*entry)(data);
                }
                else {
                    UDEBUG1( "symbol '%s' could not be found\n", entry_symbol);
                }
                // cleanup
                dlclose(handle);
            }
            else {
                UDEBUG1( "error: %s while loading '%s'\n", dlerror(dlerr),
                         this->file );
            }
        }
    }
}

/** Lock a process memory range.
 *
 * This function must be called from system mode.
 * It locks a memory range to prevent it from being unmapped.
 */
void
cyg_process_mem_lock(cyg_process *proc, CYG_ADDRESS vaddr, size_t size ) {

    cyg_mutex_lock(&proc->mutex);

    // we only need to lock memory inside the heap segment, other segments
    // will not be modified until process is destroyed

    if (( vaddr+size > proc->heap.base )&&
        ( vaddr < proc->heap.base+proc->heap.size )) {
        // maybe we could use some finer page level locking ?
        proc->heap_lock_count++;
    }

    cyg_mutex_unlock(&proc->mutex);
}

/** Unlock a process memory range.
 *
 * This function must be called from system mode.
 * It unlocks a memory range.
 */
void
cyg_process_mem_unlock(cyg_process *proc, CYG_ADDRESS vaddr, size_t size ) {

    cyg_mutex_lock(&proc->mutex);

    // we only need to unlock memory inside the heap segment, other segments
    // will not be modified until process is destroyed

    if (( vaddr+size > proc->heap.base )&&
        ( vaddr < proc->heap.base+proc->heap.size )) {
        // maybe we could use some finer page level locking ?
        proc->heap_lock_count--;
        if ( proc->heap_lock_count < 0 ) {
            proc->heap_lock_count = 0;
        }
    }

    cyg_mutex_unlock(&proc->mutex);
}

/** Read memory range lock count.
 *
 * This function must be called from system mode.
 */
static int
cyg_process_mem_read_lock_count(cyg_process *proc, CYG_ADDRESS vaddr,
                                size_t size ) {
    int ret = 0;

    if (( vaddr+size > proc->heap.base )&&
        ( vaddr < proc->heap.base+proc->heap.size )) {
        ret = proc->heap_lock_count;
    }
    return ret;
}

/** Map physical memory in a process virtual space.
 *
 *
 */
Cyg_ErrNo
cyg_process_map( cyg_process *proc, CYG_ADDRESS vaddr, size_t size ) {
    int ap, status;
    struct segment *seg;
    Cyg_ErrNo ret = -EINVAL;
    cyg_process *self = cyg_process_self();

    // check that proc is valid
    if ( !self || ( self == proc )) {

        cyg_mutex_lock(&proc->mutex);

        // sanity checks
        if (( IS_ALIGNED( vaddr, MMU_PAGE_SIZE ) )&&
            ( IS_ALIGNED( size, MMU_PAGE_SIZE  ) )&&
            ( size > 0 )&&
            ( size <= (proc->max_mem-proc->used_mem) )) {

            // find a matching segment
            if (( vaddr >= proc->ro.base )&&
                ( vaddr+size <= proc->ro.base+proc->ro.size )) {
                ap = AP_RW_RO;
                seg = &(proc->ro);
            }
            else if (( vaddr >= proc->rw.base )&&
                     ( vaddr+size <= proc->rw.base+proc->rw.size )) {
                ap = AP_RW_RW;
                seg = &(proc->rw);
            }
            else if (( vaddr >= proc->heap.base )&&
                     ( vaddr+size <= proc->heap.base+proc->heap.size )) {
                ap = AP_RW_RW;
                seg = &(proc->heap);
            }
            else {
                seg = NULL;
            }
            if ( seg ) {
                cyg_uint32 used_mem;

                // ok, attempt map operation
                cyg_mutex_lock(&vm_mutex);

                used_mem = ram_get_used_mem();
                status = arm_mmu_map( vaddr,
                                      0, /* allocate physical ram */
                                      size,
                                      ap,
                                      1, /* cached */
                                      proc->domain,
                                      1, /* require allocation */
                                      0  /* do not use small pages */ );
                if ( status == 0 ) {
                    ret = ENOERR;
                }
                // update physical ram count
                proc->used_mem += ram_get_used_mem()-used_mem;

                cyg_mutex_unlock(&vm_mutex);
            }
        }
        cyg_mutex_unlock(&proc->mutex);
    }

    return ret;
}

/** Unmap physical memory from a process virtual space.
 *
 * For user threads, unmapping is authorized only within the heap segment.
 */
Cyg_ErrNo
cyg_process_unmap( cyg_process *proc, CYG_ADDRESS vaddr, size_t size ) {
    int ap, status;
    struct segment *seg;
    Cyg_ErrNo ret = -EINVAL;
    cyg_process *self = cyg_process_self();

    // check that proc is valid
    if ( !self || ( self == proc )) {

        cyg_mutex_lock(&proc->mutex);

        // sanity checks
        if (( IS_ALIGNED( vaddr, MMU_PAGE_SIZE ) )&&
            ( IS_ALIGNED( size, MMU_PAGE_SIZE  ) )&&
            ( size > 0 )) {

            // find a matching segment
            if (( vaddr >= proc->ro.base )&&
                ( vaddr+size <= proc->ro.base+proc->ro.size )) {
                ap = AP_RW_RO;
                seg = &(proc->ro);
            }
            else if (( vaddr >= proc->rw.base )&&
                     ( vaddr+size <= proc->ro.base+proc->rw.size )) {
                ap = AP_RW_RW;
                seg = &(proc->rw);
            }
            else if (( vaddr >= proc->heap.base )&&
                     ( vaddr+size <= proc->heap.base+proc->heap.size )) {
                ap = AP_RW_RW;
                seg = &(proc->heap);
            }
            else {
                seg = NULL;
            }
            // if we were called from userland, accept unmap only within
            // unlocked heap; otherwise, accept unmap in any valid segment
            if ( seg &&
                 ((( seg == &(proc->heap) )&&
                   (cyg_process_mem_read_lock_count(proc, vaddr, size) == 0))
                  || !self )) {

                cyg_uint32 used_mem;

                // ok, attempt unmap operation
                cyg_mutex_lock(&vm_mutex);

                used_mem = ram_get_used_mem();
                status = arm_mmu_unmap( vaddr,
                                        size,
                                        1, /* require deallocation */
                                        0  /* do not fail on unmapped mem */);
                if ( status == 0 ) {
                    ret = ENOERR;
                }
                // update physical ram count
                proc->used_mem -= used_mem-ram_get_used_mem();

                cyg_mutex_unlock(&vm_mutex);
            }
        }
        cyg_mutex_unlock(&proc->mutex);
    }

    return ret;
}

static dynamo_thread_t *
cyg_process_allocate_thread(void) {
    dynamo_thread_t *thread = NULL;
    int i;
    // look for an empty slot
    for ( i = 0; i < DYNAMO_MAX_THREADS; i++ ) {
        if ( dynamo_thread_pool[i].data.process == NULL ) {
            thread = &dynamo_thread_pool[i];
            break;
        }
    }
    return thread;
}

void
cyg_process_free_thread(dynamo_thread_t *thread) {
    thread->data.process = NULL;
}

/** Add a thread to a process.
 *
 */
Cyg_ErrNo
cyg_process_add_thread( cyg_process         *proc,
                        cyg_addrword_t      sched_info,
                        cyg_thread_entry_t  *entry,
                        cyg_addrword_t      entry_data,
                        char                *name,
                        void                *stack_base,
                        cyg_ucount32        stack_size,
                        cyg_handle_t        *handle,
                        cyg_thread          *_thread ) {

    Cyg_ErrNo ret = -EINVAL;
    dynamo_thread_t *thread;
    cyg_process *self = cyg_process_self();

    cyg_mutex_lock(&process_mutex);

    // check that proc is valid
    if ( !self || ( self == proc )) {

        ret = ENOERR;
        cyg_mutex_lock(&proc->mutex);

        // verify number of threads
        if ( proc->nb_threads >= proc->max_threads ) {
            ret = -EINVAL;
        }
        // allocate thread
        if ( ret == ENOERR ) {
            thread = cyg_process_allocate_thread();
            if ( !thread ) {
                ret = -ENOMEM;
            }
            else {
                // store extra thread information
                thread->data.process = proc;
                thread->data.entry = entry;
                thread->data.ustack_base = (CYG_ADDRESS)stack_base;
                thread->data.ustack_top = thread->data.ustack_base+stack_size;
                // adjust priority
                if ( sched_info < proc->sched_info ) {
                    sched_info = proc->sched_info;
                }
            }
        }
        // now create thread with redirected entry
        if ( ret == ENOERR ) {
            cyg_thread_create( sched_info,
                               cyg_process_thread_entry,/* override entry */
                               entry_data,
                               name,
                               thread->data.kstack,
                               CYGOPT_DYNAMO_KSTACK_SIZE,
                               handle,
                               &thread->obj             /* override object */);
            proc->nb_threads++;
        }
        cyg_mutex_unlock(&proc->mutex);
    }
    cyg_mutex_unlock(&process_mutex);

    return ret;
}

/** Thread entry replacement.
 *
 * This function is used as replacement for the user-provided entry of a
 * thread. It performs a few privilege-related setups and calls the real entry
 * point.
 */
void
cyg_process_thread_entry(cyg_addrword_t entry_data) {

    cyg_process *proc;
    dynamo_thread_t *thread = (dynamo_thread_t *)cyg_thread_self();

    proc = thread->data.process;

    // compute user thread DACR, client for domains '0' and process domain
    thread->data.dacr = (1 << (2*proc->domain))|1;

    // setup DACR
    arm_mmu_set_dacr( thread->data.dacr, CYGOPT_DYNAMO_DACR_MASK );

    // switch to user mode
    asm volatile ( "mrs r0,cpsr     \n\t"
                   "bic r0,r0,#0x1f \n\t"
                   "orr r0,r0,#0x10 \n\t"
                   "msr cpsr,r0     \n\t"
                   "nop" : : : "r0", "cc" );

    // switch to user mode stack
    asm volatile ( "mov sp,%0" : : "r" (thread->data.ustack_top));

    // and execute thread code
    thread->data.entry(entry_data);

    // return to system mode and exit
    SYSCALL(cyg_thread_exit)();
}

/** Check process user access to a memory range.
 *
 */
int
cyg_process_check_access(cyg_process *proc,
                         CYG_ADDRESS vaddr, size_t size, int mode) {
    int ap, dom, ok = 0;

    cyg_mutex_lock(&vm_mutex);

    // first get start address permission and domain
    ap = arm_mmu_get_ap(vaddr);
    dom = arm_mmu_get_domain(vaddr);

    // verify access: a dynamo process thread has CLIENT access to its own
    // domain and to domain 0, in USER mode
    if (( dom == 0 )||(( dom > 0 )&&( dom == proc->domain ))) {
        if ( mode == CHECK_RO ) {
            ok = (( ap == AP_RW_RO )||( ap == AP_RW_RW ));
        }
        else {
            ok = ( ap == AP_RW_RW );
        }
    }
    // access is granted to start address, now check the whole range
    if ( ok ) {
        // check access permission
        if ( arm_mmu_check_ap(vaddr, size, ap) != 0 ) {
            ok = 0;
        }
        // check domain
        while ( ok ) {
            if ( arm_mmu_get_domain(vaddr+size-1) != dom ) {
                ok = 0;
            }
            else {
                if ( size <= MMU_SECTION_SIZE ) {
                    break;
                }
                else {
                    size -= MMU_SECTION_SIZE;
                }
            }
        }
    }
    cyg_mutex_unlock(&vm_mutex);

    return ( ok )? 0 : -1;
}

int
cyg_process_range_get(CYG_ADDRESS vaddr, size_t size, int mode) {
    int ret = -1;

    cyg_process *proc = cyg_process_self();

    if ( proc ) {
        cyg_process_mem_lock(proc, vaddr, size);
        ret = cyg_process_check_access(proc, vaddr, size, mode);
        if ( ret < 0 ) {
            cyg_process_mem_unlock(proc, vaddr, size);
        }
    }
    return ret;
}

void
cyg_process_range_release(CYG_ADDRESS vaddr, size_t size) {
    cyg_process *proc = cyg_process_self();
    if ( proc ) {
        cyg_process_mem_unlock(proc, vaddr, size);
    }
}

void cyg_process_imb(void *base, unsigned int size)
{
    CYG_ADDRESS addr;
    cyg_process *proc = cyg_process_self();

    addr = (CYG_ADDRESS)base;

    if (proc &&
        (addr >= proc->rw.base)&&(addr+size <= proc->rw.base+proc->rw.size)) {

        arm_mmu_clean_dcache_range(addr, size);
        arm_mmu_invalidate_icache_range(addr, size);
    }
}
