//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================

/**
*******************************************************************************
* @file  memprof.c
* @brief A simple memory profiling package for eCos
*
* Copyright (C) 2007 Parrot S.A.
*
* @author     ivan.djelic@parrot.com
* @date       2007-11-19
*******************************************************************************
*/

#include <cyg/kernel/kapi.h>
#include <cyg/infra/diag.h>
#include <cyg/hal/hal_intr.h>
#include <stdlib.h>
#include <string.h>

#include <pkgconf/system.h>
#include <pkgconf/memalloc.h>

#define MEMPROF_MAGIC        (0x31415927) //!< tag for allocated zone
#define MEMPROF_MAGIC_FREED  (0xdeadbeef) //!< tag for freed zone
#define MEMPROF_FENCE        (0x1eaf)     //!< tag for end of allocated zone

static struct memprof_tag *memprof_tag_head = NULL;
static cyg_uint32         memprof_heap_alloc = 0;
static cyg_uint32         memprof_heap_alloc_max = 0;

static struct memprof_tag common_tag = {0};
static const struct memprof_tag invalid_tag = {.magic = MEMPROF_MAGIC_FREED};

/* original eCos interface */
extern void *memalloc_calloc(size_t , size_t);
extern void memalloc_free(void *);
extern void *memalloc_malloc(size_t);
extern void *memalloc_realloc(void *, size_t);

#define MEMPROF_FAIL(_tag, _fmt, args...) {                             \
    struct memprof_tag *__tag = (struct memprof_tag *)_tag;             \
    diag_printf("memprof: ERROR: " _fmt, ##args);                       \
    if (__tag) memprof_print_context(__tag);                            \
    while (1);                                                          \
}

#define MEMPROF_WARN(_tag, _fmt, args...) {                             \
    struct memprof_tag *__tag = (struct memprof_tag *)_tag;             \
    diag_printf("memprof: WARNING: " _fmt, ##args);                     \
    if (__tag) memprof_print_context(__tag);                            \
}

static void memprof_print_context(struct memprof_tag *tag)
{
    diag_printf("memprof: memory was allocated in function %s(), line %d\n",
                tag->func, tag->line);
}

#define LOCK()       cyg_scheduler_lock()
#define UNLOCK()     cyg_scheduler_unlock()
#define INIT_LOCK()  do {} while (0)

static void memprof_update_samples(struct memprof_tag *tag)
{
    if (tag->calls < tag->sample[tag->sample_index]) {
        tag->sample[tag->sample_index] = tag->calls;
    }
    tag->sample_count++;
    if (tag->sample_count >= tag->sample_size) {
        tag->sample_count = 0;
        tag->sample_index++;

        if (tag->sample_index >= MEMPROF_TAG_SAMPLES) {
            int i;
            for (i = 0; i < MEMPROF_TAG_SAMPLES/2; i++) {
                tag->sample[i] = (tag->sample[2*i] < tag->sample[2*i+1])?
                    tag->sample[2*i] : tag->sample[2*i+1];
            }
            tag->sample_index = MEMPROF_TAG_SAMPLES/2;
            tag->sample_size *= 2;
        }
        tag->sample[tag->sample_index] = tag->calls;
    }
}

void memprof_register(struct memprof_tag *tag, const char *func, int line)
{
    cyg_uint32 old;

    if (!tag->magic) {

        HAL_DISABLE_INTERRUPTS(old);

        /* magic may have changed due to concurrency */
        if (!tag->magic) {

            memset(tag, 0, sizeof(struct memprof_tag));

            tag->magic = MEMPROF_MAGIC;
            tag->func = func;
            tag->line = line;
            tag->sample_size = 1;
            tag->sample[0] = (cyg_uint32)(-1);

            if (memprof_tag_head == NULL) {

                INIT_LOCK();

                /* try to detect inclusion of this package in release builds */
#if !defined(CYGPKG_INFRA_DEBUG)
                MEMPROF_WARN(NULL,"do not enable memprof in RELEASE builds!\n");
#endif
            }

            tag->next = memprof_tag_head;
            memprof_tag_head = tag;
        }

        HAL_RESTORE_INTERRUPTS(old);
    }
}

/**
 * Format an allocated block.
 *
 * Note: this should be called with lock held.
 */
static void * memprof_format_block(struct memprof_tag *tag, void *addr,
                                   size_t size)
{
    cyg_uint8 *pb;
    cyg_uint32 *ptr = addr;

    tag->alloc += size;
    tag->alloc_total += (cyg_uint64)size;
    tag->calls++;
    tag->calls_total++;

    memprof_heap_alloc += size;
    if (memprof_heap_alloc > memprof_heap_alloc_max) {
        memprof_heap_alloc_max = memprof_heap_alloc;
    }

    /* write cookie (4 bytes) */
    *ptr++ = (cyg_uint32)tag;

    /* store allocated size (4 bytes) */
    *ptr++ = (cyg_uint32)size;

    /* mark end of allocated buffer (2 bytes) */
    pb = ((cyg_uint8 *)ptr) + size;
    pb[0] = MEMPROF_FENCE & 0xff;
    pb[1] = MEMPROF_FENCE >> 8;

    return ptr;
}

/**
 * Check an allocated block for corruption.
 *
 * Note: this should be called with lock held.
 */
static cyg_uint32 *memprof_check_block(void *ptr)
{
    cyg_uint8 *pb;
    struct memprof_tag *tag;
    cyg_uint32 *cookie = (cyg_uint32 *)(((cyg_uint32)ptr)-8);

    tag = (struct memprof_tag *)cookie[0];

    if (tag->magic != MEMPROF_MAGIC) {
        if (tag->magic == MEMPROF_MAGIC_FREED) {
            MEMPROF_FAIL(NULL, "referencing freed memory (0x%08x) !\n",
                         (cyg_uint32)ptr);
        }
        else {
            MEMPROF_FAIL(NULL, "bad free pointer (0x%08x) or corrupt memory!\n",
                         (cyg_uint32)ptr);
        }
    }

    if ((tag->alloc < cookie[1])||(tag->calls == 0)) {
        MEMPROF_FAIL(tag, "corrupt memory block (0x%08x) ?!\n",(cyg_uint32)ptr);
    }

    pb = ((cyg_uint8 *)&cookie[2])+cookie[1];
    if ((pb[0]|(pb[1] << 8)) != MEMPROF_FENCE) {
        MEMPROF_FAIL(tag, "overflow or corruption at 0x%08x !\n",
                     (cyg_uint32)pb);
    }

    return cookie;
}

void * memprof_alloc(struct memprof_tag *tag, size_t size, int zero)
{
    void *ptr;

#ifdef CYGSEM_MEMALLOC_MALLOC_ZERO_RETURNS_NULL
    if (size == 0) {
        MEMPROF_WARN(tag, "allocating 0 bytes, returning NULL !\n");
        return NULL;
    }
#endif

    ptr = memalloc_malloc(size+10);
    if (ptr) {

        LOCK();
        ptr = memprof_format_block(tag, ptr, size);
        UNLOCK();

        if (zero) {
            memset(ptr, 0, size);
        }
    }

    return ptr;
}

void *memprof_realloc(struct memprof_tag *tag, void *ptr, size_t size)
{
    cyg_uint32 oldsize, *cookie;

    if (ptr == NULL) {
        return memprof_alloc(tag, size, 0);
    }

    LOCK();
    cookie = memprof_check_block(ptr);
    oldsize = cookie[1];
    tag = (struct memprof_tag *)cookie[0];
    UNLOCK();

    ptr = memalloc_realloc(cookie, size+10);

    LOCK();
    if (ptr) {
        tag->calls--;
        memprof_update_samples(tag);
        tag->alloc -= oldsize;
        memprof_heap_alloc -= oldsize;
        ptr = memprof_format_block(tag, ptr, size);
    }
    UNLOCK();

    return ptr;
}

/* now provide malloc(), calloc() and realloc() for code compiled without
 * the memprof package.
 */
#undef malloc
#undef calloc
#undef realloc

void *malloc(size_t size)
{
    memprof_register(&common_tag, "common malloc", 0);
    return memprof_alloc(&common_tag, size, 0);
}

void *calloc(size_t nmemb, size_t size)
{
    memprof_register(&common_tag, "common malloc", 0);
    return memprof_alloc(&common_tag, nmemb*size, 1);
}

void *realloc(void *ptr, size_t size)
{
    memprof_register(&common_tag, "common malloc", 0);
    return memprof_realloc(&common_tag, ptr, size);
}

void free(void *ptr)
{
    struct memprof_tag *tag;
    cyg_uint32 *cookie;

    if (ptr) {

        LOCK();

        cookie = memprof_check_block(ptr);
        tag = (struct memprof_tag *)cookie[0];

        tag->calls--;
        memprof_update_samples(tag);
        tag->alloc -= cookie[1];
        memprof_heap_alloc -= cookie[1];

        /* invalidate block */
        cookie[0] = (cyg_uint32)&invalid_tag;

        UNLOCK();

        memalloc_free(cookie);
    }
    else {
        MEMPROF_WARN(NULL, "NULL pointer passed to free()\n");
    }
}

#define FUNCNAME_LEN 18
#define XSTR(_x)     STR(_x)
#define STR(_x)      #_x

#define SLINE                                   \
    "----------------------------------------"  \
    "---------------------------------------\n"

#define DLINE                                   \
    "========================================"  \
    "=======================================\n"

static void
print_size(cyg_uint32 size)
{
    if (size < 10*1024) {
        diag_printf("%6d", size);
    }
    else if (size < 1000*1024) {
        diag_printf("%3d kB", size >> 10);
    }
    else {
        diag_printf("%3d MB", size >> 20);
    }
}

void memprof_report(void)
{
    int len, i;
    struct memprof_tag *tag;
    struct mallinfo info = mallinfo();

    LOCK();

    diag_printf(DLINE "HEAP: total=%d kB free=%d kB alloc=%d kB "
                "peak=%d kB (%d %%) ovhd=%d kB\n" DLINE,
                info.arena >> 10,
                info.fordblks >> 10,
                memprof_heap_alloc >> 10,
                memprof_heap_alloc_max >> 10,
                100*(memprof_heap_alloc_max >> 10)/(info.arena >> 10),
                (info.arena-info.fordblks-memprof_heap_alloc) >> 10);

    diag_printf("  %" XSTR(FUNCNAME_LEN) "s:line|alloced (%c)|avg sz|"
                "unfree/tot|unfree minima history\n" SLINE, "function name",
                '%');

    for (tag = memprof_tag_head; tag != NULL; tag = tag->next) {

        /* display function name */
        len = strlen(tag->func);
        if (len > FUNCNAME_LEN) {
            for (i = 0; i < FUNCNAME_LEN/2-1; i++) {
                diag_write_char(tag->func[i]);
            }
            diag_printf("..%s", tag->func+len-FUNCNAME_LEN/2+1);
        }
        else {
            diag_printf("%" XSTR(FUNCNAME_LEN) "s", tag->func);
        }
        diag_printf("():%-4d|", tag->line);

        print_size(tag->alloc);
        diag_printf("(%2d%c)|", tag->alloc*100/memprof_heap_alloc, '%');

        print_size((tag->calls_total)?
                   (cyg_uint32)(tag->alloc_total/tag->calls_total) : 0);

        diag_printf("|%4u/%-5u|", tag->calls, tag->calls_total);

        /* display unfree statistics */
        if (tag->sample_index) {
            int i;
            for (i = 0; i <= tag->sample_index; i++) {
                diag_printf("%d,", tag->sample[i]);
            }
        }
        else {
            diag_printf("no call to free() !");
        }
        diag_printf("\n");
    }
    diag_printf(DLINE);

    UNLOCK();
}
