//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================

/**
*******************************************************************************
* @file  getrtd.c
* @brief Real-Time Monitor host tools
*
* This utility is used to download an RTmon dump file from a target.
*
* Usage:
* getrtd [-h] [-o <dumpfile> [-l]] < /dev/ttySx
*
* Examples:
*   getrtd < /dev/ttySx > dump
*   netcat sonde2 8 | getrtd > dump
*   cat /dev/ttyS1 | getrtd -o mydump -l
*
* Copyright (C) 2006 Parrot S.A.
*
* @author     ivan.djelic@parrot.fr
* @date       2006-12-08
*******************************************************************************
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <time.h>
#include <errno.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "ecos_types.h"
#include "rtmon_common.h"
#include "posix_crc.h"

#define PFX "getrtd: "

#define RTMON_BLOCK_SIZE     (8192)
#define MAX_FILENAME         (4096)


#define PRINT_MEMORY(_addr_,_nb_)   {                                      \
    int i;                                                                 \
    for(i=0; i<_nb_; i++){                                                 \
      fprintf( stderr, PFX "\t  0x%08x :", &( ((uint32_t*)_addr_)[i]) );   \
      fprintf( stderr, "\t0x%08x 0x%08x 0x%08x 0x%08x\n",                  \
              ((uint32_t*)_addr_)[(i*4)],                                  \
              ((uint32_t*)_addr_)[(i*4)+1],                                \
              ((uint32_t*)_addr_)[(i*4)+2],                                \
              ((uint32_t*)_addr_)[(i*4)+3]);                               \
    }   }   


//! Download automaton states
enum {
    RTMON_ST_MAGIC,                          //!< Waiting for MAGIC header
    RTMON_ST_HEADER,                         //!< Downloading header
    RTMON_ST_THD_TAG,                        //!< Dowloading thread tag table
    RTMON_ST_USER_EVENT_TAG,                 //!< Dowloading user event tag table
    RTMON_ST_USER_U32_TAG,
    RTMON_ST_BUFFER,                         //!< Dowloading buffer
};

static const uint64_t magic = RTMON_MAGIC;
static const char const *pmagic = (const char const *)&magic;

static void
usage(void) {
    fprintf ( stderr,
              "getrtd: Download and save an RTmon dump\n"
              "usage: getrtd [-h] [-o <filename> [-l]] < infile\n"
              " -h\t\t: Display this help\n"
              " -o <filename>\t: Write dump to <filename> instead of stdout\n"
              " -l\t\t: Run in (infinite) loop mode and save each new dump "
              "in\n\t\t  <filenameXXXX> where XXXX is a time stamp and "
              "<filename> was\n\t\t  set with option -o\n" );
    exit(1);
}

/** Append date and time to a string to get a suitable filename.
 *
 * @param prefix String to which time stamp is appended
 * @return A pointer to a static buffer containing the resulting string
 */
static char *
append_stamp( const char *prefix ) {
    struct stat st;
    char        timebuf[80];
    static char filename[MAX_FILENAME];
    time_t      now;
    char        *res;

    res = NULL;

    do {
        now = time(NULL);

        strftime( timebuf,79, "%y%m%d_%Hh%Mm%Ss", localtime(&now) );
        snprintf( filename, MAX_FILENAME, "%s_%s", prefix, timebuf );

        // do not clobber existing files
        if (( stat( filename, &st ) != 0 )&&( errno == ENOENT )) {
            res = filename;
        }

    } while ( !res );

    return res;
}

/** Download an RTmon dump from stdin and output a dump file.
 *
 * @param outfile Output file name. When this parameter is NULL, dump file
 * is written to stdout.
 * @param unique A boolean flag. When set, a unique stamp is appended to
 * string 'outfile' to get the final output filename.
 * @return 0 upon success, 1 upon failure.
 */
static int getdump( const char *outfile, int unique, int verbose ) {
   int           i, c, match, state, forward;
   size_t        ret;
   struct        rtmon_header hdr;
   uint8_t       *buf                 = NULL;
   uint8_t       *buf_thd_tag         = NULL;
   event_tag_t   *buf_user_event_tag  = NULL;
   event_tag_t   *buf_user_u32_tag    = NULL;
   uint32_t      rcount, nbread, bs, crc;
   static        uint32_t last_crc32 = 0;
   int           done = 0;

   c = 0;
   state = RTMON_ST_MAGIC;

   // Disable buffering on stdout
   setvbuf( stdout, NULL, _IONBF, 0);

   fprintf( stderr, PFX "scanning for file header...\n" );

   do {
      switch (state) {

      case RTMON_ST_MAGIC:
         // scan for magic number
         match = 1;
         for ( i = 0;(( i < sizeof(magic)) && match); i++ ) 
         {
            c = getchar();
            // copy caracter to stdout if we are dumping data to a file
            forward = ( outfile != NULL );
            if ( c != pmagic[i] ){
               match = 0;
               if ( i > 0 ){
                  forward = 0;
                  ungetc( c, stdin );
               }
            }
            if ( forward ){
               putchar(c);
            }
         }// end for
         if ( match ){
            state = RTMON_ST_HEADER;
            fprintf( stderr, PFX "got magic number...\n" );
            hdr.magic = magic;
         }
         break;

      case RTMON_ST_HEADER:
         // read header
         ret = fread( &hdr.size, sizeof(struct rtmon_header)-sizeof(magic), 1, stdin );
         if ( ret != 1 ){
            fprintf( stderr, PFX "[ EE ] ERROR reading header: %s\n", strerror(errno) );
            fprintf( stderr, PFX "retrying...\n" );
            state = RTMON_ST_MAGIC;
         }else if (hdr.size == 0){
            fprintf( stderr, PFX "[ EE ] ERROR bufsize=%d bytes\n", (int)hdr.size );
            fprintf( stderr, PFX "\tyou may have not called rtmon_start()\n");
            state = RTMON_ST_BUFFER;
         }else{
            fprintf( stderr, PFX "[ OK ] Header\n");
            fprintf( stderr, PFX "\tBuffer size               = %d bytes \n", (int)hdr.size );
            fprintf( stderr, PFX "\tNbr of threads labels     = %d\n", (int)hdr.nb_thread );
            fprintf( stderr, PFX "\tNbr of user events labels = %d\n", (int)hdr.nb_user_event_tag );
            fprintf( stderr, PFX "\tNbr of user u32 labels    = %d\n", (int)hdr.nb_user_u32_tag );
            state = RTMON_ST_THD_TAG;
         }
         break;

      case RTMON_ST_THD_TAG:
         // read header
         assert(hdr.nb_thread);
         buf_thd_tag = (uint8_t *)malloc(RTMON_TAGSIZE*hdr.nb_thread);
         assert(buf_thd_tag);
         ret = fread( buf_thd_tag, (RTMON_TAGSIZE*hdr.nb_thread), 1, stdin );
         if ( ret != 1 ) 
         {
            fprintf( stderr, PFX "[ EE ] ERROR reading thread tag: %s\n", strerror(errno) );
            fprintf( stderr, PFX "retrying...\n" );
            state = RTMON_ST_HEADER;
         }else{
            assert(hdr.nb_thread);
            fprintf( stderr, PFX "[ OK ] thread tag: nb_thread=%d\n", (int)hdr.nb_thread );

            if(hdr.nb_user_event_tag != 0)    { state = RTMON_ST_USER_EVENT_TAG;}
            else if(hdr.nb_user_u32_tag != 0) { state = RTMON_ST_USER_U32_TAG;  }
            else                              { state = RTMON_ST_BUFFER;        }
         }
         break;

      case RTMON_ST_USER_EVENT_TAG:
         // read header
         buf_user_event_tag = (event_tag_t *)malloc(sizeof(event_tag_t)*hdr.nb_user_event_tag);
         assert(buf_user_event_tag);
         ret = fread( buf_user_event_tag, sizeof(event_tag_t)*hdr.nb_user_event_tag, 1, stdin );

         if(verbose){
            //PRINT_MEMORY(buf_user_event_tag,sizeof(event_tag_t)*hdr.nb_user_event_tag);
            fprintf( stderr, PFX "[    ] List of user events labels\n"); 
            for(i=0; i<hdr.nb_user_event_tag; i++)
               fprintf( stderr, PFX "\t%d : %s\n",(int)buf_user_event_tag[i].id, buf_user_event_tag[i].tag);
         }

         if ( ret != 1 ){
            state = RTMON_ST_HEADER;
            fprintf( stderr, PFX "[ EE ] ERROR reading user event tag: %s\n",strerror(errno) );
            fprintf( stderr, PFX "retrying...\n" );
         }else{
            if(hdr.nb_user_u32_tag != 0) state = RTMON_ST_USER_U32_TAG;
            else                         state = RTMON_ST_BUFFER;
            fprintf( stderr, PFX "[ OK ] user event tag: nb_user_event_tag=%d\n",
                     (int)hdr.nb_user_event_tag );
         }
         break;

      case RTMON_ST_USER_U32_TAG:
         assert(hdr.nb_user_u32_tag);
         buf_user_u32_tag = (event_tag_t *)malloc(sizeof(event_tag_t)*hdr.nb_user_u32_tag);
         assert(buf_user_u32_tag);
         ret = fread( buf_user_u32_tag, sizeof(event_tag_t)*hdr.nb_user_u32_tag, 1, stdin );

         if(verbose){
            /*PRINT_MEMORY(buf_user_u32_tag,sizeof(event_tag_t)*hdr.nb_user_u32_tag);*/
            fprintf( stderr, PFX "[    ] List of user u32 labels\n"); 
            for(i=0; i<hdr.nb_user_u32_tag; i++)
               fprintf( stderr, PFX "\t%d : %s\n",(int)buf_user_u32_tag[i].id, buf_user_u32_tag[i].tag); 
         }

         if ( ret != 1 ){
            state = RTMON_ST_HEADER;
            fprintf( stderr, PFX "[ EE ] ERROR reading user u32 tag: %s\n", strerror(errno) );
            fprintf( stderr, PFX "retrying...\n" );
         }else{
            state = RTMON_ST_BUFFER;
            fprintf( stderr, PFX "[ OK ] user u32 tag\n" );
         }
         break;


      case RTMON_ST_BUFFER:
         // read buffer
         assert(hdr.size);
         buf = (uint8_t *)malloc(hdr.size);
         assert(buf);

         rcount = hdr.size;
         nbread = 0;
         ret = 0;
         fprintf( stderr, PFX "reading buffer : [00%%]" );

         while ( rcount > 0 ) {
            bs = ( rcount >= RTMON_BLOCK_SIZE )?
               RTMON_BLOCK_SIZE : rcount;
            if ( fread( &buf[nbread], 1, bs, stdin ) != bs ) {
               ret = -1;
               break;
            }
            rcount -= bs;
            nbread += bs;
            fprintf( stderr, "\b\b\b\b\b[%02d%%]",(int)(nbread*100/hdr.size));
         }
         fprintf( stderr, "\b\b\b\b\b\b done !\n" );

         if ( ret != 0 ) {
            state = RTMON_ST_MAGIC;
            fprintf( stderr, PFX "[ EE ] ERROR reading buffer: %s\n", strerror(errno));
            fprintf( stderr, PFX "retrying...\n" );
         }
         else {
            // check data integrity
            fprintf( stderr, PFX "[    ] checking data integrity... \n" );
            crc = posix_crc( buf, hdr.size );

            if ( crc != hdr.crc32 ) {
               state = RTMON_ST_MAGIC;
               fprintf( stderr,
                        "bad CRC (0x%08x, should be 0x%08x) !\n",
                        (unsigned int)crc,
                        (unsigned int)hdr.crc32 );
               fprintf( stderr, PFX "retrying...\n" );
            }else{
               fprintf( stderr, PFX "[ OK ] Buffer integrity is correct \n" );
               done = 1;
            }
         }
         if ( state == RTMON_ST_MAGIC ) {
            free(buf);                buf = NULL;
            free(buf_thd_tag);        buf_thd_tag = NULL;
            free(buf_user_event_tag); buf_user_event_tag = NULL;
            free(buf_user_u32_tag);   buf_user_u32_tag = NULL;
         }
         break;

      default:
         assert(0);
         break;
      }
   } while ( !done && ( c != EOF ));

   if ( !done ) {
      // transfer was not completed
      fprintf( stderr, PFX "[ EE ] premature EOF, no output written !\n" );
      ret = 1;
   }else{
      // did we already save this dump ?
      if ( hdr.crc32 == last_crc32 ) {
         fprintf( stderr, PFX "this dump was already saved...\n" );
         ret = 0;
      }
      else {
         // save dump
         FILE *out;
         int errcode = 0;

         // open output file
         if ( outfile && unique ){
               outfile = append_stamp(outfile);
         }

         if( outfile ){
            out = fopen( outfile, "wb" );
            if ( !out ) {
               fprintf( stderr, PFX "[ EE ] cannot open %s for writing: %s\n",
                        outfile, strerror(errno));
               errcode = 1;
            }
         }else{
            fprintf( stderr, PFX "writing output to 'stdout'...\n");
            out = stdout;
         }

         // output file header
         if ( !errcode ) {
            ret = fwrite( &hdr, sizeof(struct rtmon_header), 1, out );
            errcode = ( ret != 1 );
         }
         // output thread tag buffer
         if ( !errcode ) {
            ret = fwrite( buf_thd_tag, 1, (hdr.nb_thread*RTMON_TAGSIZE), out );
            errcode = ( ret != (hdr.nb_thread*RTMON_TAGSIZE) );
         }

         // output user event tag buffer
         if ( !errcode ) {
            ret = fwrite( buf_user_event_tag, 1, (sizeof(event_tag_t)*hdr.nb_user_event_tag), out );
            errcode = ( ret != sizeof(event_tag_t)*hdr.nb_user_event_tag);
         }

         // output user u32 tag buffer
         if ( !errcode ) {
            ret = fwrite( buf_user_u32_tag, 1, (sizeof(event_tag_t)*hdr.nb_user_u32_tag), out );
            errcode = ( ret != sizeof(event_tag_t)*hdr.nb_user_u32_tag);
         }

         // output dump buffer
         if ( !errcode ) {
            ret = fwrite( buf, 1, hdr.size, out );
            errcode = ( ret != hdr.size );
         }

         if ( errcode ) {
            fprintf( stderr, PFX "[ EE ] cannot write dump to '%s': %s\n",
                     ( outfile )? outfile : "stdout", strerror(errno));
            if ( out && ( out != stdout )) {
               unlink(outfile);
            }
            ret = 1;
         }
         else {
            fprintf( stderr, PFX "[    ] dump file successfully written !\n" );
            last_crc32 = hdr.crc32;
            ret = 0;
         }
         // cleanup
         if ( out && ( out != stdout )){
            fclose(out);
         }
      }
   }

   // cleanup
   if( buf )                free(buf);
   if( buf_thd_tag )        free(buf_thd_tag);
   if( buf_user_event_tag ) free(buf_user_event_tag);
   if( buf_user_u32_tag )   free(buf_user_u32_tag);

   return ret;
}


int
main( int argc, char *argv[] ) {

    int ret, loopmode = 0, verbose = 0;
    char *outpat = NULL;
    int c;

    // parse options
    while (( c = getopt( argc, argv, "hvlo:" )) != -1 ) {
        switch (c) {

        case 'o':
            outpat = optarg;
            break;

        case 'l':
            loopmode = 1;
            break;

        case 'v':
            verbose = 1;
            break;

        case 'h':
        default:
            usage();
            break;
        }
    }

    if ( loopmode && !outpat ) {
        // loopmode is only valid if an output file pattern was specified
        usage();
    }

    do {
        ret = getdump( outpat, loopmode, verbose );
    } while (( ret == 0 )&& loopmode );

    return ret;
}
