

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <errno.h>
#include <assert.h>
#include <stdarg.h>

#define _GNU_SOURCE
#include <getopt.h>

#include <search.h>

#include "lxt_write.h"

#include "ecos_types.h"
#include "rtmon_common.h"
#include "posix_crc.h"
#include "rtd2lxt.h"

#define PFX             "rtd2lxt-pkt: "

static uint32_t thd_state_current[THREAD_ID_MAX];


/*-----------------------------------------------------------------------------
 * LXT Symbols functions
 *---------------------------------------------------------------------------*/
#define DEFINE_LT_SYMBOL(_fname_, _ltsymbol_, _type_, _string_)    \
    struct lt_symbol *_ltsymbol_ = NULL;                                \
    static struct lt_symbol * _fname_(void) {                           \
        if ( !_ltsymbol_ ) {                                            \
            _ltsymbol_ = add_symbol( _type_, _string_ );                \
        }                                                               \
        return _ltsymbol_;                                              \
    }


static struct lt_symbol * add_symbol( uint32_t flags, const char *fmt, ... )
{
    static char linebuf[LINEBUF_MAX];
    struct lt_symbol * ret;
    va_list ap;

    va_start(ap, fmt);
    vsnprintf( linebuf, LINEBUF_MAX, fmt, ap );
    va_end(ap);

    ret = lt_symbol_find( lt, linebuf );

    if ( !ret ) {
        ret = lt_symbol_add( lt, linebuf, 0, 0, 0, flags );
        assert(ret);
    }

    return ret;
}

/** Add ISR tags
 */
struct lt_symbol *isr_lts[RTMON_PKT_IRQ_MAX];
static struct lt_symbol *isr_symbol(int vec)
{
    if ( !isr_lts[vec] ) {
        if ( hdr.irq_tag[vec][0] ) {
            isr_lts[vec] = add_symbol( LT_SYM_F_BITS, "ISR %s",
                                       hdr.irq_tag[vec] );
        }
        else {
            isr_lts[vec] = add_symbol( LT_SYM_F_BITS, "ISR %d", vec );
        }
    }
    return isr_lts[vec];
}

/** Add DSR tags
 */
struct lt_symbol *dsr_lts[RTMON_PKT_IRQ_MAX];
static struct lt_symbol *dsr_symbol(int vec)
{
    if ( !dsr_lts[vec] ) {
        if ( hdr.irq_tag[vec][0] ) {
            dsr_lts[vec] = add_symbol( LT_SYM_F_BITS, "DSR %s",
                                       hdr.irq_tag[vec] );
        }
        else {
            dsr_lts[vec] = add_symbol( LT_SYM_F_BITS, "DSR %d", vec );
        }
    }
    return dsr_lts[vec];
}


/** Add thread tags
 *
 * Thread tags are added in thd_tag table when a RTMON_PKT_THD_CREATE packet
 * is read.
 */
struct lt_symbol *thread_lts[THREAD_ID_MAX];

static struct lt_symbol *thread_symbol(int id)
{
    assert( id < THREAD_ID_MAX );

    if ( !thread_lts[id] )
    {
        if (( id < THREAD_ID_MAX )&& thd_tag[id][0] ){
            thread_lts[id] = add_symbol( LT_SYM_F_BITS, "%s", thd_tag[id] );

        }else {
            thread_lts[id] = add_symbol( LT_SYM_F_BITS, "Thread %d", id );
        }
    }
    return thread_lts[id];
}

/** Add thread state tags
 *
 */
struct lt_symbol *thread_state_lts[THREAD_ID_MAX];
static struct lt_symbol *thread_state_symbol(int id, uint32_t *thd_state)
{
    assert( id < THREAD_ID_MAX );

    if ( !thread_state_lts[id] )
    {
        *thd_state = 0xFF;
        if (( id < THREAD_ID_MAX )&& thd_tag[id][0] ){
            //thread_state_lts[id] = add_symbol( LT_SYM_F_STRING, "state[id=%d]", id );
            thread_state_lts[id] = add_symbol( LT_SYM_F_STRING, "%s state", thd_tag[id] );
        }
    }
    return thread_state_lts[id];
}



DEFINE_LT_SYMBOL(thd_prio_chg_symbol, thd_prio_chg_lts,
                      LT_SYM_F_BITS,"Priority change");

DEFINE_LT_SYMBOL(current_prio_symbol, current_thd_prio_lts,
                      LT_SYM_F_INTEGER,"Current thread prio");

DEFINE_LT_SYMBOL(sched_symbol, sched_lts,
                      LT_SYM_F_INTEGER,"Scheduler lock count");

DEFINE_LT_SYMBOL(schedbit_symbol, schedbit_lts,
                      LT_SYM_F_BITS,"Scheduler locked");

DEFINE_LT_SYMBOL(sched_func_symbol, sched_func_lts,
                    LT_SYM_F_INTEGER,"Scheduler caller");

DEFINE_LT_SYMBOL(sched_sfunc_symbol, sched_sfunc_lts,
                    LT_SYM_F_STRING, "Scheduler caller");

DEFINE_LT_SYMBOL(trace_sfunc_symbol, trace_sfunc_lts,
                    LT_SYM_F_STRING, "Caller" );

DEFINE_LT_SYMBOL(trace_func_symbol, trace_func_lts,
                    LT_SYM_F_INTEGER, "Caller" );

DEFINE_LT_SYMBOL(pc_sfunc_symbol, pc_sfunc_lts,
                    LT_SYM_F_STRING, "PC" );

DEFINE_LT_SYMBOL(pc_func_symbol, pc_func_lts,
                    LT_SYM_F_INTEGER, "PC" );

DEFINE_LT_SYMBOL(prof_sfunc_symbol, prof_sfunc_lts,
                    LT_SYM_F_STRING, "Profiling" );

DEFINE_LT_SYMBOL(prof_func_symbol, prof_func_lts,
                    LT_SYM_F_INTEGER, "Profiling" );

DEFINE_LT_SYMBOL(aai_reg_its_symbol, aai_reg_its_lts,
                    LT_SYM_F_INTEGER, "AAI IT src" );

DEFINE_LT_SYMBOL(aai_reg_iten_symbol, aai_reg_iten_lts,
                    LT_SYM_F_INTEGER, "AAI reg ITEN" );

DEFINE_LT_SYMBOL(aai_reg_cfg_symbol,aai_reg_cfg_lts,
                    LT_SYM_F_INTEGER, "AAI reg CFG" );

DEFINE_LT_SYMBOL(aai_pc_sfunc_symbol,aai_pc_sfunc_lts,
                    LT_SYM_F_STRING, "AAI function" );

DEFINE_LT_SYMBOL(aai_pc_func_symbol,aai_pc_func_lts,
                    LT_SYM_F_INTEGER, "AAI function" );

/** Add User u32 tags
 *
 */
struct lt_symbol *user_val_lts[RTMON_PKT_USER_U32_MAX];
static struct lt_symbol *user_val_symbol(int id)
{
   uint32_t i=0;

   while( (i < RTMON_PKT_USER_U32_MAX ) &&
          (i < hdr.nb_user_u32_tag    ) )
   {
         if (( user_u32_tag[i].id == id) && (user_u32_tag[i].tag[0])){
            user_val_lts[id] = add_symbol( LT_SYM_F_INTEGER, "%s", user_u32_tag[i].tag );
            return user_val_lts[id];
         }
         i++;
   }
   // tag has not been found
   user_val_lts[id] = add_symbol( LT_SYM_F_INTEGER, "User value #%d", id );

   return user_val_lts[id];
}

/** Add User event tags
 *
 */
struct lt_symbol *user_lts[RTMON_PKT_EVENT_MAX];
static struct lt_symbol *user_symbol(int id)
{
   uint32_t i=0;

   while( (i < RTMON_PKT_EVENT_MAX   ) &&
          (i < hdr.nb_user_event_tag ) )
   {
      if (( user_event_tag[i].id == id) && (user_event_tag[i].tag[0])){
            user_lts[id] = add_symbol( LT_SYM_F_BITS, "%s", user_event_tag[i].tag );
            return user_lts[id];
      }
      i++;
   }
   // tag has not been found
   user_lts[id] = add_symbol( LT_SYM_F_BITS, "User event #%d", id );

   return user_lts[id];
}

/** Add Driver event tags
 *
 */
struct lt_symbol *driver_lts[RTMON_PKT_EVENT_MAX];
static struct lt_symbol *driver_symbol(int id) {
    if ( !driver_lts[id] ) {
        driver_lts[id] = add_symbol( LT_SYM_F_BITS, "Driver event #%d", id );
    }
    return driver_lts[id];
}



/*---------------------------------------------------------------------------*/
/* Packet management                                                         */
/*---------------------------------------------------------------------------*/

int emit_clock( uint32_t pkt )
{
    int ret = 0;
    lxttime_t timeval;
    static lxttime_t oldtimeval = 0;

    timeval = ((lxttime_t)timer.ticks)*hdr.tickperiod;

    if(RTMON_PKT_STAMP(pkt) == 0)
        timeval += hdr.tickperiod;
    else {
        if (RTMON_PEND_TICK(pkt))
            if (RTMON_PKT_STAMP(pkt) < (hdr.tickperiod - 1))
                timeval += hdr.tickperiod;
            else
                //We have detected that our time stamp will be faulty
                fprintf( stderr, PFX "Workaround for Time @%lld\n", oldtimeval);
        else
            //in this case, we memorize time stamp
            timer.count = RTMON_PKT_STAMP(pkt);
    }

    timeval += RTMON_PKT_STAMP(pkt);
    timeval *= (lxttime_t)hdr.timefact_mult;
    timeval /= (lxttime_t)hdr.timefact_div;

    if ( timeval < oldtimeval ) {
        fprintf( stderr, PFX "negative time offset @%lld: %d !\n", oldtimeval,
                 (int)((int64_t)timeval-(int64_t)oldtimeval) );
        fprintf( stderr, PFX "ne=%04x  !\n",(unsigned int)RTMON_PKT_STAMP(pkt));
        timeval = oldtimeval + 1;
    }

    if( (start_time != 0) && ( (unsigned int)timeval < start_time) )
    {/* if dump beginning is truncated */
        ret = 1;
        //fprintf( stderr, PFX "start_time= %u  -- timeval= %d \n", (int)start_time, (int)timeval);
    }
    else if( (end_time != 0) && ( (unsigned int)timeval > end_time) )
    {/* if dump end is truncated */
        ret = 1;
        //fprintf( stderr, PFX "end_time= %u  -- timeval= %d \n", (int)end_time, (int)timeval);
    }

    if(ret == 0 )
    {
        oldtimeval = timeval;
        lt_set_time64( lt, timeval );
    }

    return ret;
}



static int get_pkts( int nb, uint32_t pkt[] )
{
    int i, ret = 0;

    for ( i = 0; i < nb; i++ ) {
        pkt[i] = GET_PKT();
        if ( pkt[i] == NO_PKT ) {
            ret = -1;
            break;
        }
    }

    return ret;
}


static uint32_t old_tick = 0;
static void process_pkt_tick(void)
{
    uint32_t pkt = GET_PKT();

    assert( pkt != NO_PKT );

    // update clock
    timer.ticks = (timer.ticks&0xFF000000) | RTMON_PKT_DATA24(pkt);
    if(  (RTMON_PKT_DATA24(pkt)<=0xF)
       &&(old_tick >= 0xFFFFF0) ) // just in case few ticks are missing
    {
        timer.ticks +=0x01000000;
    }
    old_tick    = RTMON_PKT_DATA24(pkt);
    if ( verbose ) {
        fprintf( stderr, PFX "      TICK: %08x\n", (unsigned int)timer.ticks );
    }
}


static void
process_pkt_isr_enter(void) {
    int vec;
    int trunc = 1;
    uint32_t pkt;
    struct lt_symbol *sym;

    pkt = GET_PKT();
    assert(( pkt != NO_PKT )&&
           ( RTMON_PKT_TYPE(pkt) == RTMON_PKT_ISR_ENTER ));

    vec = RTMON_PKT_DATA(pkt) & (RTMON_PKT_IRQ_MAX-1);

    trunc = emit_clock(pkt);
    if( trunc == 0 )
    {
        sym = isr_symbol(vec);
        lt_emit_value_bit_string( lt, sym, 0, "x" );
        stat_isr(   RTMON_PKT_TYPE(pkt),
                    vec,
                    timer.ticks+(pkt>>RTMON_PKT_FLAG_PEND_TICK),
                    RTMON_PKT_STAMP(pkt),
                    0);
        if ( verbose )
        {
            fprintf( stderr, PFX "%c%04x: %s (enter)\n",
                     ( RTMON_PEND_TICK(pkt) )? '*' : ' ',
                     (unsigned int)RTMON_PKT_STAMP(pkt),
                     sym->name);
        }
    }
}

static void
process_pkt_isr_exit(void) {
    int vec;
    int trunc = 1;
    int call_dsr;
    uint32_t pkt;
    struct lt_symbol *sym;

    pkt = GET_PKT();
    assert(( pkt != NO_PKT )&&
           ( RTMON_PKT_TYPE(pkt) == RTMON_PKT_ISR_EXIT ));

    call_dsr = (RTMON_PKT_DATA(pkt) & RTMON_ISR_CALLDSR_MASK) && RTMON_ISR_CALLDSR_MASK;
    vec     = RTMON_PKT_DATA(pkt) & (RTMON_PKT_IRQ_MAX-1);

    trunc = emit_clock(pkt);
    if( trunc == 0 )
    {
        sym = isr_symbol(vec);
        lt_emit_value_bit_string( lt, sym, 0, "z" );
        stat_isr(   RTMON_PKT_TYPE(pkt),
                    vec,
                    timer.ticks+(pkt>>RTMON_PKT_FLAG_PEND_TICK),
                    RTMON_PKT_STAMP(pkt),
                    call_dsr);

        if ( verbose ) {
            fprintf( stderr, PFX "%c%04x: %s (exit)\n",
                     ( RTMON_PEND_TICK(pkt) )? '*' : ' ',
                     (unsigned int)RTMON_PKT_STAMP(pkt),
                     sym->name);
        }
    }
}

static void
process_pkt_dsr_enter(void) {
    int vec;
    int trunc = 1;
    uint32_t pkt;
    struct lt_symbol *sym;

    pkt = GET_PKT();
    assert(( pkt != NO_PKT )&&
           ( RTMON_PKT_TYPE(pkt) == RTMON_PKT_DSR_ENTER ));

    vec = RTMON_PKT_DATA(pkt) & (RTMON_PKT_IRQ_MAX-1);

    trunc = emit_clock(pkt);
    if( trunc == 0 )
    {
        sym = dsr_symbol(vec);
        lt_emit_value_bit_string( lt, sym, 0, "x" );
        stat_dsr(   RTMON_PKT_TYPE(pkt),
                    vec,
                    timer.ticks+(pkt>>RTMON_PKT_FLAG_PEND_TICK),
                    RTMON_PKT_STAMP(pkt));
        if ( verbose )
        {
            fprintf( stderr, PFX "%c%04x: %s (enter)\n",
                     ( RTMON_PEND_TICK(pkt) )? '*' : ' ',
                     (unsigned int)RTMON_PKT_STAMP(pkt),
                     sym->name);
        }
    }
}

static void
process_pkt_dsr_exit(void) {
    int vec;
    int trunc = 1;
    uint32_t pkt;
    struct lt_symbol *sym;

    pkt = GET_PKT();
    assert(( pkt != NO_PKT )&&
           ( RTMON_PKT_TYPE(pkt) == RTMON_PKT_DSR_EXIT ));

    vec = RTMON_PKT_DATA(pkt) & (RTMON_PKT_IRQ_MAX-1);

    trunc = emit_clock(pkt);
    if( trunc == 0 )
    {
        sym = dsr_symbol(vec);
        lt_emit_value_bit_string( lt, sym, 0, "z" );
        stat_dsr(   RTMON_PKT_TYPE(pkt),
                    vec,
                    timer.ticks+ (pkt>>RTMON_PKT_FLAG_PEND_TICK),
                    RTMON_PKT_STAMP(pkt));
        if ( verbose ) {
            fprintf( stderr, PFX "%c%04x: %s (exit)\n",
                     ( RTMON_PEND_TICK(pkt) )? '*' : ' ',
                     (unsigned int)RTMON_PKT_STAMP(pkt),
                     sym->name);
        }
    }
}



static void process_thread_state(uint32_t * pkt, uint32_t id, uint32_t state, uint32_t reason)
{
   char str_state[25];
   char str_reason[10];

   int trunc = 1;
   struct lt_symbol *sym;

    if ( id >= THREAD_ID_MAX ) {
        fprintf( stderr, "Unsupported thread ID: %d\n", (int)id );
        return;
    }
    trunc = emit_clock(pkt[0]);

    if( trunc == 0 )
    {
      sym = thread_state_symbol(id, &(thd_state_current[id]));

      if( ((reason<<5)|state) != thd_state_current[id])
      {
           if(reason == NONE)          { sprintf(str_reason,"NONE"      );
           }else if(reason == WAIT)    { sprintf(str_reason,"WAIT"      );
           }else if(reason == DELAY)   { sprintf(str_reason,"DELAY"     );
           }else if(reason == TIMEOUT) { sprintf(str_reason,"TIMEOUT"   );
           }else if(reason == BREAK)   { sprintf(str_reason,"BREAK"     );
           }else if(reason == DESTRUCT){ sprintf(str_reason,"DESTRUCT"  );
           }else if(reason == EXIT)    { sprintf(str_reason,"EXIT"      );
           }else if(reason == DONE)    { sprintf(str_reason,"DONE"      );
           }else                       { sprintf(str_reason,"%d", reason);  }

           if(state == RUNNING)         { sprintf(str_state,"RUNNING[%s]",str_reason   );
           }else if(state == SLEEPING)  { sprintf(str_state,"SLEEPING[%s]",str_reason  );
           }else if(state == COUNTSLEEP){ sprintf(str_state,"COUNTSLEEP[%s]",str_reason);
           }else if(state == SUSPENDED) { sprintf(str_state,"SUSPENDED[%s]",str_reason );
           }else if(state == CREATING)  { sprintf(str_state,"CREATING[%s]",str_reason  );
           }else if(state == EXITED)    { sprintf(str_state,"EXITED[%s]",str_reason    );
           }else                        { sprintf(str_state,"%d[%d]",state,reason  ); }

         lt_emit_value_string( lt, sym, 0, str_state );
         thd_state_current[id] = ((reason<<5)|state);
      } // end if(state != thd_state_current[id])

     if ( verbose ) {
         fprintf( stderr, PFX "\t%s: %s - %d[%d]\n",
                  sym->name,str_state,state, reason);
      }
   }
}

/** Process Context Switch packets
 */
static void process_pkt_cs(uint32_t *pkt)
{

   static uint32_t current_id = 1;
   uint32_t id, priority, nextstate, nextreason, prevstate, prevreason;
   int trunc = 1;

   struct lt_symbol *sym;


   if ( verbose ){
      fprintf( stderr, PFX "pkt[1]=%08x: \n",(unsigned int)pkt[1] );
      fprintf( stderr, PFX "pkt[2]=%08x: \n",(unsigned int)pkt[2] );
   }

   id         =  RTMON_PKT_DATA24(pkt[1]) & 0xFFFF;
    priority   = (RTMON_PKT_DATA24(pkt[1]) >> 16) & 0xFF;
    prevstate  =  RTMON_PKT_DATA24(pkt[2]) & 0x1F;
   prevreason = (RTMON_PKT_DATA24(pkt[2]) >> 5 ) & 0x7;
    nextstate  = (RTMON_PKT_DATA24(pkt[2]) >> 8 ) & 0x1F;
   nextreason = (RTMON_PKT_DATA24(pkt[2]) >> 13) & 0x7;


   if ( id >= THREAD_ID_MAX ) {
        fprintf( stderr, "Unsupported thread ID: %d\n", (int)id );
        return;
   }

   trunc = emit_clock(pkt[0]);

   if(( trunc == 0 ) && (current_id != id))
   {
      stat_thread(  current_id,
                   (timer.ticks+(pkt[0]>>RTMON_PKT_FLAG_PEND_TICK)),
                   RTMON_PKT_STAMP(pkt[0]));

      //set previous thread as inactive
      sym = thread_symbol(current_id);
      lt_emit_value_bit_string( lt, sym, 0, "z" );

      //set previous thread as active
      sym = thread_symbol(id);
      lt_emit_value_bit_string( lt, sym, 0, "x" );

      //set current thread priority
      lt_emit_value_int( lt, current_prio_symbol(), 0, priority );

      if ( verbose ) {
         fprintf( stderr, PFX "%c%04x: => %s[id=%d, p=%d]\n",
                ( RTMON_PEND_TICK(pkt[0]) )? '*' : ' ',
                (unsigned int)RTMON_PKT_STAMP(pkt[0]), sym->name, id, priority);
      }

      //process thread state informations if it has been asked by the user
      if(process_thread_states)
      {
         stat_thread_activity(   (timer.ticks+(pkt[0]>>RTMON_PKT_FLAG_PEND_TICK)),
                                 RTMON_PKT_STAMP(pkt[0]),
                                 current_id, prevstate, prevreason,
                                 id        , nextstate, nextreason );
         process_thread_state(pkt, current_id, prevstate, prevreason);
         process_thread_state(pkt, id, nextstate, nextreason);
      }
   }
   /* memorize thread id */
   current_id = id;
}



/** Process Context Switch packets
 */
static void process_pkt_thdstate(uint32_t *pkt)
{
   uint32_t id, state, reason;
   char str_state[25];
   char str_reason[10];

   int trunc = 1;
   struct lt_symbol *sym;

   id     =  RTMON_PKT_DATA24(pkt[1]) & 0xFFFF;
    state  = (RTMON_PKT_DATA24(pkt[1]) >> 16) & 0x1F;
   reason = (RTMON_PKT_DATA24(pkt[1]) >> 21) & 0x7;


    if ( id >= THREAD_ID_MAX ) {
        fprintf( stderr, "Unsupported thread ID: %d\n", (int)id );
        return;
    }
    trunc = emit_clock(pkt[0]);

    if( trunc == 0 )
    {
      sym = thread_state_symbol(id, &(thd_state_current[id]));

      if(( (RTMON_PKT_DATA24(pkt[1]) >> 16) & 0xFF) != thd_state_current[id])
      {
           if(reason == NONE)          { sprintf(str_reason,"NONE"      );
           }else if(reason == WAIT)    { sprintf(str_reason,"WAIT"      );
           }else if(reason == DELAY)   { sprintf(str_reason,"DELAY"     );
           }else if(reason == TIMEOUT) { sprintf(str_reason,"TIMEOUT"   );
           }else if(reason == BREAK)   { sprintf(str_reason,"BREAK"     );
           }else if(reason == DESTRUCT){ sprintf(str_reason,"DESTRUCT"  );
           }else if(reason == EXIT)    { sprintf(str_reason,"EXIT"      );
           }else if(reason == DONE)    { sprintf(str_reason,"DONE"      );
           }else                       { sprintf(str_reason,"%d", reason);  }

           if(state == RUNNING)         { sprintf(str_state,"RUNNING[%s]",str_reason   );
           }else if(state == SLEEPING)  { sprintf(str_state,"SLEEPING[%s]",str_reason  );
           }else if(state == COUNTSLEEP){ sprintf(str_state,"COUNTSLEEP[%s]",str_reason);
           }else if(state == SUSPENDED) { sprintf(str_state,"SUSPENDED[%s]",str_reason );
           }else if(state == CREATING)  { sprintf(str_state,"CREATING[%s]",str_reason  );
           }else if(state == EXITED)    { sprintf(str_state,"EXITED[%s]",str_reason    ); }

         lt_emit_value_string( lt, sym, 0, str_state );
         thd_state_current[id] = (RTMON_PKT_DATA24(pkt[1]) >> 16) & 0xFF;
      } // end if(state != thd_state_current[id])

      if ( verbose ) {
         fprintf( stderr, PFX "%c%04x: STATE %s, s=%d r=%d\n",
                     ( RTMON_PEND_TICK(pkt[0]) )? '*' : ' ',
                     (unsigned int)RTMON_PKT_STAMP(pkt[0]),
                     sym->name, state, reason);
      }
   }
}




/** Process Priority change packets
 */
static void process_pkt_priority_change(uint32_t subtype, uint32_t *pkt)
{
    uint32_t priority;
    int trunc = 1;

    priority = RTMON_PKT_DATA24(pkt[1]) & 0xFF;

    trunc = emit_clock(pkt[0]);

    if( trunc == 0 )
    {
        if (subtype == RTMON_THD_CHGPRIO_ENTER){
            lt_emit_value_bit_string( lt, thd_prio_chg_symbol(), 0, "x" );
            stat_priority_inversion_start( (timer.ticks+(pkt[0]>>RTMON_PKT_FLAG_PEND_TICK)),
                                  RTMON_PKT_STAMP(pkt[0]) );

            if ( verbose ) {
            fprintf( stderr, PFX "%c%04x: THD PRIORITY CHANGE(enter) p=%d\n",
                     ( RTMON_PEND_TICK(pkt[0]) )? '*' : ' ',
                     (unsigned int)RTMON_PKT_STAMP(pkt[0]),priority);
            }
        }else{
            lt_emit_value_bit_string( lt, thd_prio_chg_symbol(), 0, "z" );
            stat_priority_inversion_stop( (timer.ticks+(pkt[0]>>RTMON_PKT_FLAG_PEND_TICK)),
                                  RTMON_PKT_STAMP(pkt[0]) );
            if ( verbose ) {
            fprintf( stderr, PFX "%c%04x: THD PRIORITY CHANGE(exit) p=%d\n",
                     ( RTMON_PEND_TICK(pkt[0]) )? '*' : ' ',
                     (unsigned int)RTMON_PKT_STAMP(pkt[0]),priority);
            }
        }
    }
}


/** Process Thread packets
 */
static void process_pkt_thread(void)
{
    uint32_t subtype;
    uint32_t pkt[3];

    /* First packet */
    pkt[0] = GET_PKT();
    assert(( pkt[0] != NO_PKT )&&
           ( RTMON_PKT_TYPE(pkt[0]) == RTMON_PKT_THREAD ));

    /* Second packet */
    pkt[1] = GET_PKT();

    subtype = RTMON_PKT_DATA(pkt[0]);

    switch(subtype)
    {
    case RTMON_THD_CS:
      pkt[2] = GET_PKT();
        process_pkt_cs(pkt);
        break;
    case RTMON_THD_CHGPRIO_ENTER:
    case RTMON_THD_CHGPRIO_EXIT:
        process_pkt_priority_change(subtype, pkt);
        break;
    case RTMON_THD_STATE:
        process_pkt_thdstate(pkt);
        break;
    default:
        fprintf( stderr, "Unknown thread sub id: 0x%08x\n", pkt[0] );
        break;

    }
}





/** Process packets log caller
 */
static void process_pkt_caller(void)
{
    uint32_t pkt[2];
    uint32_t func, type;
    char *tag = NULL;
    int ret = -1;
    int trunc = 1;

    if ( get_pkts( 2, pkt ) == 0 )
    {
        type = RTMON_PKT_TYPE(pkt[0]);

        if (( type == RTMON_PKT_CALLER )&&
            ( type < RTMON_PKT_MAX )    &&
            ( RTMON_PKT_TYPE(pkt[1]) == RTMON_PKT_DATA24 ))
        {
            trunc = emit_clock(pkt[0]);
            if( trunc == 0 )
            {
                // caller function
                func = (RTMON_PKT_DATA(pkt[0]) << 24)|RTMON_PKT_DATA24(pkt[1]);

                if ( exefile )
                {
                    // display source information
                    tag = get_srctag(func);
                    lt_emit_value_string( lt, trace_sfunc_symbol(), 0, tag );
                }
                else
                {   // just display hex address
                    lt_emit_value_int( lt, trace_func_symbol(), 0, func );
                }
                if ( verbose )
                {
                    fprintf( stderr,
                             PFX "%c%04x: CALLER, func = 0x%08x '%s'\n",
                             ( RTMON_PEND_TICK(pkt[0]) )? '*' : ' ',
                             (unsigned int)RTMON_PKT_STAMP(pkt[0]),
                             (unsigned int)func, tag );
                }
            }
            ret = 0;
        }
    }
    if ( ret != 0 ) {
        fprintf( stderr, PFX "bad trace caller packet @ %d\n", (int)buf_index );
    }
}


/** Process pc packets
 */
static void process_pkt_pc(void)
{
    uint32_t pkt[2];
    uint32_t func, type;
    char     *tag   = NULL;
    int      ret    = -1;
    int      trunc  = 1;

    if ( get_pkts( 2, pkt ) == 0 )
    {
        type = RTMON_PKT_TYPE(pkt[0]);

        if (( type == RTMON_PKT_PC )    &&
            ( RTMON_PKT_TYPE(pkt[1]) == RTMON_PKT_DATA24 ))
        {
            trunc = emit_clock(pkt[0]);
            if( trunc == 0 )
            {
                // caller function
                func = (RTMON_PKT_DATA(pkt[0]) << 24)|RTMON_PKT_DATA24(pkt[1]);

                if ( exefile )
                {
                    // display source information
                    tag = get_srctag(func);
                    lt_emit_value_string( lt, pc_sfunc_symbol(), 0, tag );
                }
                else
                {   // just display hex address
                    lt_emit_value_int( lt, pc_func_symbol(), 0, func );
                }
                if ( verbose )
                {
                    fprintf( stderr,
                             PFX "%c%04x: PC    , func = 0x%08x '%s'\n",
                             ( RTMON_PEND_TICK(pkt[0]) )? '*' : ' ',
                             (unsigned int)RTMON_PKT_STAMP(pkt[0]),
                             (unsigned int)func, tag );
                }
            }
            ret = 0;
        }
    }
    if ( ret != 0 ) {
        fprintf( stderr, PFX "bad trace caller packet @ %d\n", (int)buf_index );
    }
}



/** Log Profiling call
 */
static void profile_log_call(uint32_t pc, char * tag)
{
    char     tag_file[255];
    char     tag_function[255];
    char     tag_line[255];
    uint32_t idx_tag = 0;
    uint32_t idx;

    idx = 0;
    do{
        tag_function[idx] = tag[idx_tag];
        idx_tag++;
        idx++;
    }while( (tag[idx_tag] !='(')||(tag[idx_tag+1] !=')') );
    tag_function[idx]= '\0';

    idx = 0;
    do{idx_tag++;}while(tag[idx_tag-1] !='[');
    do{
        tag_file[idx] = tag[idx_tag];
        idx_tag++;
        idx++;
    }while( tag[idx_tag] !=':' );
    tag_file[idx]= '\0';

    idx = 0;
    idx_tag +=1; // jump over ':'
    do{
        tag_line[idx] = tag[idx_tag];
        idx_tag++;
        idx++;
    }while( tag[idx_tag] !=']' );
    tag_line[idx]= '\0';

    //fprintf( stderr, PFX "addr=0x%08x tag-------- '%s'\n",pc, tag );
    //fprintf( stderr, PFX "fct='%s' - file='%s' - line='%s'\n",
    //                      tag_function, tag_file, tag_line );
    fprintf( prof_file, "\nfl=%s\n", tag_file );
    fprintf( prof_file, "fn=%s\n", tag_function );
    fprintf( prof_file, "0x%08x %s 50\n",(unsigned int)pc, tag_line );
}



//! memorize previous symbol's tag
static char tag_previous[256];

/** Process Profile log packets
 */
static void process_pkt_profile(void)
{
    uint32_t pkt[2];
    uint32_t func, type;
    char *tag = NULL;
    int ret = -1;
    int trunc = 1;

    if ( get_pkts( 2, pkt ) == 0 )
    {
        type = RTMON_PKT_TYPE(pkt[0]);

        if (( type == RTMON_PKT_PROFILE )&&
            ( RTMON_PKT_TYPE(pkt[1]) == RTMON_PKT_DATA24 ))
        {
            // caller function
            func = (RTMON_PKT_DATA(pkt[0]) << 24)|RTMON_PKT_DATA24(pkt[1]);

            trunc = emit_clock(pkt[0]);
            if( trunc == 0 )
            {
                if ( exefile )
                {
                    // display source information
                    tag = get_srctag(func);
                    profile_log_call(func, tag);

                    if(strcmp(tag_previous, tag) != 0)
                    {// if this tag is different from the previous
                        lt_emit_value_string( lt, prof_sfunc_symbol(), 0, tag );
                        strcpy(tag_previous, tag);
                    }
                }
                else
                {   // just display hex address
                    lt_emit_value_int( lt, prof_func_symbol(), 0, func );
                }
                if ( verbose && ( trunc == 0 ))
                {
                    fprintf( stderr,
                             PFX "%c%04x: PROFIL, func = 0x%08x '%s'\n",
                             ( RTMON_PEND_TICK(pkt[0]) )? '*' : ' ',
                             (unsigned int)RTMON_PKT_STAMP(pkt[0]),
                             (unsigned int)func, tag );
                }
            }
            ret = 0;
        }
    }
    if ( ret != 0 ) {
        fprintf( stderr, PFX "bad trace profile packet @ %d\n", (int)buf_index );
    }
}



static void process_pkt_slock(void)
{
    uint32_t pkt[2];
    uint32_t func, type;
    char *tag;
    int slock, ret = -1;
    int trunc = 1;

    if ( get_pkts( 2, pkt ) == 0 ) {

        type = RTMON_PKT_TYPE(pkt[0]);

        if (( type >= RTMON_PKT_SLOCK_0 )&&
            ( type < RTMON_PKT_MAX )&&
            ( RTMON_PKT_TYPE(pkt[1]) == RTMON_PKT_DATA24 ))
        {
            trunc = emit_clock(pkt[0]);
            if( trunc == 0 )
            {
                // scheduler counter value
                slock = type-RTMON_PKT_SLOCK_0;

                // scheduler locked/unlocked
                if ( slock == 0 ) {
                    lt_emit_value_bit_string( lt, schedbit_symbol(), 0, "z" );
                }
                else {
                    lt_emit_value_bit_string( lt, schedbit_symbol(), 0, "x" );
                }
                // counter
                lt_emit_value_int( lt, sched_symbol(), 0, slock );

                // caller function
                func = (RTMON_PKT_DATA(pkt[0]) << 24)|RTMON_PKT_DATA24(pkt[1]);

                if ( exefile ) {
                    // display source information
                    tag = get_srctag(func);
                    lt_emit_value_string( lt, sched_sfunc_symbol(), 0, tag );
                }
                else {
                    // just display hex address
                    lt_emit_value_int( lt, sched_func_symbol(), 0, func );
                }
                if ( verbose ) {
                    fprintf( stderr,
                             PFX "%c%04x: SLOCK slock = %d, func = 0x%08x\n",
                             ( RTMON_PEND_TICK(pkt[0]) )? '*' : ' ',
                             (unsigned int)RTMON_PKT_STAMP(pkt[0]), slock,
                             (unsigned int)func );
                }
            }
            ret = 0;
        }
    }
    if ( ret != 0 ) {
        fprintf( stderr, PFX "bad sched lock packet @ %d\n", (int)buf_index );
    }
}

static void process_pkt_user_u32(void)
{
    uint32_t pkt[2];
    uint32_t val, type;
    int uval, ret = -1;
    int trunc = 1;
    struct lt_symbol *sym;

    if ( get_pkts( 2, pkt ) == 0 ) {

        type = RTMON_PKT_TYPE(pkt[0]);

        if (( type >= RTMON_PKT_USER_U32 )&&
            ( type < RTMON_PKT_USER_U32+RTMON_PKT_USER_U32_MAX )&&
            ( RTMON_PKT_TYPE(pkt[1]) == RTMON_PKT_DATA24 ))
        {
            trunc = emit_clock(pkt[0]);
            if( trunc == 0 )
            {
                // user value tag
                uval = type-RTMON_PKT_USER_U32;
                // user value
                val = (RTMON_PKT_DATA(pkt[0]) << 24)|RTMON_PKT_DATA24(pkt[1]);

                sym = user_val_symbol(uval);
                lt_emit_value_int( lt, sym, 0, val );

                if ( verbose ) {
                    fprintf( stderr, PFX "%c%04x: USER VAL #%d, val = 0x%08x\n",
                             ( RTMON_PEND_TICK(pkt[0]) )? '*' : ' ',
                             (unsigned int)RTMON_PKT_STAMP(pkt[0]), uval,
                             (unsigned int)val );
                }
            }
            ret = 0;
        }
    }
    if ( ret != 0 ) {
        fprintf( stderr, PFX "bad user value packet @ %d\n", (int)buf_index );
    }
}

/** This function process packets log of the AAI ITS register
 */
static void process_pkt_aai_reg_its(void)
{
    uint32_t pkt[2];
    uint32_t val, type;
    int      ret = -1;
    int trunc = 1;

    if ( get_pkts( 2, pkt ) == 0 )
    {
        type = RTMON_PKT_TYPE(pkt[0]);

        if (  ( type == RTMON_PKT_AAI_REG_ITS )
            &&( RTMON_PKT_TYPE(pkt[1]) == RTMON_PKT_DATA24 ))
        {
            // user value
            val = (RTMON_PKT_DATA(pkt[0]) << 24)|RTMON_PKT_DATA24(pkt[1]);
            trunc = emit_clock(pkt[0]);
            if( trunc == 0 )
            {
                lt_emit_value_int( lt, aai_reg_its_symbol(), 0, val );

                if ( verbose ) {
                    fprintf( stderr, PFX "%c%04x: AAI REGG ITS = 0x%08x\n",
                             ( RTMON_PEND_TICK(pkt[0]) )? '*' : ' ',
                             (unsigned int)RTMON_PKT_STAMP(pkt[0]),
                             (unsigned int)val );
                }
            }
            ret = 0;
        }
    }
    if ( ret != 0 ) {
        fprintf( stderr, PFX "bad aai reg its packet @ %d\n", (int)buf_index );
    }
}

/** This function process packets log of the AAI ITEN register
 */
static void process_pkt_aai_reg_iten(void)
{
    uint32_t pkt[2];
    uint32_t val, type;
    int      ret = -1;
    int trunc = 1;

    if ( get_pkts( 2, pkt ) == 0 )
    {
        type = RTMON_PKT_TYPE(pkt[0]);

        if (  ( type == RTMON_PKT_AAI_REG_ITEN )
            &&( RTMON_PKT_TYPE(pkt[1]) == RTMON_PKT_DATA24 ))
        {
            // user value
            val = (RTMON_PKT_DATA(pkt[0]) << 24)|RTMON_PKT_DATA24(pkt[1]);
            trunc = emit_clock(pkt[0]);
            if( trunc == 0 )
            {
                lt_emit_value_int( lt, aai_reg_iten_symbol(), 0, val );

                if ( verbose ) {
                    fprintf( stderr, PFX "%c%04x: AAI REG ITEN = 0x%08x\n",
                             ( RTMON_PEND_TICK(pkt[0]) )? '*' : ' ',
                             (unsigned int)RTMON_PKT_STAMP(pkt[0]),
                             (unsigned int)val );
                }
            }
            ret = 0;
        }
    }
    if ( ret != 0 ) {
        fprintf( stderr, PFX "bad aai reg iten packet @ %d\n", (int)buf_index );
    }
}

/** This function process packets log of the AAI CFG register
 */
static void process_pkt_aai_reg_cfg(void)
{
    uint32_t pkt[2];
    uint32_t val, type;
    int      ret = -1;
    int trunc = 1;

    if ( get_pkts( 2, pkt ) == 0 )
    {
        type = RTMON_PKT_TYPE(pkt[0]);

        if (  ( type == RTMON_PKT_AAI_REG_CFG )
            &&( RTMON_PKT_TYPE(pkt[1]) == RTMON_PKT_DATA24 ))
        {
            // user value
            val = (RTMON_PKT_DATA(pkt[0]) << 24)|RTMON_PKT_DATA24(pkt[1]);
            trunc = emit_clock(pkt[0]);
            if( trunc == 0 )
            {
                lt_emit_value_int( lt, aai_reg_cfg_symbol(), 0, val );

                if ( verbose ) {
                    fprintf( stderr, PFX "%c%04x: AAI REG CFG = 0x%08x\n",
                             ( RTMON_PEND_TICK(pkt[0]) )? '*' : ' ',
                             (unsigned int)RTMON_PKT_STAMP(pkt[0]),
                             (unsigned int)val );
                }
            }
            ret = 0;
        }
    }
    if ( ret != 0 ) {
        fprintf( stderr, PFX "bad aai reg cfg packet @ %d\n", (int)buf_index );
    }
}



/** Process AAI log PC
 */
static void process_pkt_aai_pc(void)
{
    uint32_t pkt[2];
    uint32_t func, type;
    char *tag = NULL;
    int ret = -1;
    int trunc = 1;

    if ( get_pkts( 2, pkt ) == 0 )
    {
        type = RTMON_PKT_TYPE(pkt[0]);

        if (( type == RTMON_PKT_AAI_PC )&&
            ( RTMON_PKT_TYPE(pkt[1]) == RTMON_PKT_DATA24 ))
        {
            // caller function
            func = (RTMON_PKT_DATA(pkt[0]) << 24)|RTMON_PKT_DATA24(pkt[1]);

            trunc = emit_clock(pkt[0]);
            if( trunc == 0 )
            {
                if ( exefile )
                {
                    // display source information
                    tag = get_srctag(func);
                    lt_emit_value_string( lt, aai_pc_sfunc_symbol(), 0, tag );
                }
                else
                {   // just display hex address
                    lt_emit_value_int( lt, aai_pc_func_symbol(), 0, func );
                }
                if ( verbose )
                {
                    fprintf( stderr,
                             PFX "%c%04x: AAI PC, func = 0x%08x '%s'\n",
                             ( RTMON_PEND_TICK(pkt[0]) )? '*' : ' ',
                             (unsigned int)RTMON_PKT_STAMP(pkt[0]),
                             (unsigned int)func, tag );
                }
            }
            ret = 0;
        }
    }
    if ( ret != 0 ) {
        fprintf( stderr, PFX "bad trace aai pc packet @ %d\n", (int)buf_index );
    }
}

/** Process User Event Start packets
 */
static void process_pkt_user_start(void) {
    int vec;
    uint32_t pkt;
    struct lt_symbol *sym;
    int trunc = 1;

    pkt = GET_PKT();
    assert(( pkt != NO_PKT )&&
           ( RTMON_PKT_TYPE(pkt) == RTMON_PKT_USER0_START ));

    vec = RTMON_PKT_DATA(pkt);

    trunc = emit_clock(pkt);
    if( trunc == 0 )
    {
        sym = user_symbol(vec);
        stat_log_user_start( vec, (timer.ticks+(pkt>>RTMON_PKT_FLAG_PEND_TICK)), RTMON_PKT_STAMP(pkt) );

        lt_emit_value_bit_string( lt, sym, 0, "x" );
        if ( verbose ) {
            fprintf( stderr, PFX "%c%04x: USER #%d (start)\n",
                     ( RTMON_PEND_TICK(pkt) )? '*' : ' ',
                     (unsigned int)RTMON_PKT_STAMP(pkt),
                     vec );
        }
    }
}

/** Process User Event Stop packets
 */
static void process_pkt_user_stop(void)
{
    int vec;
    uint32_t pkt;
    struct lt_symbol *sym;
    int trunc = 1;

    pkt = GET_PKT();
    assert(( pkt != NO_PKT )&&
           ( RTMON_PKT_TYPE(pkt) == RTMON_PKT_USER0_STOP ));

    vec = RTMON_PKT_DATA(pkt);

    trunc = emit_clock(pkt);
    if( trunc == 0 )
    {
        sym = user_symbol(vec);
        stat_log_user_stop( vec, (timer.ticks+(pkt>>RTMON_PKT_FLAG_PEND_TICK)), RTMON_PKT_STAMP(pkt) );

        lt_emit_value_bit_string( lt, sym, 0, "z" );
        if ( verbose ) {
            fprintf( stderr, PFX "%c%04x: USER #%d (stop)\n",
                     ( RTMON_PEND_TICK(pkt) )? '*' : ' ',
                     (unsigned int)RTMON_PKT_STAMP(pkt),
                     vec );
        }
    }
}


static void process_pkt_driver_event_start(void) {
    int vec;
    uint32_t pkt;
    struct lt_symbol *sym;
    int trunc = 1;

    pkt = GET_PKT();
    assert(( pkt != NO_PKT )&&
           ( RTMON_PKT_TYPE(pkt) == RTMON_PKT_DRV_EVENT_START ));

    vec = RTMON_PKT_DATA(pkt) & (RTMON_PKT_EVENT_MAX-1);

    trunc = emit_clock(pkt);
    if( trunc == 0 )
    {
        sym = driver_symbol(vec);
        //stat_log_user_start( vec, (timer.ticks+(pkt>>RTMON_PKT_FLAG_PEND_TICK)), RTMON_PKT_STAMP(pkt) );

        lt_emit_value_bit_string( lt, sym, 0, "x" );
        if ( verbose ) {
            fprintf( stderr, PFX "%c%04x: DRIVER #%d (start)\n",
                     ( RTMON_PEND_TICK(pkt) )? '*' : ' ',
                     (unsigned int)RTMON_PKT_STAMP(pkt),
                     vec );
        }
    }
}

static void process_pkt_driver_event_stop(void)
{
    int vec;
    uint32_t pkt;
    struct lt_symbol *sym;
    int trunc = 1;

    pkt = GET_PKT();
    assert(( pkt != NO_PKT )&&
           ( RTMON_PKT_TYPE(pkt) == RTMON_PKT_DRV_EVENT_STOP ));

    vec = RTMON_PKT_DATA(pkt) & (RTMON_PKT_EVENT_MAX-1);

    trunc = emit_clock(pkt);
    if( trunc == 0 )
    {
        sym = driver_symbol(vec);
        //stat_log_user_stop( vec, (timer.ticks+(pkt>>RTMON_PKT_FLAG_PEND_TICK)), RTMON_PKT_STAMP(pkt) );

        lt_emit_value_bit_string( lt, sym, 0, "z" );
        if ( verbose ) {
            fprintf( stderr, PFX "%c%04x: DRIVER #%d (stop)\n",
                     ( RTMON_PEND_TICK(pkt) )? '*' : ' ',
                     (unsigned int)RTMON_PKT_STAMP(pkt),
                     vec );
        }
    }
}


/** Process packets of RTMON dump file
 *
 */
int process_pkt(void)
{
    uint32_t pkt;
    int type, ret = -1;

    assert(lt);

    /* Read next data packet */
    pkt = PEEK_PKT();

    if ( pkt != NO_PKT )
    {
        /* extract packet type from data */
        type = RTMON_PKT_TYPE(pkt);

        if ( verbose )
            fprintf( stderr, PFX "packet=%08x: \n",(unsigned int)pkt );

        switch (type)
        {
        case RTMON_PKT_NONE:// skip empty packets
            GET_PKT();
            break;

        case RTMON_PKT_ISR_ENTER:
            process_pkt_isr_enter();
            break;

        case RTMON_PKT_ISR_EXIT:
            process_pkt_isr_exit();
            break;

        case RTMON_PKT_DSR_ENTER:
            process_pkt_dsr_enter();
            break;

        case RTMON_PKT_DSR_EXIT:
            process_pkt_dsr_exit();
            break;

        case RTMON_PKT_DRV_EVENT_START:
            process_pkt_driver_event_start();
            break;

        case RTMON_PKT_DRV_EVENT_STOP:
            process_pkt_driver_event_stop();
            break;

        case RTMON_PKT_USER0_START:
            process_pkt_user_start();
            break;

        case RTMON_PKT_USER0_STOP:
            process_pkt_user_stop();
            break;

        case RTMON_PKT_THREAD:
            process_pkt_thread();
            break;

        case RTMON_PKT_CALLER:
            process_pkt_caller();
            break;

        case RTMON_PKT_PC:
            process_pkt_pc();
            break;

        case RTMON_PKT_PROFILE:
            process_pkt_profile();
            break;

        case RTMON_PKT_AAI_REG_ITS:
            process_pkt_aai_reg_its();
            break;

        case RTMON_PKT_AAI_REG_ITEN:
            process_pkt_aai_reg_iten();
            break;

        case RTMON_PKT_AAI_REG_CFG:
            process_pkt_aai_reg_cfg();
            break;

        case RTMON_PKT_AAI_PC:
            process_pkt_aai_pc();
            break;

        case RTMON_PKT_TICK:
            process_pkt_tick();
            break;

        case RTMON_PKT_DATA24:// ignore orphan data packets
            GET_PKT();
            break;

        default:

            // user value ?
            if (( type >= RTMON_PKT_USER_U32 )&&
                ( type < RTMON_PKT_USER_U32+RTMON_PKT_USER_U32_MAX ))
            {
                process_pkt_user_u32();
            }
            else
            {
                // or scheduler lock value ?
                if (( type >= RTMON_PKT_SLOCK_0 )&&
                    ( type < RTMON_PKT_MAX ))
                {
                    process_pkt_slock();
                }
                else
                {
                    // skip bad packets
                    fprintf( stderr, PFX "bad packet type: %d\n", type);
                    GET_PKT();
                }
            }
            break;
        }// end switch
        ret = 0;
    }
    return ret;
}






void init_savefile( void )
{
    memset( isr_lts     , 0, RTMON_PKT_IRQ_MAX      * sizeof(struct lt_symbol *));
    memset( dsr_lts     , 0, RTMON_PKT_IRQ_MAX      * sizeof(struct lt_symbol *));
    memset( thread_lts  , 0, THREAD_ID_MAX          * sizeof(struct lt_symbol *));
    memset( thread_state_lts  , 0, THREAD_ID_MAX    * sizeof(struct lt_symbol *));
    memset( user_lts    , 0, RTMON_PKT_EVENT_MAX    * sizeof(struct lt_symbol *));
    memset( user_val_lts, 0, RTMON_PKT_USER_U32_MAX * sizeof(struct lt_symbol *));
    memset( driver_lts  , 0, RTMON_PKT_EVENT_MAX    * sizeof(struct lt_symbol *));
}

void write_savefile(const char *name)
{
   int i;
   FILE *fp;

   fp = fopen( name, "wb" );
   if ( fp != NULL ) {

      // interrupt traces
      fprintf( fp, "@%x\n-Interrupt routines\n@%x\n",
               TR_BLANK, TR_BIN|TR_RJUSTIFY );

      for ( i = 0; i < RTMON_PKT_IRQ_MAX; i++ ) {
         if ( isr_lts[i] ) {
               fprintf( fp, "%s\n", isr_lts[i]->name );
         }
         if ( dsr_lts[i] ) {
               fprintf( fp, "%s\n", dsr_lts[i]->name );
         }
      }

      // scheduler locking/unlocking
      if ( schedbit_lts || sched_lts || sched_sfunc_lts || sched_func_lts)
      fprintf( fp, "@%x\n-Scheduler\n@%x\n", TR_BLANK, TR_BIN|TR_RJUSTIFY );

      if ( schedbit_lts ) {
         fprintf( fp, "%s\n", schedbit_lts->name );
      }

      if ( sched_lts ) {
         fprintf( fp, "@%x\n", TR_ANALOG_STEP|TR_RJUSTIFY );
         fprintf( fp, "%s\n", sched_lts->name );
      }

      if ( exefile ) {
         if ( sched_sfunc_lts )
         {
               fprintf( fp, "@%x\n", TR_ASCII|TR_RJUSTIFY );
               fprintf( fp, "%s\n", sched_sfunc_lts->name );
         }
      }
      else {
         if ( sched_func_lts ) {
               fprintf( fp, "@%x\n", TR_HEX|TR_RJUSTIFY );
               fprintf( fp, "%s\n", sched_func_lts->name );
         }
      }

      if(prof_sfunc_lts || prof_func_lts ||
         trace_sfunc_lts|| trace_func_lts||
         pc_sfunc_lts   || pc_sfunc_lts     )
      {
         fprintf( fp, "@%x\n", TR_BIN|TR_RJUSTIFY );
         fprintf( fp, "@%x\n-Trace\n@%x\n", TR_BLANK, TR_BIN|TR_RJUSTIFY );
      }
      if ( exefile )
      {
         if ( prof_sfunc_lts ) {
               fprintf( fp, "@%x\n", TR_ASCII|TR_RJUSTIFY );
               fprintf( fp, "%s\n", prof_sfunc_lts->name );
         }
         if ( trace_sfunc_lts ){
               fprintf( fp, "@%x\n", TR_ASCII|TR_RJUSTIFY );
               fprintf( fp, "%s\n", trace_sfunc_lts->name );
         }
         if ( pc_sfunc_lts ) {
               fprintf( fp, "@%x\n", TR_ASCII|TR_RJUSTIFY );
               fprintf( fp, "%s\n", pc_sfunc_lts->name );
         }
      }
      else {
         if ( prof_func_lts ) {
               fprintf( fp, "@%x\n", TR_HEX|TR_RJUSTIFY );
               fprintf( fp, "%s\n", prof_func_lts->name );
         }
         if ( trace_func_lts ) {
               fprintf( fp, "@%x\n", TR_HEX|TR_RJUSTIFY );
               fprintf( fp, "%s\n", trace_func_lts->name );
               }
         if ( pc_func_lts ) {
               fprintf( fp, "@%x\n", TR_HEX|TR_RJUSTIFY );
               fprintf( fp, "%s\n", pc_func_lts->name );
         }
      }

      /** AAI Module Traces
      */
      if(aai_reg_its_lts || aai_reg_iten_lts || aai_reg_cfg_lts ||
         aai_pc_sfunc_lts|| aai_pc_sfunc_lts   ) {
         fprintf( fp, "@%x\n", TR_BIN|TR_RJUSTIFY );
         fprintf( fp, "@%x\n-AAI\n@%x\n", TR_BLANK, TR_BIN|TR_RJUSTIFY );
      }
      if ( aai_reg_its_lts ) {
         fprintf( fp, "@%x\n", TR_HEX|TR_RJUSTIFY );
         fprintf( fp, "%s\n", aai_reg_its_lts->name );
      }
      if ( aai_reg_cfg_lts ) {
         fprintf( fp, "@%x\n", TR_HEX|TR_RJUSTIFY );
         fprintf( fp, "%s\n", aai_reg_cfg_lts->name );
      }
      if ( aai_reg_iten_lts ) {
         fprintf( fp, "@%x\n", TR_HEX|TR_RJUSTIFY );
         fprintf( fp, "%s\n", aai_reg_iten_lts->name );
      }
      if ( exefile ) {
         if ( aai_pc_sfunc_lts ) {
               fprintf( fp, "@%x\n", TR_ASCII|TR_RJUSTIFY );
               fprintf( fp, "%s\n", aai_pc_sfunc_lts->name );
         }
      }
      else {
         if ( aai_pc_func_lts ) {
               fprintf( fp, "@%x\n", TR_HEX|TR_RJUSTIFY );
               fprintf( fp, "%s\n", aai_pc_func_lts->name );
         }
      }


      /** thread context switches
      */
      fprintf( fp, "@%x\n-\n-eCos Threads\n@%x\n",
               TR_BLANK, TR_BIN|TR_RJUSTIFY );

    if ( thd_prio_chg_lts ){
         fprintf( fp, "@%x\n", TR_ANALOG_STEP|TR_RJUSTIFY );
         fprintf( fp, "%s\n", thd_prio_chg_lts->name );
      }

    if ( current_thd_prio_lts ){
         fprintf( fp, "@%x\n", TR_DEC|TR_RJUSTIFY );
         fprintf( fp, "%s\n", current_thd_prio_lts->name );
      }

      for ( i = 0; i < THREAD_ID_MAX; i++ ) {
         if ( thread_lts[i] ) {
               fprintf( fp, "%s\n", thread_lts[i]->name );
         }
         if ( thread_state_lts[i] ) {
               fprintf( fp, "%s\n", thread_state_lts[i]->name );
         }
      }


      /** then output driver events
      */
      fprintf( fp, "@%x\n-\n-Driver events\n@%x\n",
               TR_BLANK, TR_BIN|TR_RJUSTIFY );

      for ( i = 0; i < RTMON_PKT_EVENT_MAX; i++ ) {
         if ( driver_lts[i] ) {
               fprintf( fp, "%s\n", driver_lts[i]->name );
         }
      }

      /** then output user events
      */
      fprintf( fp, "@%x\n-\n-User events\n@%x\n",
               TR_BLANK, TR_BIN|TR_RJUSTIFY );

      for ( i = 0; i < RTMON_PKT_EVENT_MAX; i++ ) {
         if ( user_lts[i] ) {
               fprintf( fp, "%s\n", user_lts[i]->name );
         }
      }

      for ( i = 0; i < RTMON_PKT_USER_U32_MAX; i++ ) {
         if ( user_val_lts[i] ) {
               fprintf( fp, "@%x\n", TR_HEX|TR_RJUSTIFY );
               fprintf( fp, "%s\n", user_val_lts[i]->name );
         }
      }

      fclose(fp);
   }
}





