//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================

/**
*******************************************************************************
* @file  stat.c
* @brief Real-Time Monitor host tools
*
* This functions are used to generate a statistic report upon RTMON traces
*
* Copyright (C) 2006 Parrot S.A.
*
* @author     gregoire.etienne@parrot.com
* @date       2007-07-04
*
* @author     leonard.bacinello.ext@parrot.com
* @date       2008-07-07
*******************************************************************************
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <errno.h>
#include <assert.h>
#include <stdarg.h>

#define _GNU_SOURCE
#include <getopt.h>

#include <search.h>

#include "lxt_write.h"

#include "ecos_types.h"
#include "rtmon_common.h"
#include "posix_crc.h"
#include "rtd2lxt.h"

#define PFX             "rtd2lxt-stat: "

#define UNAVAILABLE 0xFFFFFFFF	//! Uninitialized value

#define RTMON_IRQ_MAX   (32)    //! Maximum numbers of IRQ

#define ISR_DUR_OK      (8)     //! Duration limit 1 for an ISR(�s)
#define ISR_DUR_WARNING (17)    //! Duration limit 2 for an ISR(�s)

#define DSR_DUR_OK      (50)    //! Duration limit 1 for a DSR(�s)
#define DSR_DUR_WARNING	(110)   //! Duration limit 2 for a DSR(�s)

#define BACKGD_COLOR	   0xFFFFFF
#define COLOR_OK		   0xF6DC12
#define COLOR_WARNING	0xED7F10
#define COLOR_DANGER	   0xFF0000
#define THREAD_COLOR	   0x33FF33
#define CUMUL_COLOR	   0x00FF33


#define FPRINT_CRITIC_TABLE(_col_ok_, _dur_ok_, _col_warn_, _dur_warn_, _col_danger_)  \
    fprintf( fp, "<table border=1 cellspacing=2 class=\"critic\">\n"); \
    fprintf( fp, "<tr><TD width=100 align=left><b>Level:</b></TD><TD width=100 align=center >OK</TD><TD width=100 align=center >WARN</TD><TD width=100 align=center >CRITICAL</TD></tr>\n" );   \
    fprintf( fp, "<tr><TD width=100 align=left><b>Duration:</b></TD><TD width=100  align=center style=\"background: #%06x\">%dus < </TD><TD width=100  align=center style=\"background: #%06x\">%dus < </TD><TD width=100  align=center style=\"background: #%06x\">-</TD></tr>\n", _col_ok_, _dur_ok_, _col_warn_, _dur_warn_, _col_danger_ );   \
    fprintf( fp, "</table><p>")



typedef struct{
	uint32_t ticks;
	uint32_t count;
} time_unit_t;

/* This structure stores all needed informations
 * to get max min and average duration of an event
 */
typedef struct{
	time_unit_t min;	/* Minimal duration 	*/
	time_unit_t tmin;	/* Date of Min duration	*/
	time_unit_t max;	/* Maximal duration 	*/
	time_unit_t tmax;	/* Date of Max duration	*/
	time_unit_t moy;	/* Average duration 	*/
	time_unit_t temp;
	uint32_t    nb;
} time_measure_t;

/* Thread info structure
 */
typedef struct {
   time_measure_t active;           /* active time measures     */
   time_measure_t running;          /* running time measures    */
   time_unit_t    thd_active_time;  /* overall active time save */ 
   uint8_t        state;            /* thread state             */
}thd_info_t;

thd_info_t thd_info[THREAD_ID_MAX];

static time_measure_t stat_vec_time[32];
static time_measure_t stat_isr_time[32];
static time_measure_t stat_dsr_time[32];
static time_measure_t stat_user_log_time[RTMON_PKT_EVENT_MAX];
static time_measure_t stat_priority_inversion_time;
static time_unit_t    stat_thd_time[THREAD_ID_MAX];
static time_unit_t    stat_thd_temp;


#define NEW_MIN      0x1
#define NEW_MAX      0x2
#define NEW_MINMAX   0x3

#define UPDATE_STATS(_time_measure_,_ret_,_ticks_, _count_) {  \
            _ret_ =  stat_comp_time( &(_time_measure_->temp),  \
                            &(_time_measure_->min ),           \
                            &(_time_measure_->max ) );         \
            if(_ret_ & NEW_MAX){                               \
                _time_measure_->tmax.ticks = _ticks_;          \
                _time_measure_->tmax.count = _count_;          \
            }                                                  \
            if(_ret_ & NEW_MIN){                               \
                _time_measure_->tmin.ticks = _ticks_;          \
                _time_measure_->tmin.count = _count_;          \
            }                                                  \
            _time_measure_->temp.count = UNAVAILABLE;          \
            _time_measure_->nb++;                              \
         }


/** Define a color 
 * 
 * @param time measured
 * @param time limit ok
 * @param time limit warn
 * @return color
 */
static uint32_t stat_set_color( uint32_t micros, 
                                uint32_t ok,
                                uint32_t warn)
{
	uint32_t color= 0;

	if(micros < ok)        color = COLOR_OK;
	else if(micros < warn) color = COLOR_WARNING;
	else                   color = COLOR_DANGER;

   return color;
}



/** translate ticks into microseconds
 * 
 * @param time unit in ticks and count
 * @return time in microseconds
 */
static uint32_t tick_2_micros(time_unit_t *temp)
{
    uint32_t micros = 0;
    uint64_t tempi  = 0;

    if( (temp->count >= 0)&&(temp->count <= hdr.tickperiod) )
    {
        //fprintf( stderr, PFX "%s (%d,%d) = ", __FUNCTION__,temp->ticks, temp->count);
        tempi = (uint64_t)(temp->ticks * hdr.tickperiod) + temp->count;
        tempi *= (uint64_t)(hdr.timefact_mult/10); // divided by 10 because timefact unit is 0,1ms
        tempi /= (uint64_t)hdr.timefact_div;
        micros = (uint32_t)tempi;
        //fprintf( stderr, "%d usec\n",micros);
    }

    return micros;
}

#if 0
static void stat_print_time(time_measure_t *ti )
{
      printf( "\tmin (%d,%d) - max (%d,%d) - ave (%d,%d) - temp (%d,%d)\n", 
              ti->min.ticks, ti->min.count,
              ti->max.ticks, ti->max.count,
              ti->moy.ticks, ti->moy.count,
              ti->temp.ticks, ti->temp.count );
}
#endif

/** Count elapsed time in ticks 
 *
 * @param previous time unit in ticks and count (which return computed time)
 * @param new tick value
 * @param new count value
 */
static void stat_count_time(time_unit_t *temp, uint32_t ticks, uint32_t count )
{
    // values are in the same tick unit
    if( ( temp->ticks == ticks))
    {
        temp->ticks = 0;

        if( temp->count <= count){
            temp->count = count - temp->count;
        }else{
            fprintf( stderr, PFX "ERROR %s\n",__FUNCTION__);
        }
    // values are in different tick units 
    }else if( temp->ticks < ticks){
        temp->ticks = ticks - temp->ticks -1;
        temp->count = (count + hdr.tickperiod) - temp->count;
        if(temp->count >= hdr.tickperiod)
        {
            temp->count -= hdr.tickperiod;
            temp->ticks += 1;
        }
    }else{
        fprintf( stderr, PFX "%s Error tick offset %d - %d\n",
                              __FUNCTION__,(int)temp->ticks, (int)ticks);
    }

}

/** Add time in ticks
 * 
 * @param time unit in ticks and count (which return computed time)
 * @param tick value
 * @param count value
 */
void stat_add_time(time_unit_t *temp, uint32_t ticks, uint32_t count )
{
   // add only if value is correct. Incorrect values can happen
   // in case of negative offsets
   if(count <= hdr.tickperiod)
   {
        temp->ticks += ticks;

        if( (temp->count+count) >= hdr.tickperiod){
            temp->ticks ++;
            temp->count = (temp->count + count) - hdr.tickperiod;
        }else{
            temp->count += count;
        }
   }
}


/** Compare a time unit to a min and a max
 *  
 * @param time unit (which return computed time)
 * @param minimum time value
 * @param maximum time value
 *
 * @retval -1 if min value has been changed
 * @retval  0 if neither min nor max values have been changed
 * @retval  1 if max value has been changed
 * @retval  2 if both min and max values have been changed
 */
int stat_comp_time(time_unit_t *temp, time_unit_t *min, time_unit_t *max )
{
    int ret = 0;

    if( min->ticks == UNAVAILABLE){
        min->count = temp->count;
        min->ticks = temp->ticks;
        max->count = temp->count;
        max->ticks = temp->ticks;
	    ret = NEW_MINMAX;
    }else if( ( temp->ticks == min->ticks) && ( temp->count < min->count) ){
        min->count = temp->count;
        ret = NEW_MIN;
    }else if( temp->ticks < min->ticks){
        min->ticks = temp->ticks;
        min->count = temp->count;
        ret = NEW_MIN;
    }else if( temp->ticks > max->ticks){
        max->count = temp->count;
        max->ticks = temp->ticks;
        ret = NEW_MAX;
    }else if( ( temp->ticks == max->ticks) &&  ( temp->count > max->count) ){
        max->count = temp->count;
        ret = NEW_MAX;
    }

    return ret;
}

/****************************************************************************** 
 * stat_log_user_start( )
 * ----------------------------------------------------------------------------
 * 
 *****************************************************************************/
void stat_log_user_start( uint32_t event, uint32_t ticks, uint32_t count )
{
   stat_user_log_time[event].temp.ticks = ticks;
   stat_user_log_time[event].temp.count = count;
}

/****************************************************************************** 
 * stat_log_user_stop( )
 * ----------------------------------------------------------------------------
 * 
 *****************************************************************************/
void stat_log_user_stop( uint32_t event, uint32_t ticks, uint32_t count )
{
   int ret;
   time_measure_t *user_event = &(stat_user_log_time[event]);

   //if buffer is not truncated
   if(user_event->temp.count != UNAVAILABLE)
   {
      // compute user_event duration
      stat_count_time( &(user_event->temp), ticks, count );
      // add duration to user_event average 
      stat_add_time( &(user_event->moy),
                       user_event->temp.ticks,
                       user_event->temp.count );
      // compare duration to min and max
      UPDATE_STATS(user_event, ret, ticks, count);
   }
}


/****************************************************************************** 
 * stat_priority_inversion_start( )
 * ----------------------------------------------------------------------------
 * 
 *****************************************************************************/
void stat_priority_inversion_start( uint32_t ticks, uint32_t count )
{
   stat_priority_inversion_time.temp.ticks = ticks;
   stat_priority_inversion_time.temp.count = count;
}

/****************************************************************************** 
 * stat_priority_inversion_stop( )
 * ----------------------------------------------------------------------------
 * 
 *****************************************************************************/
void stat_priority_inversion_stop( uint32_t ticks, uint32_t count )
{
   int ret = 0;
   time_measure_t *measure = &(stat_priority_inversion_time);

   //if buffer is not truncated
   if(measure->temp.count != UNAVAILABLE)
   {
      // compute user_event duration
      stat_count_time( &(measure->temp), ticks, count );
      // add duration to user_event average 
      stat_add_time( &(measure->moy),
                       measure->temp.ticks,
                       measure->temp.count );
      // compare duration to min and max
      UPDATE_STATS(measure, ret, ticks, count);
   }
}





/****************************************************************************** 
 * stat_isr( )
 * ----------------------------------------------------------------------------
 * 
 *****************************************************************************/
void stat_isr(  uint32_t    type,
                uint32_t    vec, 
                uint32_t    ticks, 
                uint32_t    count, 
                int         call_dsr )
{
   int ret = 0;
    time_measure_t *isr     = &(stat_isr_time[vec]);
    time_measure_t *latency = &(stat_vec_time[vec]);

    if ( type == RTMON_PKT_ISR_ENTER )
    {
        isr->temp.ticks = ticks;
        isr->temp.count = count;
    }
    else if(  (type == RTMON_PKT_ISR_EXIT )
            &&(isr->temp.count != UNAVAILABLE) )
    {
        stat_count_time( &(isr->temp), ticks, count );
        stat_add_time( &(isr->moy),isr->temp.ticks,isr->temp.count );

        UPDATE_STATS(isr, ret, ticks, count);

        if(call_dsr == 1){
            latency->temp.ticks = ticks;
            latency->temp.count = count;
        }else
            latency->temp.count = UNAVAILABLE;
    }

}

/****************************************************************************** 
 * stat_dsr( )
 * ----------------------------------------------------------------------------
 * 
 *****************************************************************************/
void stat_dsr(  uint32_t type, 
                uint32_t vec,
                uint32_t ticks, 
                uint32_t count )
{
   int ret = 0;
    time_measure_t *dsr     = &(stat_dsr_time[vec]);
    time_measure_t *latency = &(stat_vec_time[vec]);

    if ( type == RTMON_PKT_DSR_ENTER )
    {
        dsr->temp.ticks = ticks;
        dsr->temp.count = count;

        if(stat_vec_time[vec].temp.count != UNAVAILABLE)
        {
            stat_count_time( &(latency->temp), ticks, count );
            stat_add_time( &(latency->moy),latency->temp.ticks,latency->temp.count );
            UPDATE_STATS(latency, ret, ticks, count);
        }
    }
    else if(  (type == RTMON_PKT_DSR_EXIT )
            &&(stat_dsr_time[vec].temp.count != UNAVAILABLE) )
    {
        stat_count_time( &(dsr->temp), ticks, count );
        stat_add_time(  &(dsr->moy),dsr->temp.ticks,dsr->temp.count );
        UPDATE_STATS(dsr, ret, ticks, count);
    }
}

/****************************************************************************** 
 * stat_thread_activity( )
 * ----------------------------------------------------------------------------
 * Measures thread activity
 *****************************************************************************/
void stat_thread_activity( uint32_t ticks, uint32_t count,
                           uint32_t pid, uint8_t pstate, uint8_t preason,
                           uint32_t nid, uint8_t nstate, uint8_t nreason )
{
   int comp;

   // computes elapsed time
   if(thd_info[pid].active.temp.ticks!= UNAVAILABLE)
   {
      stat_count_time( &(thd_info[pid].active.temp), ticks, count );

      stat_add_time  ( &(thd_info[pid].active.moy),
                         thd_info[pid].active.temp.ticks,
                         thd_info[pid].active.temp.count );
   
      thd_info[pid].active.nb++;      
   }

   /*--------------------------------------------------------------------------
    * if previous thread goes "unactive", we gonna count elapsed time and add
    * it to statistics time
    *------------------------------------------------------------------------*/
   if(  ( pstate == SLEEPING )
      &&(thd_info[pid].running.temp.ticks != UNAVAILABLE) )
   { 
      time_measure_t *prevthd_time = &(thd_info[pid].running);

      // computes elapsed time
      stat_count_time( &(thd_info[pid].running.temp), ticks, count );
      // add time 
      stat_add_time  ( &(thd_info[pid].running.moy),
                       thd_info[pid].running.temp.ticks,
                       thd_info[pid].running.temp.count );

      UPDATE_STATS( prevthd_time, comp, ticks, count );
      
      if(comp & NEW_MAX) {
         thd_info[pid].active.max.ticks = thd_info[pid].active.moy.ticks;
         thd_info[pid].active.max.count = thd_info[pid].active.moy.count;
      }

      if(comp & NEW_MIN){
         thd_info[pid].active.min.ticks = thd_info[pid].active.moy.ticks;
         thd_info[pid].active.min.count = thd_info[pid].active.moy.count;
      }

      // reset 
      thd_info[pid].running.temp.ticks = UNAVAILABLE;
      
      // add actual active time to overall active time  
      stat_add_time  ( &(thd_info[pid].thd_active_time),
                       thd_info[pid].active.moy.ticks,
                       thd_info[pid].active.moy.count );
      
      // reset the amount of active time
      thd_info[pid].active.moy.ticks  = 0;
      thd_info[pid].active.moy.count  = 0;
   }

   thd_info[pid].active.temp.ticks = UNAVAILABLE;
   thd_info[pid].active.temp.count = 0;

   /*--------------------------------------------------------------------------
    * if next thread were not RUNNING(active), we gonna start counting time
    *------------------------------------------------------------------------*/
   if(thd_info[nid].state != RUNNING)
   {
      thd_info[nid].running.temp.ticks = ticks;
      thd_info[nid].running.temp.count = count;
   }

   //always start counting active time
   thd_info[nid].active.temp.ticks = ticks;
   thd_info[nid].active.temp.count = count;

   //new state is know the current state
   thd_info[nid].state = nstate;
   thd_info[pid].state = pstate;

}

/****************************************************************************** 
 * stat_thread_activity_close( )
 * ----------------------------------------------------------------------------
 * Terminate thread activity measures
 *****************************************************************************/
void stat_thread_activity_close( uint32_t i, uint32_t ticks, uint32_t count )
{
   int comp;

   if (  (thd_info[i].state == RUNNING)
       &&(thd_info[i].running.temp.count != UNAVAILABLE) //measure running
       &&(thd_info[i].running.min.count  == UNAVAILABLE) )//already measured event ) 
   {
      time_measure_t *prevthd_time = &(thd_info[i].running);

      // computes elapsed time
      stat_count_time( &(thd_info[i].running.temp), ticks, count );
      // add time 
      stat_add_time  ( &(thd_info[i].running.moy),
                       thd_info[i].running.temp.ticks,
                       thd_info[i].running.temp.count );

      UPDATE_STATS( prevthd_time, comp, ticks, count );

      if(comp & NEW_MAX) {
         thd_info[i].active.max.ticks = thd_info[i].active.moy.ticks;
         thd_info[i].active.max.count = thd_info[i].active.moy.count;
      }

      if(comp & NEW_MIN){
         thd_info[i].active.min.ticks = thd_info[i].active.moy.ticks;
         thd_info[i].active.min.count = thd_info[i].active.moy.count;
      }
   }
}


/****************************************************************************** 
 * stat_thread( )
 * ----------------------------------------------------------------------------
 * 
 *****************************************************************************/
void stat_thread( uint32_t id, uint32_t ticks, uint32_t count )
{
  if ( ( stat_thd_temp.count != UNAVAILABLE )
     &&( stat_thd_temp.ticks != UNAVAILABLE ) )
  {
      time_unit_t previous_time;

      previous_time.ticks = stat_thd_temp.ticks;
      previous_time.count = stat_thd_temp.count;

      // computes elapsed time
      stat_count_time( &(previous_time), ticks, count );
      // add time 
      stat_add_time  ( &(stat_thd_time[id]),
                       previous_time.ticks,
                       previous_time.count );
  }

   stat_thd_temp.ticks = ticks;
   stat_thd_temp.count = count;
}



/****************************************************************************** 
 * stat_init( )
 * ----------------------------------------------------------------------------
 * Initialize datas
 *****************************************************************************/

#define INIT_MEASURE(_measure_)         \
    _measure_.min.count = UNAVAILABLE;  \
    _measure_.min.ticks = UNAVAILABLE;  \
    _measure_.moy.count = 0;            \
    _measure_.moy.ticks = 0;            \
    _measure_.max.count = 0;            \
    _measure_.nb  = 0;                  \
    _measure_.temp.count = UNAVAILABLE; \
    _measure_.temp.ticks = UNAVAILABLE


void stat_init( )
{
   uint32_t i;

   for(i=0;i<RTMON_IRQ_MAX;i++){
      // interrupt traces
      INIT_MEASURE(stat_isr_time[i]);
      INIT_MEASURE(stat_dsr_time[i]);
      INIT_MEASURE(stat_vec_time[i]);
   }

   for(i=0;i<32;i++){
      INIT_MEASURE(stat_user_log_time[i]);
   }

   INIT_MEASURE(stat_priority_inversion_time);


   for(i=0;i<THREAD_ID_MAX;i++){
        stat_thd_time[i].ticks = 0;
        stat_thd_time[i].count = 0;
        thd_info[i].thd_active_time.ticks = 0;
        thd_info[i].thd_active_time.count = 0;
   }

   for(i=0;i<THREAD_ID_MAX;i++){
      INIT_MEASURE(thd_info[i].running);
      INIT_MEASURE(thd_info[i].active);
      thd_info[i].state = EXITED;
   }
   
   stat_thd_temp.count = UNAVAILABLE;
   stat_thd_temp.ticks = UNAVAILABLE;
}


/****************************************************************************** 
 * write_stat_latency( )
 * ----------------------------------------------------------------------------
 * Writes interrupt latency html report
 *****************************************************************************/
void write_stat_latency( FILE *fp)
{
    uint32_t        i,color, micros, time;
    time_measure_t  *latency = NULL;

    for(i=0;i<RTMON_IRQ_MAX;i++)
    {
        latency = &(stat_vec_time[i]);

        // interrupt traces
        if(latency->nb != 0)
        {
            micros = tick_2_micros(&(latency->min));
            time   = tick_2_micros(&(latency->tmin));
		      color = stat_set_color(micros,DSR_DUR_OK,DSR_DUR_WARNING);
            fprintf( fp, "<b>Latency of %s - IRQ vector %d:</b>",
                          hdr.irq_tag[i], (int)i);
            fprintf( fp, "<table border=0 cellspacing=2>"
                         "<tr><td width=50 align=right>min</td>"
                         "<td width=%d align=center style=\"background: #%06x\"></td>"
                         "<td >%dus</td><td ><tt>(reached at %dus)</tt></td></tr>\n</table>\n",
                         2*micros, (unsigned int)color, (int)micros, (int)time );

            micros = tick_2_micros(&(latency->moy)) / latency->nb;
		      color = stat_set_color(micros,DSR_DUR_OK,DSR_DUR_WARNING);
            fprintf( fp, "<table border=0 cellspacing=2>\n"
                         "<tr><td width=50 align=right>ave</td>"
                         "<td width=%d align=center style=\"background: #%06x\"></td>"
                         "<td >%dus</td></tr>\n</table>\n",
                         2*micros, (unsigned int)color, (int)micros );

            micros = tick_2_micros(&(latency->max));
            time   = tick_2_micros(&(latency->tmax));
		      color = stat_set_color(micros,DSR_DUR_OK,DSR_DUR_WARNING);
            fprintf( fp, "<table border=0 cellspacing=2>\n");
            fprintf( fp, "<tr><td width=50 align=right>max</td>"
                         "<td width=%d align=center style=\"background: #%06x\"></td>"
                         "<td >%dus</td><td ><tt>(reached at %dus)</tt></td></tr>\n</table>\n",
                             2*micros, (unsigned int)color, (int)micros, (int)time );

            fprintf( fp, "<p>\n");
        }
    }
}




/****************************************************************************** 
 * write_stat_dsr( )
 * ----------------------------------------------------------------------------
 * Writes interrupt DSR html report
 *****************************************************************************/
void write_stat_dsr( FILE *fp)
{
   uint32_t        i,color, micros, time;
   time_measure_t  *dsr = NULL;


   fprintf( fp, "DSR criticity characterization:<br>\n");
   FPRINT_CRITIC_TABLE(COLOR_OK, DSR_DUR_OK, COLOR_WARNING, DSR_DUR_WARNING, COLOR_DANGER );
   fprintf( fp, "</table><p>");

   for(i=0;i<RTMON_IRQ_MAX;i++)
   {
      dsr = &(stat_dsr_time[i]);

      // interrupt traces
      if(dsr->nb != 0)
      {
         micros = tick_2_micros(&(dsr->min));
         time   = tick_2_micros(&(dsr->tmin));
		   color = stat_set_color(micros,DSR_DUR_OK,DSR_DUR_WARNING);
         fprintf( fp, "<b>DSR %s - IRQ vector %d:</b>\n", hdr.irq_tag[i], i);
         fprintf( fp, "The following result is calculated from %d measures\n", dsr->nb);
         fprintf( fp, "<table border=0 cellspacing=2>"
                      "<tr><td width=50 align=right>min</td>"
                      "<td width=%d align=center style=\"background: #%06x\"></td>"
                      "<td>%dus</td><td ><tt>(reached at %dus)</tt></td></tr>\n</table>\n", 
                      2*micros, (unsigned int)color, (int)micros, (int)time  );

         micros = tick_2_micros(&(dsr->moy)) / dsr->nb;
		   color = stat_set_color(micros,DSR_DUR_OK,DSR_DUR_WARNING);
         fprintf( fp, "<table border=0 cellspacing=2>\n"
                      "<tr><td width=50 align=right>ave</td>"
                      "<td width=%d align=center style=\"background: #%06x\"></td>"
                      "<td >%dus</td></tr>\n</table>\n", 
                      2*micros, (unsigned int)color, (int)micros);

         micros = tick_2_micros(&(dsr->max));
         time   = tick_2_micros(&(dsr->tmax));
		   color = stat_set_color(micros,DSR_DUR_OK,DSR_DUR_WARNING);
         fprintf( fp, "<table border=0 cellspacing=2>\n"
                      "<tr><td width=50 align=right>max</td>"
                      "<td width=%d align=center style=\"background: #%06x\"></td>"
                      "<td>%dus</td><td ><tt>(reached at %dus)</tt></td></tr>\n</table>\n", 
                      2*micros, (unsigned int)color, (int)micros, (int)time   );
         fprintf( fp, "<p>\n");
      }
   }
}




/****************************************************************************** 
 * write_stat_isr( )
 * ----------------------------------------------------------------------------
 * Writes interrupt ISR html report
 *****************************************************************************/
void write_stat_isr( FILE *fp)
{
   uint32_t        i,color, micros, time;
   time_measure_t  *isr = NULL;

   fprintf( fp, "ISR criticity characterization:<br>\n");
   FPRINT_CRITIC_TABLE(COLOR_OK, ISR_DUR_OK, COLOR_WARNING, ISR_DUR_WARNING, COLOR_DANGER );

   for(i=0;i<RTMON_IRQ_MAX;i++)
   {
      isr = &(stat_isr_time[i]);

      if(stat_isr_time[i].nb != 0)
      {
         micros = tick_2_micros(&(isr->min));
         time   = tick_2_micros(&(isr->tmin));
         color  = stat_set_color(micros,ISR_DUR_OK,ISR_DUR_WARNING);
         fprintf( fp, "<b>ISR %s - IRQ vector %d:</b>\n", hdr.irq_tag[i], i);
         fprintf( fp, "The following result is calculated from %d measures\n", isr->nb);
         fprintf( fp, "<table border=0 cellspacing=2>"
                        "<tr><td width=50 align=right>min</td>"
                        "<td width=%d align=center style=\"background: #%06x\"></td>"
                        "<td>%dus</td><td ><tt>(reached at %dus)</tt></td></tr>\n</table>\n", 
                        7*micros, (unsigned int)color, (int)micros, (int)time );

         micros = tick_2_micros(&(isr->moy)) / isr->nb;

         color  = stat_set_color(micros,ISR_DUR_OK,ISR_DUR_WARNING);
         fprintf( fp, "<table border=0 cellspacing=2>\n"
                        "<tr><td width=50 align=right>ave</td>"
                        "<td width=%d align=center style=\"background: #%06x\"></td>"
                        "<td >%dus </td></tr>\n</table>\n", 
                        7*micros, (unsigned int)color, (int)micros);

         micros = tick_2_micros(&(isr->max));
         time   = tick_2_micros(&(isr->tmax));
         color = stat_set_color(micros,ISR_DUR_OK,ISR_DUR_WARNING);
         fprintf( fp, "<table border=0 cellspacing=2>\n"
                        "<tr><td width=50 align=right>max</td>"
                        "<td width=%d align=center style=\"background: #%06x\"></td>"
                        "<td>%dus</td><td > <tt>(reached at %dus)</tt></td></tr>\n</table>\n", 
                        7*micros, (unsigned int)color, (int)micros, (int)time );
         fprintf( fp, "<p>\n");
      }
   }
}


/****************************************************************************** 
 * write_stat_priority_inversion( )
 * ----------------------------------------------------------------------------
 * Writes priority inversion html report
 *****************************************************************************/
void write_stat_priority_inversion( FILE *fp, uint32_t trace_length)
{
    uint32_t        color, micros, time;
    time_measure_t  *measures = NULL;

    color  = COLOR_OK;
    measures = &(stat_priority_inversion_time);

    if(measures->nb != 0)
    {
        fprintf( fp, "<b>PRIORITY INVERSION:</b>\n"
                     "The following result is calculated from %d measures\n",
                     measures->nb);

        micros = tick_2_micros(&(measures->moy)) ;
        fprintf( fp, "<table border=0 cellspacing=2>\n"
                     "<tr><td width=90 align=right>cumulative</td>"
                     "<td width=%d align=center style=\"background: #%06x\"></td>"
                     "<td >%dus - %d %%</td></tr>\n</table>\n", 
                     micros/5000,THREAD_COLOR, (int)micros, (micros/10)/trace_length);

        micros = tick_2_micros(&(measures->min));
        time   = tick_2_micros(&(measures->tmin));
        fprintf( fp, "<table border=0 cellspacing=2>"
                     "<tr><td width=90 align=right>min</td>"
                     "<td width=%d align=center style=\"background: #%06x\"></td>"
                     "<td>%dus</td><td ><tt>(reached at %dus)</tt></td></tr>\n</table>\n", 
                     micros/5000, (unsigned int)color, (int)micros, (int)time );

        micros = tick_2_micros(&(measures->moy)) / measures->nb;
        fprintf( fp, "<table border=0 cellspacing=2>\n"
                     "<tr><td width=90 align=right>ave</td>"
                     "<td width=%d align=center style=\"background: #%06x\"></td>"
                     "<td >%dus </td></tr>\n</table>\n", 
                     micros/5000,(unsigned int)color, (int)micros);

        micros = tick_2_micros(&(measures->max));
        time   = tick_2_micros(&(measures->tmax));
        fprintf( fp, "<table border=0 cellspacing=2>"
                     "<tr><td width=90 align=right>max</td>"
                     "<td width=%d align=center style=\"background: #%06x\"></td>"
                     "<td>%dus</td><td ><tt>(reached at %dus)</tt></td></tr>\n</table>\n",  
                            micros/5000, (unsigned int)color, (int)micros, (int) time );

        fprintf( fp, "<p>\n");
    }
}



/****************************************************************************** 
 * write_stat_user_event( )
 * ----------------------------------------------------------------------------
 * Writes user event html report
 *****************************************************************************/
void write_stat_log_user( FILE *fp, uint32_t trace_length)
{
    uint32_t        i,color, micros, time;
    time_measure_t  *user_event = NULL;

    color  = COLOR_OK;

    for(i=0;i<RTMON_PKT_EVENT_MAX;i++)
    {
        user_event = &(stat_user_log_time[i]);

        if(stat_user_log_time[i].nb != 0)
        {
           fprintf( fp, "<b>EVENT %d:</b>\n", i);
           fprintf( fp, "The following result is calculated from %d measures\n",
                         user_event->nb);

           micros = tick_2_micros(&(user_event->moy)) ;
           fprintf( fp, "<table border=0 cellspacing=2>\n"
                        "<tr><td width=90 align=right>cumulative</td>"
                        "<td width=%d align=center style=\"background: #%06x\"></td>"
                        "<td >%dus - %d %%</td></tr>\n</table>\n", 
                        micros/5000,THREAD_COLOR, (int)micros, (micros/10)/trace_length);

           micros = tick_2_micros(&(user_event->min));
           time   = tick_2_micros(&(user_event->tmin));
           fprintf( fp, "<table border=0 cellspacing=2>"
                        "<tr><td width=90 align=right>min</td>"
                        "<td width=%d align=center style=\"background: #%06x\"></td>"
                        "<td>%dus</td><td ><tt>(reached at %dus)</tt></td></tr>\n</table>\n", 
                        micros/5000, (unsigned int)color, (int)micros, (int)time );

           micros = tick_2_micros(&(user_event->moy)) / user_event->nb;
           fprintf( fp, "<table border=0 cellspacing=2>\n"
                        "<tr><td width=90 align=right>ave</td>"
                        "<td width=%d align=center style=\"background: #%06x\"></td>"
                        "<td >%dus </td></tr>\n</table>\n", 
                        micros/5000,(unsigned int)color, (int)micros);

           micros = tick_2_micros(&(user_event->max));
           time   = tick_2_micros(&(user_event->tmax));
           fprintf( fp, "<table border=0 cellspacing=2>"
                        "<tr><td width=90 align=right>max</td>"
                        "<td width=%d align=center style=\"background: #%06x\"></td>"
                        "<td>%dus</td><td ><tt>(reached at %dus)</tt></td></tr>\n</table>\n",  
                        micros/5000, (unsigned int)color, (int)micros, (int) time );

            fprintf( fp, "<p>\n");
        }
    }
}



/****************************************************************************** 
 * write_stat_thread( )
 * ----------------------------------------------------------------------------
 * Writes user event html report
 *****************************************************************************/
void write_stat_thread( FILE *fp, uint32_t trace_length)
{
   uint32_t i, micros;
	uint32_t color = THREAD_COLOR;

   for(i=0;i<THREAD_ID_MAX;i++)
   {
      // interrupt traces
      if( (stat_thd_time[i].count != 0)||(stat_thd_time[i].ticks != 0) )
      {
         micros = tick_2_micros(&(stat_thd_time[i]));

         fprintf( fp, "<b> Thread ID %d - %s </b>", (int)i, thd_tag[i]);
         fprintf( fp, "<table border=0 cellspacing=2>"
                      "<tr><td width=50 align=right>%d %% </td>"
                      "<td width=%d align=center style=\"background: #%06x\"></td>"
                      "<td> %d,%03d ms</td></tr>\n</table>\n", 
                      (micros/10)/trace_length, micros/5000, (unsigned int)color, 
                      (int)(micros/1000), (int)micros%1000 );
      }
   }
}

/****************************************************************************** 
 * write_stat_thread_activity( )
 * ----------------------------------------------------------------------------
 * Writes user event html report
 *****************************************************************************/
void write_stat_thread_activity( FILE         *fp, 
                                 uint32_t     trace_length, 
                                 timer_unit_t *end_timer)
{
   uint32_t    i;
   thd_info_t  *ti;
   
   fprintf( fp, 
           "<table>\n"
           "<tr align=\"center\">"
           "<th width=60 ><tt>Thd id</tt></th>"
           "<th width=60 ><tt>Thd prio</tt></th>"
           "<th width=300><tt>Thd name</tt></th>"
           "<th width=60 ><tt>Nb meas run</tt></th>"
           "<th width=60 ><tt>Nb meas act</tt></th>"
           "<th width=60 ><tt>Cumul run</tt></th>"
           "<th width=60 ><tt>Cumul act</tt></th>"
           "<th width=60 ><tt>Min run(us)</tt></th>"
           "<th width=60 ><tt>ave run(us)</tt></th>"
           "<th width=60 ><tt>Max run(us)</tt></th>"
           "<th width=60 ><tt>Min act(us)</tt></th>"
           "<th width=60 ><tt>ave act(us)</tt></th>"
           "<th width=60 ><tt>Max act(us)</tt></th>"
           "</tr>\n"
           "</table>\n");

   for(i=0;i<THREAD_ID_MAX;i++)
   {
      ti = &(thd_info[i]);

      stat_thread_activity_close( i, end_timer->ticks, end_timer->count );

      // interrupt traces
      if(ti->running.nb != 0)
      {
          fprintf( fp,
                  "<table>\n"
                  "<tr>"
                  "<td width=60 align=\"center\"><tt>%d</tt></td>"
                  "<td width=60 align=\"center\"><tt>%c%c</tt></td>"
                  "<td width=300 align=\"left\"><tt>%s</tt></td>"
                  "<td width=60 align=\"right\"><tt>%d</tt></td>"
                  "<td width=60 align=\"right\"><tt>%d</tt></td>"
                  "<td width=60 align=\"right\"><tt>%d</tt></td>"
                  "<td width=60 align=\"right\"><tt>%d</tt></td>"
                  "<td width=60 align=\"right\"><tt>%d</tt></td>"
                  "<td width=60 align=\"right\"><tt>%d</tt></td>"
                  "<td width=60 align=\"right\"><tt>%d</tt></td>"
                  "<td width=60 align=\"right\"><tt>%d</tt></td>"
                  "<td width=60 align=\"right\"><tt>%d</tt></td>"
                  "<td width=60 align=\"right\"><tt>%d</tt></td>"
                  "</tr>\n"
                  "</table>\n",

                  (int)i,
                  thd_tag[i][0],
                  thd_tag[i][1],
                  &(thd_tag[i][3]),
                  ti->running.nb,
                  ti->active.nb, 
                  (int)tick_2_micros(&(ti->running.moy) ), 
                  (int)tick_2_micros(&(ti->thd_active_time)), 
                  (int)tick_2_micros(&(ti->running.min)), 
                  (int)(tick_2_micros(&(ti->running.moy)) / ti->running.nb), 
                  (int)tick_2_micros(&(ti->running.max)),
                  (int)tick_2_micros(&(ti->active.min)), 
                  // following division by "ti->running.nb" is COHERENT 
                  (int)(tick_2_micros(&(ti->thd_active_time)) / ti->running.nb), 
                  (int)tick_2_micros(&(ti->active.max))
                      );
      }
    }
}

void write_stat_file(const char * name)
{
   FILE     *fp;
   uint32_t color = BACKGD_COLOR;
   uint32_t begin_stat, end_stat;

   if(start_time !=0)
      begin_stat = start_time / 10000; //start_time unit is 100 ns
   else
      begin_stat = first_tick * 10;

   if(end_time !=0)
      end_stat = end_time / 10000; //end_time unit is 100 ns
   else
      end_stat = timer.ticks * 10;


   fp = fopen( name, "wb" );
   if ( fp != NULL ) 
   {
      fprintf( fp, "<!DOCTYPE HTML PUBLIC >\n"
                   "<html>\n"
                   "<head>\n"
                   "<title>RTMon report</title>\n"
                   "<meta http-equiv=\"content-type\" content=\"text/html; charset=iso-8859-1\">\n"
                   "<STYLE type=\"text/css\">.rubrique{ border: solid thin; text-align: center}</STYLE>\n"
                   "<STYLE type=\"text/css\">.critic, .critic TD, .critic TH{font-family:sans-serif;font-size:8pt;}</STYLE>\n"
                   "<STYLE type=\"text/css\">body   {font-family:sans-serif;font-size:10pt;} </STYLE>\n"
                   "</head>");

      fprintf( fp, "<body style=\"background: #%6x\">\n", (unsigned int)color);
      fprintf( fp, "<h1><center>RTMon report</center></h1>\n"
                   "<p>\n"
                   "<p>\n"
                   "</p><div align=\"center\">\n"
                   "<hr>\n"
                   "<a href=\"#ISR\">ISR</a> &nbsp;|&nbsp;\n"
                   "<a href=\"#DSR\">DSR</a> &nbsp;|&nbsp;\n"
                   "<a href=\"#LATENCY\">Latency</a> &nbsp;|&nbsp;\n"
                   "<a href=\"#THREAD\">Thread</a> &nbsp;|&nbsp;\n"
                   "<a href=\"#USEREVENT\">User Events</a>\n"
                   "<hr>\n"
                   "</div>\n");

      fprintf( fp, "<table border=0 cellspacing=4>");
      fprintf( fp, "<tr><td width=300 align=right><b>flush reason:</b></td><td>%s</td></tr>\n", hdr.flush_reason );
      fprintf( fp, "<tr><td width=300 align=right><b>trace length:</b></td><td>%dms</td></tr>\n", (end_stat-begin_stat) );
      fprintf( fp, "</table>");

      fprintf( fp, "<h2 class=\"rubrique\"><a name=\"ISR\">ISR Report</a></h2><br>\n"
                   "This measure give the minimal, average and maximum length of each ISR <p>\n");
      write_stat_isr(fp);

      fprintf( fp, "<h2 class=\"rubrique\"><a name=\"DSR\">DSR Report</a></h2><br>\n"
                   "This measure give the minimal, average and maximum length of each DSR <p>\n");
      write_stat_dsr(fp);

      fprintf( fp, "<h2 class=\"rubrique\"><a name=\"LATENCY\">Latency Report</a></h2>\n"
                   "This measure presents the latency between the "
                   "end of ISR and the beginning of DSR <p>\n");
      write_stat_latency(fp);

      fprintf( fp, "<h2 class=\"rubrique\"><a name=\"THREAD\">Thread Report</a></h2><br>\n"
                   "This measure give the minimal, average and maximum length of a priority inversion period <p>\n");
      write_stat_priority_inversion(fp, (end_stat-begin_stat));
      fprintf( fp, "<hr>");
      write_stat_thread(fp, (end_stat-begin_stat));

      if(process_thread_states)
      {
         fprintf( fp, "<h2 class=\"rubrique\"><a name=\"THREAD\">Thread Activity Report</a></h2><br>\n"
                   "This measure give the minimal, average and maximum length of a running period "
                   "and give the CPU active amount of time within this period.<p>\n");
         write_stat_thread_activity(fp, (end_stat-begin_stat), &timer);
      }

      fprintf( fp, "<h2 class=\"rubrique\"><a name=\"USEREVENT\">User Event Report</a></h2><br>\n"
                   "This measure give the minimal, average and maximum length of each User event <p>\n");
      write_stat_log_user(fp, (end_stat-begin_stat));

      fprintf( fp, "<hr>");
      fprintf( fp, "</body>");
      fprintf( fp, "</html>");

   }
}














