/**
*******************************************************************************
* @file  rtmon_common.h
* @brief Real-Time Monitor
*
* This package provides a real-time monitoring tool. This file provides common
* definitions and types shared between target and host code.
*
* Copyright (C) 2006 Parrot S.A.
*
* @author   ivan.djelic@parrot.fr
* @date     2006-12-01
* @author   gregoire.etienne@parrot.com
* @date     2007-06-14
*******************************************************************************
*/
#ifndef CYGONCE_RTMON_COMMON_H
#define CYGONCE_RTMON_COMMON_H

#define RTMON_PKT_IRQ_MAX          (32) //! Number of IRQ to monitor
#define RTMON_PKT_USER_U32_MAX     (32) //! Number of user events to monitor
#define RTMON_PKT_EVENT_MAX        (256) //! Number of user events to monitor
#define RTMON_PKT_SLOCK_MAX        (16) //! Number of IRQ to monitor


//! Log packet type encoding
enum {
    RTMON_PKT_NONE           = 0,
    RTMON_PKT_ISR_ENTER      = 1,
    RTMON_PKT_ISR_EXIT       = 2,
    RTMON_PKT_DSR_ENTER      = 3,
    RTMON_PKT_DSR_EXIT       = 4,
    RTMON_PKT_DRV_EVENT_START= 5,
    RTMON_PKT_DRV_EVENT_STOP = 6,
    RTMON_PKT_THREAD         = 7,
    RTMON_PKT_DATA24         = 8,
    RTMON_PKT_CALLER         = 9,
    RTMON_PKT_PC             = 10,
    RTMON_PKT_PROFILE        = 11, //this value is hardcoded in rtmon_handler.S
    RTMON_PKT_USER_U32       = 12,                                         //32 values 
    RTMON_PKT_SLOCK_0        = RTMON_PKT_USER_U32 + RTMON_PKT_USER_U32_MAX,//16 values
    RTMON_PKT_TICK           = RTMON_PKT_SLOCK_0  + RTMON_PKT_SLOCK_MAX,   // 0x3C
    RTMON_PKT_AAI_REG_ITS    = 0x3D,
    RTMON_PKT_AAI_REG_ITEN   = 0x3F,
    RTMON_PKT_AAI_REG_CFG    = 0x40,
    RTMON_PKT_AAI_PC         = 0x41,
    RTMON_PKT_USER0_START    = 0x42,
    RTMON_PKT_USER0_STOP     = 0x43,
    RTMON_PKT_USER0_U32      = 0x44,                                        //32 values 
    RTMON_PKT_MAX            = RTMON_PKT_USER0_U32 + RTMON_PKT_USER_U32_MAX,//16 values
};

//! RTMON_PKT_THREAD subtypes
enum {
	RTMON_THD_CS			   = 0,    //! Context switch
	RTMON_THD_CHGPRIO_ENTER = 1,    //! Priority modification (enter)
	RTMON_THD_CHGPRIO_EXIT	= 2,    //! Priority modification (exit)
	RTMON_THD_STATE         = 3,    //! Thread state
};



#define RTMON_PKT_FLAG_PEND_TICK   (31)

#define RTMON_PKT_TYPE_MASK        (0x7f)
#define RTMON_PKT_DATA24_MASK      (0xffffff)

 //! Magic is "RTmonLog"
#define RTMON_MAGIC                (0x676f4c6e6f6d5452LL)
//! tag string size
#define RTMON_TAGSIZE              (32)

#define RTMON_REASON_MAXLEN        (256)

//#define RTMON_PKT_TICK(_pkt_)       ((((_pkt_)>>24)&(RTMON_PKT_TYPE_MASK))==RTMON_PKT_TICK)
#define RTMON_PKT_HEADER(_pkt_)    ((_pkt_) >> 24)
#define RTMON_PKT_TYPE(_pkt_)      (((_pkt_) >> 24) & RTMON_PKT_TYPE_MASK)
#define RTMON_PKT_STAMP(_pkt_)     ((_pkt_) & 0xffff)
#define RTMON_PKT_DATA(_pkt_)      (((_pkt_) >> 16) & 0xff)
#define RTMON_PKT_DATA24(_pkt_)    ((_pkt_) & 0xffffff)

#define RTMON_ISR_CALLDSR_MASK     (0x80)
#define RTMON_ISR_VECTOR(_byte_)   ((_pkt_) & 0x7F)


//! user event tag list
typedef struct {
   int   id;
   char  tag[RTMON_TAGSIZE];
}event_tag_t;

//! RTmon file header
struct rtmon_header {
    cyg_uint64  magic;              //! RTmon magic
    cyg_uint32  size;               //! RTmon packet buffer size
    cyg_uint32  crc32;              //! RTmon packet buffer crc
    cyg_uint32  version;            //! RTmon version
    cyg_uint32  tickperiod;         //! eCos ticks period (usually 10ms)
    cyg_uint32  timefact_mult;
    cyg_uint32  timefact_div;
    cyg_uint32  nb_thread;          //! RTmon number of created thread
    cyg_uint32  nb_user_event_tag;  //! RTmon number of user event tag
    cyg_uint32  nb_user_u32_tag;    //! RTmon number of user u32 tag
    cyg_int32   timescale;
    char        irq_tag[RTMON_PKT_IRQ_MAX][RTMON_TAGSIZE];  //! IRQ tag table
    char        flush_reason[RTMON_REASON_MAXLEN];
} __attribute__ ((packed));


#endif /* CYGONCE_RTMON_COMMON_H */
