//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================

/**
*******************************************************************************
* @file  rtmon.c
* @brief Real-Time Monitor
*
* This package provides a real-time monitoring tool.
*
* Copyright (C) 2006 Parrot S.A.
*
* @author     ivan.djelic@parrot.fr
* @date       2006-12-01
*******************************************************************************
*/

#include <cyg/kernel/kapi.h>
#include <cyg/infra/diag.h>         // diagnostic output
#include <pkgconf/rtmon.h>
#include <cyg/rtmon/rtmon.h>
#include <cyg/rtmon/rtmon_common.h>
#include <cyg/rtmon/rtmon_internal.h>
#include <cyg/rtmon/rtmon_hal.h>
#include <cyg/crc/crc.h>
#include <cyg/hal/hal_diag.h>


#include <stdio.h>
#include <stdarg.h>

#ifdef CYGPKG_IO_FILEIO
#include <cyg/fileio/fileio.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#endif

static cyg_uint32 rtmon_buf[RTMON_BUFSIZE];

struct rtmon_info rtmon = {
    .buf            = rtmon_buf,
    .isr_mask       = RTMON_EVENT_NONE,
    .dsr_mask       = RTMON_EVENT_NONE,
    .event_mask     = RTMON_EVENT_NONE,
    .driver_mask    = RTMON_EVENT_NONE,
    .dsr_latency = { 0,0,0,0,0,0,0,0,
                     0,0,0,0,0,0,0,0,
                     0,0,0,0,0,0,0,0,
                     0,0,0,0,0,0,0,0 },
    .logfile     = NULL,
};

static struct rtmon_header hdr CYGBLD_ATTRIB_ALIGN(4);

static cyg_uint64 rtmon_ustamp[RTMON_PKT_EVENT_MAX];
static cyg_uint32 rtmon_ucount[RTMON_PKT_EVENT_MAX];
static cyg_uint32 rtmon_isr_end[RTMON_PKT_IRQ_MAX];
static int rtmon_dsr_pending[RTMON_PKT_IRQ_MAX];

//! thread tag list
static char rtmon_thread_tag[RTMON_THD_ID_MAX][RTMON_TAGSIZE];
static event_tag_t rtmon_user_event_tag[RTMON_PKT_EVENT_MAX];
static event_tag_t rtmon_user_u32_tag[RTMON_PKT_USER_U32_MAX];


//! IRQ tag list
static const char
rtmon_hal_irq_tag[RTMON_PKT_IRQ_MAX][RTMON_TAGSIZE] = RTMON_HAL_IRQ_TAGS;



/** ISR wrapper function
 */
cyg_uint32 rtmon_isr_wrapper( cyg_vector_t       vector,
                         cyg_addrword_t     data,
                         HAL_SavedRegisters *regs,
                         cyg_uint32         (* handler)( cyg_vector_t vector,
                                                         cyg_addrword_t data ))
{
    int log;
    cyg_uint32 ret;

    // no need to disable interrupts here
    log = ((1 << vector) & rtmon.isr_mask );

    if ( log )
    {
        cyg_uint32 _old;
        HAL_DISABLE_INTERRUPTS(_old);
        RTMON_LOG_STAMP( RTMON_PKT_ISR_ENTER, vector );
        HAL_RESTORE_INTERRUPTS(_old);
    }
    ret = (*handler)( vector, data );

    if ( log )
    {
        cyg_uint32 _old, pkt;
        HAL_DISABLE_INTERRUPTS(_old);

        if( ret & CYG_ISR_CALL_DSR )
        {
            pkt = RTMON_LOG_STAMP( RTMON_PKT_ISR_EXIT,
                                   (vector | RTMON_ISR_CALLDSR_MASK) );
            if ( rtmon.dsr_latency[vector] &&
                !rtmon_dsr_pending[vector])
            {
                rtmon_isr_end[vector] = RTMON_MKTIME32(pkt);
                rtmon_dsr_pending[vector] = 1;
            }
        }
        else
        {
            pkt = RTMON_LOG_STAMP( RTMON_PKT_ISR_EXIT, vector );
        }
        HAL_RESTORE_INTERRUPTS(_old);
    }

    return ret;
}

/** Log DSR enter
 */
void rtmon_log_dsr_enter( cyg_uint32 vector )
{
    cyg_uint32 old;
    cyg_uint64 latency = 0;
    int overflow = 0;

    HAL_DISABLE_INTERRUPTS(old);

    if ((1 << vector) & rtmon.dsr_mask ) {
        cyg_uint32 pkt = RTMON_LOG_STAMP( RTMON_PKT_DSR_ENTER, vector );

        if ( rtmon_dsr_pending[vector] ) {
            latency = RTMON_MKTIME32(pkt) - rtmon_isr_end[vector];
            rtmon_dsr_pending[vector] = 0;
            overflow = ( latency > rtmon.dsr_latency[vector] );
            if ( overflow ) {
                // make sure monitor is stopped
                rtmon_stop();
            }
        }
    }

    HAL_RESTORE_INTERRUPTS(old);

    RTMON_ASSERT( !overflow, "DSR %s latency overflow: %d clock ticks",
                  hdr.irq_tag[vector], latency );
}

/** Scheduler locking event
 */
void rtmon_log_slock( cyg_uint32 slock, cyg_uint32 func )
{
    if ( rtmon.event_mask & RTMON_EVENT_SCHED_LOCK )
    {
        cyg_uint32 old;
        HAL_DISABLE_INTERRUPTS(old);
        RTMON_LOG_STAMP((RTMON_PKT_SLOCK_0+slock) & RTMON_PKT_TYPE_MASK,
                        func >> 24 );
        RTMON_LOG_DATA24(func);
        HAL_RESTORE_INTERRUPTS(old);
    }
}

/** Log user value
 */
void rtmon_log_user_u32( cyg_uint32 user_event, cyg_uint32 value ) {
    cyg_uint32 old;
    user_event &= (RTMON_PKT_USER_U32_MAX-1);

    HAL_DISABLE_INTERRUPTS(old);

    if ( rtmon.event_mask & RTMON_EVENT_USER_U32){
        RTMON_LOG_STAMP( RTMON_PKT_USER_U32+user_event, value >> 24 );
        RTMON_LOG_DATA24(value);
    }

    HAL_RESTORE_INTERRUPTS(old);
}


/** Log driver event stop
 */
void rtmon_log_driver_stop( cyg_uint32 driver_event )
{
    cyg_uint32 old;
    driver_event &= (RTMON_PKT_EVENT_MAX-1);

    HAL_DISABLE_INTERRUPTS(old);

    if ( rtmon.driver_mask ){
        RTMON_LOG_STAMP( RTMON_PKT_DRV_EVENT_STOP, driver_event );
    }

    HAL_RESTORE_INTERRUPTS(old);
}

/** Log driver event start
 */
void rtmon_log_driver_start(cyg_uint32 driver_event)
{
    cyg_uint32 old;
    driver_event &= (RTMON_PKT_EVENT_MAX-1);

    HAL_DISABLE_INTERRUPTS(old);

    if ( rtmon.driver_mask ){
        RTMON_LOG_STAMP( RTMON_PKT_DRV_EVENT_START, driver_event );
    }

    HAL_RESTORE_INTERRUPTS(old);
}

/** Log user event stop
 */
void rtmon_log_user_stop( cyg_uint32 user_event )
{
    cyg_uint32 old;
    user_event &= (RTMON_PKT_EVENT_MAX-1);

    HAL_DISABLE_INTERRUPTS(old);
    if ( rtmon.event_mask & RTMON_EVENT_USER){
        RTMON_LOG_STAMP( RTMON_PKT_USER0_STOP, user_event );
    }
    HAL_RESTORE_INTERRUPTS(old);
}

/** Log user event start
 */
void rtmon_log_user_start(cyg_uint32 user_event)
{
    cyg_uint32 old;
    user_event &= (RTMON_PKT_EVENT_MAX-1);

    HAL_DISABLE_INTERRUPTS(old);
    if ( rtmon.event_mask & RTMON_EVENT_USER){
        RTMON_LOG_STAMP( RTMON_PKT_USER0_START, user_event );
    }
    HAL_RESTORE_INTERRUPTS(old);
}



/** Log user event start
 */
void rtmon_log_user_start_timed(cyg_uint32 user_event) {
    cyg_uint32 old;
    user_event &= (RTMON_PKT_EVENT_MAX-1);

    HAL_DISABLE_INTERRUPTS(old);

    if ( rtmon.event_mask & RTMON_EVENT_USER){
        cyg_uint32 pkt = RTMON_LOG_STAMP( RTMON_PKT_USER0_START, user_event );
        if ( !rtmon_ucount[user_event] ) {
            // store time stamp
            rtmon_ustamp[user_event] = RTMON_MKTIME64(pkt);
        }
        rtmon_ucount[user_event]++;
    }

    HAL_RESTORE_INTERRUPTS(old);
}

/** Log user event stop
 */
void rtmon_log_user_stop_timed( cyg_uint32 user_event, cyg_uint32 tlimit )
{
    cyg_uint32 old;
    cyg_uint32 pkt = 0;
    cyg_uint64 ustamp = 0;
    int overflow = 0;


    HAL_DISABLE_INTERRUPTS(old);

    user_event &= (RTMON_PKT_EVENT_MAX-1);

    if ( rtmon.event_mask & RTMON_EVENT_USER){
        pkt = RTMON_LOG_STAMP( RTMON_PKT_USER0_STOP, user_event );
        rtmon_ucount[user_event]--;

        if ( !rtmon_ucount[user_event] && tlimit ) {
            // check for time overflow
            ustamp = RTMON_MKTIME64(pkt);
            // convert us to timer units
            tlimit = RTMON_HAL_US2TIME(tlimit);
            overflow = ( ustamp > tlimit + rtmon_ustamp[user_event] );
            if ( overflow ) {
                // make sure monitor is stopped
                rtmon_stop();
            }
        }
    }

    HAL_RESTORE_INTERRUPTS(old);

    RTMON_ASSERT( !overflow, "User #%d event overflow: "
                  "start @ %lld, stop @ %lld, delta = %d",
                  rtmon_ustamp[user_event], ustamp,
                  (int)(ustamp - rtmon_ustamp[user_event]));
}

/** Check DSR count and fail if count > 0
 */
void rtmon_log_check_dsr_count(cyg_uint32 vector)
{
    RTMON_ASSERT( !(rtmon.event_mask & RTMON_EVENT_CHECK_DSR_CNT),
                  "Two consecutive DSR posts for IRQ %d !\n", vector );
}


/**Add a thread tag of the created task to the
 * thread tag list.
 */
void rtmon_set_thread_tag( cyg_uint32 id, const char *name, cyg_uint32 prio  )
{
    if ( id < RTMON_THD_ID_MAX )
    {
        int i = 0;

        hdr.nb_thread++;

        rtmon_thread_tag[id][i++] = (prio/10) + '0';
        rtmon_thread_tag[id][i++] = (prio%10) + '0';
        rtmon_thread_tag[id][i++] = '-';
        strncpy( &(rtmon_thread_tag[id][i]), name, (RTMON_TAGSIZE-1-i) );
        rtmon_thread_tag[id][RTMON_TAGSIZE-1] = '\0';
    }
}

/** Add a tag to a user event
 */
void rtmon_set_user_event_label( cyg_uint32 user_event, const char *name  )
{
    if (  (user_event < RTMON_PKT_EVENT_MAX )
       && (hdr.nb_user_event_tag < RTMON_PKT_EVENT_MAX) )
    {
        cyg_uint32 old;

        HAL_DISABLE_INTERRUPTS(old);
        if( rtmon_user_event_tag[user_event].tag[0] == 0 ){
           strncpy( rtmon_user_event_tag[user_event].tag, name, (RTMON_TAGSIZE-1) );
           rtmon_user_event_tag[user_event].tag[RTMON_TAGSIZE-1] = '\0';
           rtmon_user_event_tag[user_event].id = user_event;
           hdr.nb_user_event_tag++;
        }
        HAL_RESTORE_INTERRUPTS(old);
    }
}

/** Add a tag to a user u32 value
 */
void rtmon_set_user_u32_label( cyg_uint32 user_event, const char *name  )
{
    if (  (user_event < RTMON_PKT_USER_U32_MAX )
       && (hdr.nb_user_u32_tag < RTMON_PKT_USER_U32_MAX) )
    {
        cyg_uint32 old;

        HAL_DISABLE_INTERRUPTS(old);
        if( rtmon_user_u32_tag[user_event].tag[0] == 0 ){
           strncpy( rtmon_user_u32_tag[user_event].tag, name, (RTMON_TAGSIZE-1) );
           rtmon_user_u32_tag[user_event].tag[RTMON_TAGSIZE-1] = '\0';
           rtmon_user_u32_tag[user_event].id = user_event;
           hdr.nb_user_u32_tag++;
        }
        HAL_RESTORE_INTERRUPTS(old);
    }
}


/** This function configure and start RTmon
 */
void rtmon_start( struct rtmon_cfg *cfg )
{
    int             i;
    cyg_uint32      old;
    cyg_handle_t    thread;
    cyg_thread_info info;

    rtmon_stop();

    rtmon_hal_init_var();

    diag_printf( "rtmon: isr_mask=0x%08x dsr_mask=0x%08x events=0x%08x\n",
                 cfg->isr_mask, cfg->dsr_mask, cfg->event_mask );
    diag_printf( "rtmon: dsr_latency[] = { " );

    for ( i = 0; i < RTMON_PKT_IRQ_MAX; i++ )
    {
        if ( cfg->dsr_latency[i] )
        {
            diag_printf( "[%s] <= %d us, ", rtmon_hal_irq_tag[i],
                         cfg->dsr_latency[i] );
        }
    }
    diag_printf( "}\n" );

    if ( cfg->logfile )
    {
        rtmon.logfile = cfg->logfile;
        diag_printf( "rtmon: logfile=\"%s\"\n", cfg->logfile );
    }

    rtmon.buf_index = 0;

    RTMON_ASSERT( rtmon.buf == rtmon_buf, "Corrupted RTmon variables !" );

    memset( rtmon.buf           , 0, RTMON_BUFSIZE*sizeof(cyg_uint32) );
    memset( rtmon_thread_tag    , 0, (RTMON_THD_ID_MAX*RTMON_TAGSIZE) );
    memset( rtmon_user_event_tag, 0, (sizeof(event_tag_t)*RTMON_PKT_EVENT_MAX) );
    memset( rtmon_user_u32_tag  , 0, (sizeof(event_tag_t)*RTMON_PKT_USER_U32_MAX) );
    rtmon.ticks = cyg_current_time();

    // adjust time limits
    for ( i = 0; i < RTMON_PKT_IRQ_MAX; i++ )
    {
        rtmon.dsr_latency[i] = RTMON_HAL_US2TIME(cfg->dsr_latency[i]);
    }
    for ( i = 0; i < RTMON_PKT_EVENT_MAX; i++ )
    {
        rtmon_ucount[i] = 0;
    }

    // setup time scale and factor
    hdr.timescale     = RTMON_HAL_TIMESCALE;
    hdr.tickperiod    = RTMON_HAL_TICKPERIOD;
    hdr.timefact_mult = RTMON_HAL_TIMEFACT_MULT;
    hdr.timefact_div  = RTMON_HAL_TIMEFACT_DIV;

    // fill-in IRQ tags
    for ( i = 0; i < RTMON_PKT_IRQ_MAX; i++ )
    {
        strncpy( hdr.irq_tag[i], rtmon_hal_irq_tag[i], RTMON_TAGSIZE );
        hdr.irq_tag[i][RTMON_TAGSIZE-1] = '\0';
        rtmon_dsr_pending[i] = 0;
    }

    // retrieve thread tags (names)
    hdr.nb_thread = 1;

    for ( i = 0; i < RTMON_THD_ID_MAX; i++ ) {
        // find a thread with this ID
        thread = cyg_thread_find(i);
        if ( thread && ( cyg_thread_get_info( thread, i, &info ) != false )) {
            rtmon_set_thread_tag( i, info.name, info.set_pri );
        }
    }

    // initialize nb user event value
    hdr.nb_user_event_tag = 0;
    hdr.nb_user_u32_tag   = 0;

    HAL_DISABLE_INTERRUPTS(old);

    // enable all events
    rtmon.isr_mask   = cfg->isr_mask;
    rtmon.dsr_mask   = cfg->dsr_mask;
    rtmon.event_mask = RTMON_EVENT_TICK|cfg->event_mask;
    rtmon.driver_mask= cfg->driver_mask;

    // log first tick packet
    RTMON_LOG_TICK(rtmon.ticks);

    // start tick timer
    rtmon_hal_start_timer();
    /* Increment counter */


    if (rtmon.event_mask & RTMON_EVENT_PROFILING) {
        rtmon_hal_start_profiling();
    }

    HAL_RESTORE_INTERRUPTS(old);

    // initial context switch information
    RTMON_LOG_CS(cyg_thread_get_current_priority(cyg_thread_self()),
                 cyg_thread_get_id(cyg_thread_self()) );
}


/** This function stops the monitor. It can only be
 * restarted calling <tt>rtmon_start()</tt>.
 */
void rtmon_stop(void)
{
    cyg_uint32 old;

    if (rtmon.event_mask & RTMON_EVENT_PROFILING) {
        rtmon_hal_stop_profiling();
    }

    // stop tick timer
    rtmon_hal_stop_timer();

    HAL_DISABLE_INTERRUPTS(old);

    // disable all events
    rtmon.isr_mask   = RTMON_EVENT_NONE;
    rtmon.dsr_mask   = RTMON_EVENT_NONE;
    rtmon.event_mask = RTMON_EVENT_NONE;
    rtmon.driver_mask= RTMON_EVENT_NONE;
    memset( rtmon.dsr_latency, 0, RTMON_PKT_IRQ_MAX*sizeof(cyg_uint32));
    rtmon.logfile    = NULL;

    HAL_RESTORE_INTERRUPTS(old);
}

static void
rtmon_fix_pkts(void) {
    cyg_uint32 i, pkt;
    for ( i = 0; i < RTMON_BUFSIZE; i++ ) {
        pkt = rtmon.buf[i];
        if ( ( RTMON_PKT_TYPE(pkt) != RTMON_PKT_TICK ) &&
             ( RTMON_PKT_TYPE(pkt) != RTMON_PKT_NONE ) &&
             ( RTMON_PKT_TYPE(pkt) != RTMON_PKT_DATA24 ))
        {
            if( (RTMON_TIMER_PERIOD) < (pkt & 0xffff) )
                diag_printf( "rtmon: invalid stamp: period=0x%08x pkt=0x%08x\n", RTMON_TIMER_PERIOD, pkt);
            else
                rtmon.buf[i] = RTMON_FIX_PKT(pkt);
        }
    }
}

/** This function computes crc
 */
static cyg_uint32 rtmon_buf_crc32(cyg_uint32 *len)
{
    cyg_uint8 *p;
    cyg_uint32 index, i, crc;

    index = rtmon.buf_index;
    cyg_posix_crc_start( &crc, len );

    for ( i = 0; i < RTMON_BUFSIZE; i++ )
    {
        if ( RTMON_PKT_HEADER(rtmon.buf[index]) != RTMON_PKT_NONE )
        {
            p = (cyg_uint8 *)&(rtmon.buf[index]);
            cyg_posix_crc_cont( &crc, len, (char*)p, sizeof(cyg_uint32) );
        }
        index++;
        if ( index >= RTMON_BUFSIZE )
        {
            index = 0;
        }
    }
    return cyg_posix_crc_end( &crc, len );
}


/** This function sends the packet buffer through diag device
 */
static void rtmon_send_buf(void) {
    cyg_uint8 *p;
    cyg_uint32 index, i, j;

    index = rtmon.buf_index;

    for ( i = 0; i < RTMON_BUFSIZE; i++ )
    {
        if ( RTMON_PKT_HEADER(rtmon.buf[index]) != RTMON_PKT_NONE )
        {
            p = (cyg_uint8 *)&(rtmon.buf[index]);
            for ( j = 0; j < sizeof(cyg_uint32); j++ )
            {
                HAL_DIAG_WRITE_CHAR(p[j]);
            }
        }
        index++;
        if ( index >= RTMON_BUFSIZE )
        {
            index = 0;
        }
    }
}

/** This function sends the thread tag buffer through diag device
 *  @param nb_thd_max number of thread created since application started
 */
static void rtmon_send_thd_tag(cyg_uint32 nb_thd_max)
{
   cyg_uint32 i_char;

   for ( i_char = 0; i_char < (RTMON_TAGSIZE*nb_thd_max); i_char++ )
   {
      HAL_DIAG_WRITE_CHAR(((cyg_uint8 *)&rtmon_thread_tag)[i_char]);
   }

}

/** This function sends the user event tag buffer through diag device
 *  @param nb_event_tag_max number of thread created since application started
 */
static void rtmon_send_user_event_tag(cyg_uint32 nb_event_tag_max)
{
    cyg_uint32 i_event, i_char;

    for ( i_event = 0; i_event < RTMON_PKT_EVENT_MAX; i_event++ )
    {
       if(rtmon_user_event_tag[i_event].tag[0] != 0 )
       {
          for ( i_char = 0; i_char < sizeof(event_tag_t); i_char++ )
          {
            HAL_DIAG_WRITE_CHAR(((cyg_uint8 *)&rtmon_user_event_tag[i_event])[i_char]);
          }
       }
    }
}

/** This function sends the user u32 tag buffer through diag device
 *  @param nb_thd_max number of thread created since application started
 */
static void rtmon_send_user_u32_tag(cyg_uint32 nb_u32_tag_max)
{
    cyg_uint32 i_event, i_char;

    for ( i_event = 0; i_event < RTMON_PKT_USER_U32_MAX; i_event++ )
    {
       if(rtmon_user_u32_tag[i_event].tag[0] != 0 )
       {
          for ( i_char = 0; i_char < sizeof(event_tag_t); i_char++ )
          {
            HAL_DIAG_WRITE_CHAR(((cyg_uint8 *)&rtmon_user_u32_tag[i_event])[i_char]);
          }
       }
    }
}

#ifdef CYGPKG_IO_FILEIO
/** This function write RTMon buffer to a file
 */
static int rtmon_write_buf(int fd)
{
    int ret = 0;
    cyg_uint32 index, i;

    index = rtmon.buf_index;

    for ( i = 0; i < RTMON_BUFSIZE; i++ ) {
        if ( RTMON_PKT_HEADER(rtmon.buf[index]) != RTMON_PKT_NONE ) {
            ret = write( fd, &rtmon.buf[index], sizeof(cyg_uint32) );
            if ( ret != sizeof(cyg_uint32) ) {
                ret = -1;
                break;
            }
        }
        index++;
        if ( index >= RTMON_BUFSIZE ) {
            index = 0;
        }
    }
    return ret;
}
#endif

/** This function sends the whole RTMon buffer on diagnostic device
 */
void rtmon_flush( const char *fmt, ... ) {
    cyg_uint32 i;
    va_list ap;

    cyg_uint32 old;

    HAL_DISABLE_INTERRUPTS(old);

    // stop all logging activity
    rtmon_stop();

    // store reason for flush
    va_start( ap, fmt );
    vsnprintf( hdr.flush_reason, RTMON_REASON_MAXLEN, fmt, ap );
    va_end(ap);
    hdr.flush_reason[RTMON_REASON_MAXLEN-1] = '\0';

    diag_printf( "rtmon: flushing buffer because:\nrtmon: %s\n",
                 hdr.flush_reason );

    // fix logfile header
    hdr.magic = RTMON_MAGIC;
    hdr.version = RTMON_VERSION;
    // fix packet time stamps
    rtmon_fix_pkts();
    // compute CRC and size
    hdr.crc32 = rtmon_buf_crc32(&hdr.size);

    diag_printf( "rtmon: bufsize = %d bytes, crc = 0x%08x\n",
                  hdr.size, hdr.crc32 );
    diag_printf( "rtmon: version = %d, nb_thread = %d \n",
                  hdr.version, hdr.nb_thread );

#ifdef CYGPKG_IO_FILEIO

    if ( rtmon.logfile ) {
        cyg_uint32 status;

        // check if interrupts are disabled or if scheduler is locked
        HAL_QUERY_INTERRUPTS(status);
#if defined(CYGPKG_HAL_ARM)
        // on ARM platforms we only want to test CPSR bit I
        status &= CPSR_IRQ_DISABLE;
#endif

        if ( !status ||( cyg_scheduler_read_lock() != 0 )) {
            // scheduler is locked, or worse, interrupts are disabled,
            // which means we probably cannot access the filesystem
            // do nothing...
        }
        else {
            int fd, hdrsize = sizeof(struct rtmon_header);

            // ok it looks like we have been called from a thread context
            // let's first boost up our priority
            cyg_thread_set_priority( cyg_thread_self(), 0 );

            // then try to flush buffer
            diag_printf( "rtmon: writing buffer to '%s'...\n", rtmon.logfile );

            fd = open( rtmon.logfile, O_WRONLY|O_TRUNC, 0644 );
            if (( fd < 0 )||
                ( write( fd, &hdr, hdrsize ) != hdrsize )||
                ( rtmon_write_buf(fd) < 0 )) {
                diag_printf( "rtmon: error writing file '%s': %s\n",
                             rtmon.logfile, strerror(errno));
            }
            if ( fd >= 0 ) {
                close(fd);
            }
        }
    }

#endif

   if(hdr.size == 0 )
   {
      diag_printf( "rtmon: buffer size is %d, please check \n", hdr.size );
      diag_printf( "rtmon: that rtmon_start has been called \n" );
      while (1) {};
   }
   else
   {
    diag_printf( "rtmon: sending buffer to diag channel...\n" );

    // we do not want to be interrupted here
    cyg_scheduler_lock();

    // send file ad infinitum, host should eventually catch magic header
    while (1) {

        // send file header
        for ( i = 0; i < sizeof(struct rtmon_header); i++ )
        {
            HAL_DIAG_WRITE_CHAR(((cyg_uint8 *)&hdr)[i]);
        }

        //send thread tag buffer
        rtmon_send_thd_tag(hdr.nb_thread);
        //send user event tag buffer
        rtmon_send_user_event_tag(hdr.nb_user_event_tag);
        //send user u32 tag buffer
        rtmon_send_user_u32_tag(hdr.nb_user_u32_tag);


        // send buffer
        rtmon_send_buf();

        // insert a 1 second break
        HAL_DELAY_US(1000000);
    }
    // never reached for now
    cyg_scheduler_unlock();

   }
}

cyg_uint32
rtmon_get_max_buffer_size(void)
{
    cyg_uint32 log_size = 0;

    log_size += sizeof(struct rtmon_header);
    log_size += sizeof(rtmon_thread_tag);
    log_size += sizeof(rtmon_user_event_tag);
    log_size += sizeof(rtmon_user_u32_tag);
    log_size += sizeof(rtmon_buf);

    return log_size;
}

cyg_uint32
rtmon_copy_log(void *log_data, const char *fmt, ...)
{
    cyg_uint32 log_size = 0;
    cyg_uint32 i, index;
    va_list ap;
    cyg_uint8 *log_data8 = (cyg_uint8*)log_data;
    cyg_uint32 old;

    //We do not want to be interrupted here
    HAL_DISABLE_INTERRUPTS(old);
    cyg_scheduler_lock();

    //Store flush reason
    va_start(ap, fmt);
    vsnprintf(hdr.flush_reason, RTMON_REASON_MAXLEN, fmt, ap);
    va_end(ap);
    hdr.flush_reason[RTMON_REASON_MAXLEN - 1] = '\0';

    //Fix logfile header
    hdr.magic = RTMON_MAGIC;
    hdr.version = RTMON_VERSION;

    //Fix packet time stamps
    rtmon_fix_pkts();

    //Compute CRC and size
    hdr.crc32 = rtmon_buf_crc32(&hdr.size);

    //Copy file header
    memcpy(&log_data8[log_size], &hdr, sizeof(struct rtmon_header));
    log_size += sizeof(struct rtmon_header);

    //Copy thread tag buffer
    memcpy(&log_data8[log_size], &rtmon_thread_tag,
           RTMON_TAGSIZE * hdr.nb_thread);
    log_size += RTMON_TAGSIZE * hdr.nb_thread;

    //Copy user event tag buffer
    for(i = 0; i < RTMON_PKT_EVENT_MAX; i++) {
        if(rtmon_user_event_tag[i].tag[0] != 0) {
            memcpy(&log_data8[log_size], &rtmon_user_event_tag[i],
                sizeof(rtmon_user_event_tag[0]));
            log_size += sizeof(rtmon_user_event_tag[0]);
        }
    }

    //Copy user u32 tag buffer
    for(i = 0; i < RTMON_PKT_USER_U32_MAX; i++) {
        if(rtmon_user_u32_tag[i].tag[0] != 0) {
            memcpy(&log_data8[log_size], &rtmon_user_u32_tag[i],
                sizeof(rtmon_user_u32_tag[0]));
            log_size += sizeof(rtmon_user_u32_tag[0]);
        }
    }

    //Copy buffer
    index = rtmon.buf_index;
    for(i = 0; i < RTMON_BUFSIZE; i++) {
        if(RTMON_PKT_HEADER(rtmon.buf[index]) != RTMON_PKT_NONE) {
            memcpy(&log_data8[log_size], &rtmon.buf[index],
                sizeof(rtmon.buf[0]));
            log_size += sizeof(rtmon.buf[0]);
        }

        index++;
        if(index >= RTMON_BUFSIZE) {
            index = 0;
        }
    }

    //Re-enable everything
    cyg_scheduler_unlock();
    HAL_RESTORE_INTERRUPTS(old);

    return log_size;
}
